/*
 
 Copyright (C) 2010,2011 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */


#include "delve.h"
#include "interface.h"
#include "dbio.h"
#include "2ndorder.h"
#include "pHMM.h"
#include "nuc_code.h"
#include "seed.h"
#include "core.h"
#include "scoreboard.h"
#include "read_sam.h"
#include "align_to_sam.h"
#include <ctype.h>
#include "align.h"

pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
int  counter = 0;

int phmm2_posterior_controller(struct db* db,struct parameters* param)
{
	struct read_info** ri = 0;
	FILE* file = 0;
	int i,j,size,numseq = 0;
	
	//float mapped_sum = 0.0f;
	
	init_logsum();
	struct hmm2* hmm = init_2nd_hmm(500, 500);
	
	struct genome_probabilities* gp = 0;	
	
	struct rescue_probabilities* rp = 0;
	
	if(param->local_weighting){
		rp =  init_rp(db->seq_info->total_len ,param->local_weighting);
	}
	//fprintf(stderr,"malloced local...\n");
	ri = malloc(sizeof(struct read_info*) * param->num_query);
	
	for(i = 0; i < param->num_query;i++){
		ri[i] = malloc(sizeof(struct read_info));
		ri[i]->seq = 0;
		ri[i]->name = 0;
		ri[i]->qual = 0;
		ri[i]->len = 0;
		ri[i]->priors = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->strand = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->hits = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->identity = malloc(sizeof(float)* (LIST_STORE_SIZE+1));
	}
	
	gp = malloc(sizeof(struct genome_probabilities));
	
	size = (db->seq_info->total_len/ 1024) + 1;
	
	gp->plus = malloc(sizeof(float*) * size);
	gp->minus = malloc(sizeof(float*) * size);
	gp->num_reads = 0;
	for(i = 0;i < size;i++){
		gp->plus[i] = 0;
		gp->minus[i] = 0;
	}
	
	
	
	if (!(file = fopen(param->infile[1] , "r" ))){
		fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[1]);
		exit(-1);
	}
	fprintf(stderr,"Reading Model\n");
	read_hmm2_from_file(hmm,param);
	read_rescue_probabilities(rp, db->seq_info->total_len ,param->local_weighting, param);
		
	
	while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
		fprintf(stderr,"Read: %d\n",numseq);
		for(i = 0;i < numseq;i++){
			if(ri[i]->strand[0] != 0){
				ri[i]->seq = reverse_complement2(ri[i]->seq,ri[i]->len);
			}
			for(j = 0;j < ri[i]->len;j++){
				if(ri[i]->seq[j] == 4){
					ri[i]->seq[j] = NCODE;
				}else if(j >= ri[i]->len-BORDER_REGION){
					ri[i]->seq[j] += 8;
				}else if(j  >= BORDER_REGION){
					ri[i]->seq[j] += 4;
				}				
			}
		}
		
		//ri = add_k(ri, db,numseq);
		if(param->local_weighting){
			ri = set_local_prob(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq);
		}
		
		
		hmm = run_posterior2(hmm,ri,db,gp, param,numseq);
	}
	
	//normalization... 
	
	/*mapped_sum = prob2scaledprob(0.0f);
	
	for(i = 0;i < size;i++){
		if(gp->plus[i]){
			for(j = 0; j < 1024;j++){
				mapped_sum = logsum(mapped_sum, gp->plus[i][j]);
			}
			
		}
		
		if(gp->minus[i]){
			for(j = 0; j < 1024;j++){
				mapped_sum = logsum(mapped_sum, gp->minus[i][j]);
			}
			
		}
	}
	
	for(i = 0;i < size;i++){
		if(gp->plus[i]){
			for(j = 0; j < 1024;j++){
				gp->plus[i][j] -= mapped_sum;
				//mapped_sum = logsum(mapped_sum, gp->plus[i][j]);
				//gp->plus[i][j] = floor(gp->plus[i][j] *1000+0.5)/1000;
			}
			
		}
		
		if(gp->minus[i]){
			for(j = 0; j < 1024;j++){
				gp->minus[i][j] -= mapped_sum;
				
				//gp->minus[i][j] = floor(gp->minus[i][j] *1000+0.5)/1000;
				
				//mapped_sum = logsum(mapped_sum, gp->minus[i][j]);
			}
			
		}
	}*/
		
	
	gp->num_reads = 1;
	
	
	
		
	//fprintf(stderr,"Printing... %d	%d\n",numseq,gp->num_reads);
	print_bedgraphs(db,param,gp,size);
		
	for(i = 0;i < size;i++){
		if(gp->plus[i]){
			free(gp->plus[i]);
		}
		if(gp->minus[i]){
			free(gp->minus[i]);
		}
	}
	
	
	free(gp->plus);
	free(gp->minus);
	free(gp);
	
	if(param->local_weighting){
		free_rp(rp);
	}
	for(i = 0; i < param->num_query;i++){
		
		free(ri[i]->strand); 
		free(ri[i]->hits);
		free(ri[i]->identity);
		free(ri[i]);
	}
	free_2nd_hmm(hmm);
	free(ri);
	
	//if(param->infile[1]){
	fclose(file);
	//}
	return 1;
	
}

struct hmm2* run_posterior2(struct hmm2* localhmm,struct read_info** ri,struct db* db,struct genome_probabilities* gp,struct parameters* param,int size)
{
	struct thread_data2* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data2)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	//hmm = init_hmm(lena, lenb, 0.002f, 0.001f, 0.02f, 1);
	
	
	interval =  (int)((double)size /(double)param->num_threads);
	
	fprintf(stderr,"Starting with %d sequences\n",size);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].db = db;
		thread_data_array[t].gp  = gp;
		thread_data_array[t].hmm = init_2nd_hmm(500, 500);
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
		
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].db = db;
	thread_data_array[param->num_threads - 1].gp  = gp;
	thread_data_array[param->num_threads - 1].hmm = init_2nd_hmm(500, 500);
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = size;
	thread_data_array[param->num_threads - 1].param = param;
	
	init_thread_hmm2s(thread_data_array,localhmm,param->num_threads);
	pthread_mutex_init(&mutex1, NULL);
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, &do_get_posteriors_thread2, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_mutex_destroy(&mutex1);
	//fprintf(stderr,"Got here....\n");
	//fprintf(stderr,"local:%d	%f	%f	%f\n",t, localhmm->genome_model_t[0][0],localhmm->genome_model_t[0][1],localhmm->genome_model_t[0][2]);
	//for(t = 0;t < param->num_threads;t++) {
	//	fprintf(stderr,"Tread:%d	%f	%f	%f\n",t,thread_data_array[t].hmm->genome_model_t[0][0],thread_data_array[t].hmm->genome_model_t[0][1],thread_data_array[t].hmm->genome_model_t[0][2]);
	//}
	
	
	//entangle_hmm2s(thread_data_array,localhmm,param->num_threads);
	//fprintf(stderr,"local:%d	%f	%f	%f\n",t, localhmm->genome_model_t[0][0],localhmm->genome_model_t[0][1],localhmm->genome_model_t[0][2]);
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	for(t = 0;t < param->num_threads;t++) {
		free_2nd_hmm(thread_data_array[t].hmm);
		//print_hmm_param(thread_data_array[t].hmm);
	}
	
	free(thread_data_array);
	return localhmm;
}


void* do_get_posteriors_thread2(void *threadarg)
{
	struct thread_data2 *data;
	data = (struct thread_data2 *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm2* hmm = data->hmm;
	struct db* db = data->db;
	struct genome_probabilities* gp = data->gp;
	
	const int start = data->start; 
	const int end = data->end;
	//float random_model_prior =  prob2scaledprob(data->param->random_model_prior);
	//float model_prior = prob2scaledprob(1.0f - data->param->random_model_prior);
	
	//posterior[i] - hmm->score + fracfloat unaligned_probability_cutoff = data->param->unaligned_probability_cutoff;
	
	int count_mismatches_posterior = data->param->count_mismatches_posterior;
	int i,j,hit;	
	int strand = 0;
	unsigned char* seq = 0;
	
	float million = prob2scaledprob(1000000);
	
	//float pseudocount = 0.0f;
	seq = malloc(sizeof(unsigned char) * 500);

	float sum = 0.0f;
	float max= 0.0f;
	float unaligned = 0.0f;
	//float genome_sum = 0;
	int c,g;
	int alignment_counter = 0;
	float* scores = malloc(sizeof(float)* 100); 
	float* random_scores = malloc(sizeof(float)* 100); 
	float* genome_scores = malloc(sizeof(float)* 100); 
	float* genome_posterior = 0;
	unsigned int a,b,f;
	
	struct posteriors_collector** pc = 0;
	pc = malloc(sizeof(struct posteriors_collector*) * MAX_LOCAL_POSTERIORS);
	for(i = 0; i < MAX_LOCAL_POSTERIORS;i++){
		pc[i] = malloc(sizeof(struct posteriors_collector  ));
	}
	//fprintf(stderr,"%d-%d\n",start,end);
	for(i = start; i < end;i++){
		hit = 0;
		j = 0;
		while(ri[i]->identity[hit] != -2.0f){
			if(ri[i]->strand[hit] !=0){
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
			}else{
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
			}
			//seq = transform_genome_sequence(seq , ri[i]->len+20);
			//hmm =  score_genome(hmm,seq,ri[i]->len+20 );
			
			hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ,-1); 
			hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),0);
			
			
			scores[hit] = hmm->score;// + ri[i]->priors[hit];
			
			genome_scores[hit] = ri[i]->priors[hit];
			
			random_scores[hit] =  hmm->random_score ;
			
			//scores[j] = hmm->score + ri[i]->priors[hit];
			//random_scores[j] =  hmm->random_score;
			j++;
			hit++;
		}
		
		sum = genome_scores[0];
		for(c = 1; c < hit;c++){
			sum = logsum(sum,genome_scores[c]);
		}
		for(c = 0; c < hit;c++){
			genome_scores[c] = genome_scores[c] -sum + prob2scaledprob(1.0 - data->param->unaligned_prior);
		}
		
		hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),1);
		sum = hmm->random_score + prob2scaledprob( data->param->unaligned_prior);
		for(c = 0; c < hit;c++){
			sum += random_scores[c] + prob2scaledprob(1.0 - scaledprob2prob(genome_scores[c]));
		}
		//fprintf(stderr,"%f	%f\n", sum, prob2scaledprob( (1.0 / 0.9)  - 1.0 ));
		
		unaligned = sum ;
		
		
		for(c = 0;c < hit;c++){
			sum = prob2scaledprob(1.0f);
			for(j = 0;j < hit;j++){
				if(c == j){
					//read is aligned at this seed position 
					sum += scores[j] + genome_scores[j];
				}else{
					//read not aligned at this seed
					sum += random_scores[j] + prob2scaledprob(1.0 - scaledprob2prob(genome_scores[j])) ;
					//sum += random_scores[j] + prob2scaledprob(0.1 * 1.0/ (float) hit);
				}
			}
			
			//read is aligned to genome  
			//sum += prob2scaledprob(0.9);
			scores[c] = sum;
		}
		

		sum = unaligned;
		for(c = 0; c < hit;c++){
			sum = logsum(sum,scores[c]);
			
		}
		unaligned = unaligned - sum;
		
		max = prob2scaledprob(0.0);
		
		for(c = 0 ; c < hit;c++){
			
			scores[c] = scores[c] - sum;
			if(scores[c] > max){
				max = scores[c];
			}
		}
		//sum = scores[0];
		//for(c = 1; c < j;c++){
		//	sum = logsum(sum,scores[c]);
		//}		
		
		hit = 0;
		while(ri[i]->identity[hit] != -2.0f){
			//fprintf(stderr,"%d	%f	%f	%f	%f\n",hit, model_prior,random_model_prior,scores[hit], random_scores[hit] );
			//fprintf(stderr,"%f\n",scaledprob2prob( (scores[hit] + model_prior) - ( logsum( scores[hit] + model_prior,  random_scores[hit] + random_model_prior))));
			//if(scaledprob2prob(  (scores[hit] + model_prior) - ( logsum( scores[hit] + model_prior,  random_scores[hit] + random_model_prior))) > unaligned_probability_cutoff){
			if(max > unaligned && scaledprob2prob(scores[hit]) > (1.0 / (double)LIST_STORE_SIZE)){
				if(ri[i]->strand[hit] != 0){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
					pc[alignment_counter]->strand = 16;
				//	local_strands[alignment_counter] = 16;
				}else{
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
					pc[alignment_counter]->strand = 0;
				//	local_strands[alignment_counter] = 0;
				}
				
				pc[alignment_counter]->pos = ri[i]->hits[hit]-10;
				pc[alignment_counter]->len = ri[i]->len+20;
				
				//seq = transform_genome_sequence(seq , ri[i]->len+20);
				//hmm =  score_genome(hmm,seq,ri[i]->len+20 );
				hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,-1 ); 
				
				hmm = glocal_backward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20); 
				pc[alignment_counter]->local_posteriors  = read_genome_posteriors(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,scores[hit] ,count_mismatches_posterior );
				
				if(ri[i]->strand[hit] != 0){
					pc[alignment_counter]->local_posteriors = reverse_genome_posteriors(pc[alignment_counter]->local_posteriors , ri[i]->len+20);
				}
				
				alignment_counter++;
				if(alignment_counter == MAX_LOCAL_POSTERIORS){
					//fprintf(stderr,"Thread number %ld\n", (long int)pthread_self());
					pthread_mutex_lock( &mutex1 );
					
					for(c= 0; c < alignment_counter;c++){
						gp->num_reads++;
						a = (pc[c]->pos ) >> 10;
						b = (pc[c]->pos +  pc[c]->len) >> 10;
						genome_posterior = pc[c]->local_posteriors;
						
						
						/*if(c == 0){
							fprintf(stderr,"%c	%d	%d	%d	",c,pc[c]->pos,pc[c]->len,pc[c]->strand);
							for(f = 0;f < pc[c]->len;f++){
								fprintf(stderr,"%f ",pc[c]->local_posteriors[f]);
								
							}
							fprintf(stderr,"\n");
						}*/
						
						if(pc[c]->strand == 0){
							if(a == b){
								if(!gp->plus[a]){
									gp->plus[a] = malloc(sizeof(float) * 1024);
									for(g = 0; g < 1024;g++){
										gp->plus[a][g] = 0.0f;
									}
								}
								f = a;
								//fprintf(stderr,"Region:%d\n",c);
								a = (pc[c]->pos)  & 0x3FF;
								b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
								//fprintf(stderr,"%d	%d\n",a,b);
								for(g = a;g <= b;g++){
									//	fprintf(stderr,"reding from:%d\n",j - a);
									gp->plus[f][g] += floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0;
								}
							}else{
								if(!gp->plus[a]){
									gp->plus[a] = malloc(sizeof(float) * 1024);
									for(g = 0; g < 1024;g++){
										gp->plus[a][g] = 0.0f;
									}
								}
								if(!gp->plus[b]){
									gp->plus[b] = malloc(sizeof(float) * 1024);
									for(g = 0; g < 1024;g++){
										gp->plus[b][g] = 0.0f;
									}
								}
								f = a;
								a = (pc[c]->pos)  & 0x3FF;
								
								//fprintf(stderr,"%d	%d\n",a,b);
								for(g = a;g < 1024;g++){
									//fprintf(stderr,"reding from:%d	(%p)\n",j - a,gp->plus[c]);
									gp->plus[f][g] +=  floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0;
									
								}
								//fprintf(stderr,"break\n");
								f = b;
								b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
								for(g = 0;g <= b;g++){
									//fprintf(stderr,"reding from:%d\n",j + 1024 - a);
									gp->plus[f][g] += floor(scaledprob2prob(genome_posterior[g + 1024 - a]+million) + 0.5) / 1000000.0;
									//logsum(gp->plus[f][g],genome_posterior[g + 1024 - a]);
								}
							}
						}else{
							if(a == b){
								if(!gp->minus[a]){
									gp->minus[a] = malloc(sizeof(float) * 1024);
									for(g = 0; g < 1024;g++){
										gp->minus[a][g] = 0.0f;
									}
								}
								f = a;
								a = (pc[c]->pos)  & 0x3FF;
								b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
								for(g = a;g <= b;g++){
									gp->minus[f][g] += floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0; 
									//logsum(gp->minus[f][g],genome_posterior[g - a]);
								}
							}else{
								if(!gp->minus[a]){
									gp->minus[a] = malloc(sizeof(float) * 1024);
									for(g = 0; g < 1024;g++){
										gp->minus[a][g] = 0.0f;
									}
								}
								if(!gp->minus[b]){
									gp->minus[b] = malloc(sizeof(float) * 1024);
									for(g = 0; g < 1024;g++){
										gp->minus[b][g] = 0.0f;
									}
								}
								f = a;
								a = (pc[c]->pos)  & 0x3FF;
								
								
								for(g = a;g < 1024;g++){
									gp->minus[f][g] += floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0; 
									//logsum(gp->minus[f][g],genome_posterior[g - a]);
								}
								
								f= b;
								b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
								for(g = 0;g <= b;g++){
									gp->minus[f][g] += floor(scaledprob2prob(genome_posterior[g + 1024 - a]+million) + 0.5) / 1000000.0; 
									//logsum(gp->minus[f][g],genome_posterior[g + 1024 -a]);
								}
							}
						}
					}
					counter++;
					//fprintf(stderr,"Counter value: %d\n",counter);
					
					pthread_mutex_unlock( &mutex1 );
					for(c= 0; c < alignment_counter;c++){
						//fprintf(stderr,"%d	%p	%ud	%ud\n",c,pc[c]->local_posteriors,pc[c]->pos,pc[c]->strand);
						free(pc[c]->local_posteriors);
					}
					alignment_counter = 0;
				}

			}
			//hmm =  read_genome_posteriors(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20); 
					
			hit++;
		}
		//exit(0);
	}
	
	if(alignment_counter){
		//fprintf(stderr,"Thread number %ld\n", (long int)pthread_self());
		pthread_mutex_lock( &mutex1 );
		for(c= 0; c < alignment_counter;c++){
			gp->num_reads++;
			a = (pc[c]->pos ) >> 10;
			b = (pc[c]->pos +  pc[c]->len) >> 10;
			genome_posterior = pc[c]->local_posteriors;
			
			
			/*if(c == 0){
			 fprintf(stderr,"%c	%d	%d	%d	",c,pc[c]->pos,pc[c]->len,pc[c]->strand);
			 for(f = 0;f < pc[c]->len;f++){
			 fprintf(stderr,"%f ",pc[c]->local_posteriors[f]);
			 
			 }
			 fprintf(stderr,"\n");
			 }*/
			
			if(pc[c]->strand == 0){
				if(a == b){
					if(!gp->plus[a]){
						gp->plus[a] = malloc(sizeof(float) * 1024);
						for(g = 0; g < 1024;g++){
							gp->plus[a][g] = 0.0f;
						}
					}
					f = a;
					//fprintf(stderr,"Region:%d\n",c);
					a = (pc[c]->pos)  & 0x3FF;
					b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
					//fprintf(stderr,"%d	%d\n",a,b);
					for(g = a;g <= b;g++){
						//	fprintf(stderr,"reding from:%d\n",j - a);
						gp->plus[f][g] += floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0;
					}
				}else{
					if(!gp->plus[a]){
						gp->plus[a] = malloc(sizeof(float) * 1024);
						for(g = 0; g < 1024;g++){
							gp->plus[a][g] = 0.0f;
						}
					}
					if(!gp->plus[b]){
						gp->plus[b] = malloc(sizeof(float) * 1024);
						for(g = 0; g < 1024;g++){
							gp->plus[b][g] = 0.0f;
						}
					}
					f = a;
					a = (pc[c]->pos)  & 0x3FF;
					
					//fprintf(stderr,"%d	%d\n",a,b);
					for(g = a;g < 1024;g++){
						//fprintf(stderr,"reding from:%d	(%p)\n",j - a,gp->plus[c]);
						gp->plus[f][g] +=  floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0;
						
					}
					//fprintf(stderr,"break\n");
					f = b;
					b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
					for(g = 0;g <= b;g++){
						//fprintf(stderr,"reding from:%d\n",j + 1024 - a);
						gp->plus[f][g] += floor(scaledprob2prob(genome_posterior[g + 1024 - a]+million) + 0.5) / 1000000.0;
						//logsum(gp->plus[f][g],genome_posterior[g + 1024 - a]);
					}
				}
			}else{
				if(a == b){
					if(!gp->minus[a]){
						gp->minus[a] = malloc(sizeof(float) * 1024);
						for(g = 0; g < 1024;g++){
							gp->minus[a][g] = 0.0f;
						}
					}
					f = a;
					a = (pc[c]->pos)  & 0x3FF;
					b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
					for(g = a;g <= b;g++){
						gp->minus[f][g] += floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0; 
						//logsum(gp->minus[f][g],genome_posterior[g - a]);
					}
				}else{
					if(!gp->minus[a]){
						gp->minus[a] = malloc(sizeof(float) * 1024);
						for(g = 0; g < 1024;g++){
							gp->minus[a][g] = 0.0f;
						}
					}
					if(!gp->minus[b]){
						gp->minus[b] = malloc(sizeof(float) * 1024);
						for(g = 0; g < 1024;g++){
							gp->minus[b][g] = 0.0f;
						}
					}
					f = a;
					a = (pc[c]->pos)  & 0x3FF;
					
					
					for(g = a;g < 1024;g++){
						gp->minus[f][g] += floor(scaledprob2prob(genome_posterior[g - a]+million) + 0.5) / 1000000.0; 
						//logsum(gp->minus[f][g],genome_posterior[g - a]);
					}
					
					f= b;
					b = (pc[c]->pos +  pc[c]->len) & 0x3FF;
					for(g = 0;g <= b;g++){
						gp->minus[f][g] += floor(scaledprob2prob(genome_posterior[g + 1024 - a]+million) + 0.5) / 1000000.0; 
						//logsum(gp->minus[f][g],genome_posterior[g + 1024 -a]);
					}
				}
			}
		}
		
		counter++;
		//fprintf(stderr,"Counter value: %d\n",counter);
		pthread_mutex_unlock( &mutex1 );
		for(c= 0; c < alignment_counter;c++){
			free(pc[c]->local_posteriors);
		}
	}
	

	for(i = 0; i < MAX_LOCAL_POSTERIORS;i++){
		free(pc[i]);// = malloc(sizeof(struct posteriors_collector  ));
	}
	free(pc); //= malloc(sizeof(struct posteriors_collector*) * 1000);
	free(genome_scores);
	free(random_scores);
	free(scores);
	free(seq);
	pthread_exit((void *) 0);
}


int phmm2_controller(struct db* db,struct parameters* param)
{
	struct read_info** ri = 0;
	FILE* file = 0;
	int i,j,numseq,iter = 0;
	int total_numseq = 0;
	init_logsum();
	struct hmm2* hmm = init_2nd_hmm(500, 500);
	
	//int tmp;
	struct rescue_probabilities* rp = 0;
	if(param->local_weighting){
		rp =  init_rp(db->seq_info->total_len ,param->local_weighting);
	}
	//fprintf(stderr,"malloced local...\n");
	ri = malloc(sizeof(struct read_info*) * param->num_query);
	
	for(i = 0; i < param->num_query;i++){
		ri[i] = malloc(sizeof(struct read_info));
		ri[i]->seq = 0;
		ri[i]->name = 0;
		ri[i]->qual = 0;
		ri[i]->len = 0;
		ri[i]->priors = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->strand = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->hits = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->identity = malloc(sizeof(float)* (LIST_STORE_SIZE+1));
	}
	//fprintf(stderr,"malloced local ri\n");

	/*if(!isatty(0)){
	 file = stdin;
	 }else{
	 */
	if (!(file = fopen(param->infile[1] , "r" ))){
		fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[1]);
		exit(-1);
	}
	//}
	
	if(param->hmm_file &&  access( param->hmm_file, R_OK) != -1 ) {
		fprintf(stderr,"Reading Model\n");
		read_hmm2_from_file(hmm,param);
		read_rescue_probabilities(rp, db->seq_info->total_len ,param->local_weighting, param);
	}else {
		fprintf(stderr,"Estimating Random model...");
		for(iter = 0; iter < 1;iter++){ 
			while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
				for(i = 0;i < numseq;i++){
					if(ri[i]->strand[0] != 0){
						ri[i]->seq = reverse_complement2(ri[i]->seq,ri[i]->len);
					}
					for(j = 0;j < ri[i]->len;j++){
						if(ri[i]->seq[j] == 4){
							ri[i]->seq[j] = NCODE;
						}else if(j >= ri[i]->len-BORDER_REGION){
							ri[i]->seq[j] += 8;
						}else if(j  >= BORDER_REGION){
							ri[i]->seq[j] += 4;
						}
					}
				}			
				total_numseq += numseq;
				hmm = run_pHMM2_random(hmm,ri,db,param,numseq);
			}
			hmm = re_estimate_2ndorder_random(hmm);
			rewind(file);
		}
		fprintf(stderr,"%d	%d\n",hmm->train_count,total_numseq);
		fprintf(stderr,"	done.\n");
		
		param->unaligned_prior = 1.0 - ((double)hmm->train_count / (double)total_numseq);
		
		//exit(0);
		hmm->train_count = 0;
		/*fprintf(stderr,"Jump staring 2nd order pHMM...");
		tmp =  param->num_query;
		param->train_pair_hmm = 1;
		param->num_query = 1000000;
		for(iter = 0; iter < 10;iter++){ 
			total_numseq = 0;
			param->seed_errors = 0;
			param->total_len = 0;
			while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
				for(i = 0;i < numseq;i++){
					if(ri[i]->strand[0] != 0){
						ri[i]->seq = reverse_complement2(ri[i]->seq,ri[i]->len);
					}
					for(j = 0;j < ri[i]->len;j++){
						if(ri[i]->seq[j] == 4){
							ri[i]->seq[j] = NCODE;
						}else if(j >= ri[i]->len-BORDER_REGION){
							ri[i]->seq[j] += 8;
						}else if(j  >= BORDER_REGION){
							ri[i]->seq[j] += 4;
						}
					}
					
				}
				
				fprintf(stderr,"Reading	%d\n",numseq);
				//ri = add_k(ri, db,numseq);
				//if(iter > -1){
				//	ri = set_local_prob(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq);
				//}
				hmm = run_pHMM2(hmm,ri, db,param,numseq);
				//if(iter > -1){
				//	rp = update_rp(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq,0,0);
				//}
				total_numseq += numseq;
				if(total_numseq >= 1000000){
					break;
				}
				
			}
			rewind(file);
			fprintf(stderr,"Iter:%d\n",iter);
			for(i = 0; i < 65536;i++){
				hmm->tprob[i] = logsum(hmm->tprob[i], prob2scaledprob(1.0)); 
			}			
			
			//for(i = 0; i < 25;i++){
			//	hmm->random_genome_model_t[i] = logsum(hmm->random_genome_model_t[i], prob2scaledprob(1.0));
			//	hmm->random_read_model_t[i] = logsum(hmm->random_read_model_t[i], prob2scaledprob(1.0));
			//}
			
			hmm = re_estimate_2ndorder(hmm);
		}
		
		param->num_query = tmp;
		param->train_pair_hmm = 1;
		fprintf(stderr,"	done.\n");*/
		param->train_pair_hmm = 1;
		for(iter = 0; iter < param->iterations;iter++){ 
			total_numseq = 0;
			param->seed_errors = 0;
			param->total_len = 0;
			hmm->train_count =0;
			while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
				for(i = 0;i < numseq;i++){
					if(ri[i]->strand[0] != 0){
						ri[i]->seq = reverse_complement2(ri[i]->seq,ri[i]->len);
					}
					for(j = 0;j < ri[i]->len;j++){
						if(ri[i]->seq[j] == 4){
							ri[i]->seq[j] = NCODE;
						}else if(j >= ri[i]->len-BORDER_REGION){
							ri[i]->seq[j] += 8;
						}else if(j  >= BORDER_REGION){
							ri[i]->seq[j] += 4;
						}
					}
					
				}
								
				fprintf(stderr,"Reading	%d\n",numseq);
				ri = add_k(ri, db,numseq);
				if(iter > param->iterations / 2){
					ri = set_local_prob(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq);
				}
				hmm = run_pHMM2(hmm,ri, db,param,numseq);
				if(iter > param->iterations / 2){
					rp = update_rp(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq,0,0);
				}
				total_numseq += numseq;
				if(hmm->train_count > param->training_sequences){
					break;
				}
				
			}
			rewind(file);
			hmm = re_estimate_2ndorder(hmm);
			hmm = re_estimate_2ndorder_random(hmm);
			if(iter > param->iterations / 2){
				rp = update_rp(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq,1, 1.0f *total_numseq );
			}
			//fprintf(stderr,"Iter:%d	%d	%d	%f (should be r)\n",iter,param->total_len,total_numseq  , (double)param->total_len  / (double) total_numseq);
			/*if(!param->mean_error){
				param->mean_error = (double)param->seed_errors/  (double)param->total_len;
				fprintf(stderr,"Iter:%d	%d	%d	%f\n",iter, param->seed_errors,param->total_len,(double)param->seed_errors/  (double)param->total_len  );
			}else{
				//param->random_model_prior =  (double)param->total_len  / (double) total_numseq;
				fprintf(stderr,"Iter:%d	%d	%d	%f (should be r)\n",iter,param->total_len,total_numseq  , (double)param->total_len  / (double) total_numseq);
			}*/
			param->random_model_prior =  (double)param->total_len  / (double) total_numseq;
			//fprintf(stderr,"Iter:%d	%d	%d	%f (should be r)\n",iter,param->total_len,total_numseq  , (double)param->total_len  / (double) total_numseq);
			fprintf(stderr,"Iter:%d\n",iter);
		}
		
		
				
		
		if(param->hmm_file){
			write_hmm2_to_file(hmm,param);
			write_rescue_probabilities(rp, db->seq_info->total_len ,param->local_weighting, param);
		}
		rewind(file);
	}
	
	
	//fprintf(stderr,"Random prior: %f\n",param->random_model_prior);
	/*rewind(file);
	if(param->fdr){
		while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
			for(i = 0;i < numseq;i++){
				if(ri[i]->strand[0] != 0){
					ri[i]->seq = reverse_complement2(ri[i]->seq,ri[i]->len);
				}
				for(j = 0;j < ri[i]->len;j++){
					if(ri[i]->seq[j] == 4){
						ri[i]->seq[j] = NCODE;
					}else if(j >= ri[i]->len-BORDER_REGION){
						ri[i]->seq[j] += 8;
					}else if(j  >= BORDER_REGION){
						ri[i]->seq[j] += 4;
					}
				}
			}
			
			//ri = add_k(ri, db,numseq);
			if(param->local_weighting){
				ri = set_local_prob(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq);
			}
			
			hmm =  get_fdr_cutoff(hmm,ri,db,param,numseq);
			//ajhfke hark,ahcja ghckhajf shri = run_score_alignments_hmm2(hmm,ri,db,param,numseq);
		}
		rewind(file);
		float sum = 0.0f;
		for(i = 0;i < 100001;i++){
			sum+= hmm->p_value_ranks[i];
		}
		float c_sum = 0.0f;
		
		for(i = 0;i < 100001;i++){
			//fprintf(stderr,"%d		%f	%d	%f	%f\n",i, (double)i / 100000.0, hmm->p_value_ranks[i], sum / (sum - c_sum + 2), (1.0 - (double)i / 100000.0)  * sum / (sum - c_sum + 2));
			if((1.0 - (double)i / 100000.0)  * sum / (sum - c_sum + 2) <= param->fdr){
				param->fdr = (double)i / 100000.0;
				break;
			}
			c_sum += hmm->p_value_ranks[i];
		}
		fprintf(stderr,"FDR cutoff = %f\n",param->fdr);
		free(hmm->p_value_ranks);
	}
	rewind(file);*/
	
	print_samheader(db,param);
	
		
	while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
		for(i = 0;i < numseq;i++){
			if(ri[i]->strand[0] != 0){
				ri[i]->seq = reverse_complement2(ri[i]->seq,ri[i]->len);
			}
			for(j = 0;j < ri[i]->len;j++){
				if(ri[i]->seq[j] == 4){
					ri[i]->seq[j] = NCODE;
				}else if(j >= ri[i]->len-BORDER_REGION){
					ri[i]->seq[j] += 8;
				}else if(j  >= BORDER_REGION){
					ri[i]->seq[j] += 4;
				}
			}
		}
		
		//ri = add_k(ri, db,numseq);
		if(param->local_weighting){
			ri = set_local_prob(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq);
		}
		
		
		ri = run_score_alignments_hmm2(hmm,ri,db,param,numseq);
	}
	/*int r1,g1,r2,g2;
	for(i = 0; i < 65536;i++){
		r1 = (i >> 12) & 15;
		g1= (i >> 8) & 15;
		r2 = (i >> 4) & 15;
		g2 = i & 15; 
		
		if((r1 & 3) == (r2 & 3)){
			if((g1 & 3) == (g2 & 3)){
				fprintf(stderr,"%d	%d	%d	%d	%f	%f	\n",r1,g1,r2,g2,hmm->prob[i],scaledprob2prob(hmm->prob[i]));
			}
		}
		
		//if(r1  == EGAP && r2 != EGAP){
			//if(scaledprob2prob(hmm->prob[i]) > 0.1){
		//	fprintf(stderr,"%d	%d	%d	%d	%f	%f	\n",r1,g1,r2,g2,hmm->prob[i],scaledprob2prob(hmm->prob[i]));
			//}
		//}
		
	}*/		
	if(param->local_weighting){
		free_rp(rp);
	}
	for(i = 0; i < param->num_query;i++){
		
		free(ri[i]->strand); 
		free(ri[i]->hits);
		free(ri[i]->identity);
		free(ri[i]);
	}
	free_2nd_hmm(hmm);
	free(ri);
	
	//if(param->infile[1]){
	fclose(file);
	//}
	return 1;
}


struct hmm2* get_fdr_cutoff(struct hmm2* hmm,struct read_info** ri,struct db* db,struct parameters* param,int size)
{
	struct thread_data2* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data2)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	init_logsum();
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	//hmm = init_hmm(lena, lenb, 0.002f, 0.001f, 0.02f, 1);
	hmm->p_value_ranks = malloc(sizeof(int) * 100001);
	for(t = 0; t < 100001;t++){
		hmm->p_value_ranks[t] = 0;
	}
	
	interval =  (int)((double)size /(double)param->num_threads);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].db = db;
		thread_data_array[t].hmm =  init_2nd_hmm(500, 500);
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
		thread_data_array[t].p_value_ranks = hmm->p_value_ranks;
		
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].db = db;
	thread_data_array[param->num_threads - 1].hmm = init_2nd_hmm(500, 500);
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = size;
	thread_data_array[param->num_threads - 1].param = param;
	thread_data_array[param->num_threads - 1].p_value_ranks = hmm->p_value_ranks;
	
	init_thread_hmm2s(thread_data_array,hmm,param->num_threads);
	
	pthread_mutex_init(&mutex1, NULL);
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, do_get_fdr_thread_hmm2, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_mutex_destroy(&mutex1);
	
	for(t = 0;t < param->num_threads;t++) {
		free_2nd_hmm(thread_data_array[t].hmm);
	}
	free(thread_data_array);
	return hmm;
}

void* do_get_fdr_thread_hmm2(void *threadarg)
{
	struct thread_data2 *data;
	data = (struct thread_data2 *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm2* hmm = data->hmm;
	struct db* db = data->db;
	const int start = data->start; 
	const int end = data->end;
	const struct parameters* param = data->param;
	int i,j,c;
	int hit = 0;
	int strand;
	unsigned char* seq = 0;
	int num_scores = 0;
	//int k = 0;
	unsigned long int*  top_scores = 0;
	unsigned long int top_cut = 0ul;
	
	float random_model_prior =  param->random_model_prior;
	float model_prior = 1.0f - random_model_prior;
	float* mapping_scores = 0;
	
	//unsigned char noqual[] = "*";
	//unsigned char* qual = 0;
	float sum;//, genome_sum, prior_sum;
	
	
	mapping_scores = malloc(sizeof(float)* LIST_STORE_SIZE);
	top_scores = malloc(sizeof(unsigned long int)* LIST_STORE_SIZE);
	float* random_scores = malloc(sizeof(float)* 100); 
	seq = malloc(sizeof(unsigned char) * 500);
	float* local_p_values = malloc(sizeof(float) * 100000);
	int* p_value_ranks = data->p_value_ranks;
	
	int p_value_count = 0;
	
	for(i = start;i < end;i++){
		num_scores = 0;
		//fprintf(stderr,"%d\n",i);
		
		//if(ri[i]->identity[0] == -2.0f){
		//	unaligned_to_sam(ri[i]);			
		//}
		hit = 0;
		while(ri[i]->identity[hit] != -2.0f){ // option != -2 
			
			
			if(ri[i]->strand[hit] != 0){
				//reverse_complement2(ri[i]->seq,ri[i]->len);
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);

				
			}else{
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);

			}
			seq = transform_genome_sequence(seq , ri[i]->len+20);
			
			hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,-1 ); 
			hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),0);
			mapping_scores[num_scores] = hmm->score + ri[i]->priors[hit];
			
			random_scores[num_scores] =  hmm->random_score;
			
			num_scores++;
			hit++;
		}
		
		if(num_scores){
			sum = mapping_scores[0];
			for(j = 1; j < num_scores;j++){
				sum = logsum(sum,mapping_scores[j]);
			}
			
			for(j = 0; j < LIST_STORE_SIZE;j++){
				top_scores[j] = 0ul;
			}
			
			
			for(j = 0; j < num_scores;j++){
				BinaryInsertionSortOne(top_scores,LIST_STORE_SIZE,(unsigned long int) (scaledprob2prob( mapping_scores[j]  -sum ) * 1000000.0f));
			}
			
			j = 1;
			top_cut = top_scores[param->upper_limit-j];
			while(top_cut == 0.0f){
				j++;
				top_cut = top_scores[param->upper_limit-j];
			}
			
			
			for(j = 0; j < num_scores;j++){
				local_p_values[p_value_count] = scaledprob2prob(  (mapping_scores[j] + prob2scaledprob(model_prior)) - ( logsum( mapping_scores[j] + prob2scaledprob(model_prior),  random_scores[j] + prob2scaledprob(random_model_prior)))) * 100000.0;
				p_value_count++;
				if(p_value_count == 100000){
					
					//printf("Thread number %ld\n", (long int)pthread_self());
					pthread_mutex_lock( &mutex1 );
					counter++;
					//printf("Counter value: %d\n",counter);
					for(c =0;c < p_value_count;c++){
						p_value_ranks[(int)local_p_values[c]]++;
					}
					pthread_mutex_unlock( &mutex1 );
					p_value_count = 0;
				}
			}
			
		}
	}
	if(p_value_count){
		//printf("Thread number %ld\n", (long int)pthread_self());
		pthread_mutex_lock( &mutex1 );
		counter++;
		//printf("Counter value: %d\n",counter);
		for(c =0;c < p_value_count;c++){
			p_value_ranks[(int)local_p_values[c]]++;
		}
		pthread_mutex_unlock( &mutex1 );
		p_value_count = 0;
	}
	
	free(local_p_values);
	free(random_scores);
	free(mapping_scores);
	free(top_scores);	
	free(seq);
	pthread_exit((void *) 0);
}

struct read_info** run_score_alignments_hmm2(struct hmm2* hmm,struct read_info** ri,struct db* db,struct parameters* param,int size)
{
	struct thread_data2* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data2)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	init_logsum();
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	//hmm = init_hmm(lena, lenb, 0.002f, 0.001f, 0.02f, 1);
	
	
	interval =  (int)((double)size /(double)param->num_threads);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].db = db;
		thread_data_array[t].hmm =  init_2nd_hmm(500, 500);
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].db = db;
	thread_data_array[param->num_threads - 1].hmm = init_2nd_hmm(500, 500);
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = size;
	thread_data_array[param->num_threads - 1].param = param;
	
	init_thread_hmm2s(thread_data_array,hmm,param->num_threads);
	
	
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, do_score_alignments_thread_hmm2, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	
	for(t = 0;t < param->num_threads;t++) {
		free_2nd_hmm(thread_data_array[t].hmm);
	}
	free(thread_data_array);
	return ri;
}

void* do_score_alignments_thread_hmm2(void *threadarg)
{
	struct thread_data2 *data;
	data = (struct thread_data2 *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm2* hmm = data->hmm;
	struct db* db = data->db;
	const int start = data->start; 
	const int end = data->end;
	//const struct parameters* param = data->param;
	int i,j,c;
	int hit = 0;
	int strand;
	unsigned char* seq = 0;
	unsigned char* aln = 0;
	int num_scores = 0;
	//int k = 0;
	unsigned long int*  top_scores = 0;
	//unsigned long int top_cut = 0ul;
	
	//float random_model_prior =  param->random_model_prior;
	//float model_prior = 1.0f - random_model_prior;
	
	//float unaligned_probability_cutoff = param->fdr;// fparam->unaligned_probability_cutoff;
	float* scores = 0;
	
	//unsigned char noqual[] = "*";
	//unsigned char* qual = 0;
	float sum;//genome_sum;
	float unaligned;
	float max;
	int flag;
	scores = malloc(sizeof(float)* LIST_STORE_SIZE);
	top_scores = malloc(sizeof(unsigned long int)* (LIST_STORE_SIZE+1));
	float* random_scores = malloc(sizeof(float)* 100); 
	float* genome_scores = malloc(sizeof(float)* 100); 
	seq = malloc(sizeof(unsigned char) * 500);
	
	
	for(i = start;i < end;i++){
		hit = 0;
		num_scores = 0;
		//fprintf(stderr,"%d\n",i);
			
		if(ri[i]->identity[0] == -2.0f){
			unaligned_to_sam(ri[i]);			
		}
		hit = 0;
		while(ri[i]->identity[hit] != -2.0f){ // option != -2 
			
			
			if(ri[i]->strand[hit] != 0){
				//reverse_complement2(ri[i]->seq,ri[i]->len);
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
				//reverse_complement2(ri[i]->seq,ri[i]->len);
				//reverse_complement2(ri[i]->seq,ri[i]->len);
			}else{
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				//hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ); 
			}
			seq = transform_genome_sequence(seq , ri[i]->len+20);

			hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,-1 ); 
			hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),0);
			//hmm =  score_genome(hmm,seq,ri[i]->len+20 );
			//fprintf(stderr,"%d	%f	%f\n",hit, hmm->score,hmm->genome_score  );
			//scores[num_scores] = hmm->score + ri[i]->priors[hit];
			
			scores[num_scores] = hmm->score;// +  ri[i]->priors[hit];
			genome_scores[hit] = ri[i]->priors[hit];
			random_scores[num_scores] =  hmm->random_score;
			
			num_scores++;
			hit++;
		}
		
		if(num_scores){
			sum = genome_scores[0];
			for(c = 1; c < hit;c++){
				sum = logsum(sum,genome_scores[c]);
			}
			for(c = 0; c < hit;c++){
				genome_scores[c] = genome_scores[c] -sum+ prob2scaledprob(1.0 - data->param->unaligned_prior);//   0.9);
			}
			
			hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),1);
			sum = hmm->random_score + prob2scaledprob( data->param->unaligned_prior);// 0.1);
			for(c = 0; c < hit;c++){
				sum += random_scores[c] + prob2scaledprob(1.0 - scaledprob2prob(genome_scores[c]));
			}
			//fprintf(stderr,"%f	%f\n", sum, prob2scaledprob( (1.0 / 0.9)  - 1.0 ));
			
			unaligned = sum ;

			
			for(c = 0;c < hit;c++){
				sum = prob2scaledprob(1.0f);
				//if(i < 5){
				//	fprintf(stderr,"%d	%f	%f	%f	%f	%f	%f	",c,scores[c],genome_scores[c], random_scores[c], scaledprob2prob(scores[c]), scaledprob2prob(genome_scores[c]), scaledprob2prob(random_scores[c]));
				//}
				//fprintf(stderr,"%f\n",sum);
				for(j = 0;j < hit;j++){
					if(c == j){
						sum += scores[j] + genome_scores[j];
					}else{
						sum += random_scores[j] + prob2scaledprob(1.0 - scaledprob2prob(genome_scores[j])) ;
						//sum += random_scores[j] + prob2scaledprob(0.1 * 1.0/ (float) hit);
					}
				}
				//if(i < 5){
				//	fprintf(stderr,"TOTAL:%f	%f\n",sum,scaledprob2prob(sum));
				//}
				//sum += prob2scaledprob(0.9);
				scores[c] = sum;
			}
			
				
			sum = unaligned;
			//sum = scores[0];
			for(c = 0; c < hit;c++){
				sum = logsum(sum,scores[c]);
			}
			unaligned = unaligned - sum;
			
			max = prob2scaledprob(0.0);
			
			for(c = 0 ; c < hit;c++){
				
				scores[c] = scores[c] - sum;
				if(scores[c] > max){
					max = scores[c];
				}
			}
			
		
			
			hit = 0;
			//c =0;
			for(j = 0; j < num_scores;j++){
				flag = 0;
				if(unaligned  > max){
					flag = 0x200;
				}
				if(scores[j]  == max){
					
				}else if(scaledprob2prob(scores[j]) > (1.0 / (double)LIST_STORE_SIZE)){
					flag |= 0x100;
				}else{
					flag = -1;
				}
				
				if(flag != -1){
					if(ri[i]->strand[j] != 0){
						//reverse_complement2(ri[i]->seq,ri[i]->len);
						strand = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,1);
						seq = transform_genome_sequence(seq , ri[i]->len+20);
						//reverse_complement2(ri[i]->seq,ri[i]->len);
						aln = glocal_viterbi_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20); 
						aln = reverse_path(aln);
						ri[i]->seq = reverse_complement_16(ri[i]->seq,ri[i]->len);
						align_to_sam(
									 aln,
									 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
									 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
									 ri[i]->strand[j] | flag, 
									 scaledprob2prob( scores[j]),
									 ri[i],
									 0
									 );
						free(aln);
						ri[i]->seq = reverse_complement_16(ri[i]->seq,ri[i]->len);
						
					}else{
						strand = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,0);
						seq = transform_genome_sequence(seq , ri[i]->len+20);
						aln = glocal_viterbi_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ); 
						align_to_sam(
									 aln,
									 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
									 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
									 ri[i]->strand[j] | flag, 
									 scaledprob2prob(scores[j] ),
									 ri[i],
									 0
									 );
						free(aln);
					}
				}
			}
			
			
			/*sum = scores[0];
			genome_sum = genome_scores[0];
			for(c = 1; c < hit;c++){
				sum = logsum(sum,scores[c]);
				genome_sum = logsum(genome_sum, genome_scores[c]);
			}
			for(c = 0; c < hit;c++){
				if(i < 5){
					fprintf(stderr,"%d	%f	%f	%f	%f	",c,genome_scores[c]-genome_sum,scores[c] -sum,  scaledprob2prob(genome_scores[c]-genome_sum), scaledprob2prob(scores[c] -sum)   );
				}
				scores[c] = scores[c] -sum  + genome_scores[c]-genome_sum;
				if(i < 5){
					fprintf(stderr,"%f	%f\n",scores[c] ,scaledprob2prob(scores[c] ));
				}
			}*/
			
			
			/*sum = scores[0];
			for(j = 1; j < num_scores;j++){
				sum = logsum(sum,scores[j]);
			}*/
			
			for(j = 0; j < LIST_STORE_SIZE+1;j++){
				top_scores[j] = 0ul;
			}
			if(strcmp("B1B91D34D_1",(const char*) ri[i]->name) == 0){
				fprintf(stderr,"u:%f	%f	(sum:%f)\n",max,scaledprob2prob(max),sum);
				for(c = 0 ; c < num_scores;c++){	
					fprintf(stderr,"%d:%f	%f	(sum:%f)\n",c,scores[c],scaledprob2prob(scores[c]),sum);
				}
			}
			/*
			for(j = 0; j < num_scores;j++){
				
				BinaryInsertionSortOne(top_scores,LIST_STORE_SIZE,(unsigned long int) (scaledprob2prob( scores[j]  -sum ) * 1000000.0f));
			}
			
			//BinaryInsertionSortOne(top_scores,LIST_STORE_SIZE,(unsigned long int) (scaledprob2prob( max  -sum ) * 1000000.0f));
			
						
			
			
			
			
			
			
			
			j = 1;
			top_cut = top_scores[param->upper_limit-j];
			while(top_cut == 0.0f){
				j++;
				top_cut = top_scores[param->upper_limit-j];
			}
						
			hit = 0;
			
			if(num_scores == 1){
				hit = 0;
				c = 0;
				for(j = 0; j < num_scores;j++){
					if((unsigned long int) (scaledprob2prob( scores[j]  -sum) * 1000000.0f) >= top_cut){
					//	if(scaledprob2prob(  (scores[j] + prob2scaledprob(model_prior)) - ( logsum( scores[j] + prob2scaledprob(model_prior),  random_scores[j] + prob2scaledprob(random_model_prior)))) >= unaligned_probability_cutoff){						
						c++;
						if(ri[i]->strand[j] != 0){
							//reverse_complement2(ri[i]->seq,ri[i]->len);
							strand = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,1);
							seq = transform_genome_sequence(seq , ri[i]->len+20);
							//reverse_complement2(ri[i]->seq,ri[i]->len);
							aln = glocal_viterbi_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 
																); 
							aln = reverse_path(aln);
							
							
							ri[i]->seq = reverse_complement_16(ri[i]->seq,ri[i]->len);
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 ri[i]->mapq,
										 ri[i],
										 0
										 );
							free(aln);
							ri[i]->seq = reverse_complement_16(ri[i]->seq,ri[i]->len);
							
						}else{
							strand = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,0);
							seq = transform_genome_sequence(seq , ri[i]->len+20);
							aln = glocal_viterbi_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ); 
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 ri[i]->mapq,
										 ri[i],
										 0
										 );
							free(aln);
						}
					//	}else{
					///		unaligned_to_sam(ri[i]);	
					//		hit = param->upper_limit;
					//		break;
					//	}
						hit++;
						if(hit == param->upper_limit){
							break;
						}
					}
				}
				if(!c){
					unaligned_to_sam(ri[i]);	
					hit = param->upper_limit;
					//	break;
				}
				
			}else{
				hit = 0;
				c =0;
				for(j = 0; j < num_scores;j++){
					if((unsigned long int) (scaledprob2prob(scores[j]  -sum) * 1000000.0f) >= top_cut){
						c++;
					//	if(scaledprob2prob(  (scores[j] + prob2scaledprob(model_prior)) - ( logsum( scores[j] + prob2scaledprob(model_prior),  random_scores[j] + prob2scaledprob(random_model_prior)))) >= unaligned_probability_cutoff){

						if(ri[i]->strand[j] != 0){
							//reverse_complement2(ri[i]->seq,ri[i]->len);
							strand = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,1);
							seq = transform_genome_sequence(seq , ri[i]->len+20);
							//reverse_complement2(ri[i]->seq,ri[i]->len);
							aln = glocal_viterbi_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20); 
							aln = reverse_path(aln);
							ri[i]->seq = reverse_complement_16(ri[i]->seq,ri[i]->len);
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 scaledprob2prob( scores[j] - sum),
										 ri[i],
										 0
										 );
							free(aln);
							ri[i]->seq = reverse_complement_16(ri[i]->seq,ri[i]->len);
							
						}else{
							strand = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,0);
							seq = transform_genome_sequence(seq , ri[i]->len+20);
							aln = glocal_viterbi_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ); 
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 scaledprob2prob(scores[j] - sum),
										 ri[i],
										 0
										 );
							free(aln);
						}
					//	}else{
					//		unaligned_to_sam(ri[i]);	
					//		hit = param->upper_limit;
					//		break;
					//	}
						hit++;
						if(hit == param->upper_limit){
							break;
						}
					}
				}
				
				if(!c){
					unaligned_to_sam(ri[i]);	
					hit = param->upper_limit;
					//	break;
				}
								
			}*/
		}
	}
	free(random_scores);
	free(genome_scores );
	free(scores);
	free(top_scores);	
	free(seq);
	pthread_exit((void *) 0);
}




struct hmm2* run_pHMM2(struct hmm2* localhmm,struct read_info** ri,struct db* db,struct parameters* param,int size)
{
	struct thread_data2* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data2)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	//hmm = init_hmm(lena, lenb, 0.002f, 0.001f, 0.02f, 1);
	
	
	interval =  (int)((double)size /(double)param->num_threads);
	
	fprintf(stderr,"Starting with %d sequences\n",size);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].db = db;
		thread_data_array[t].hmm = init_2nd_hmm(500, 500);
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].db = db;
	thread_data_array[param->num_threads - 1].hmm = init_2nd_hmm(500, 500);
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = size;
	thread_data_array[param->num_threads - 1].param = param;
	
	init_thread_hmm2s(thread_data_array,localhmm,param->num_threads);
	
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, do_baum_welch_thread2, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	//fprintf(stderr,"local:%d	%f	%f	%f\n",t, localhmm->genome_model_t[0][0],localhmm->genome_model_t[0][1],localhmm->genome_model_t[0][2]);
	//for(t = 0;t < param->num_threads;t++) {
	//	fprintf(stderr,"Tread:%d	%f	%f	%f\n",t,thread_data_array[t].hmm->genome_model_t[0][0],thread_data_array[t].hmm->genome_model_t[0][1],thread_data_array[t].hmm->genome_model_t[0][2]);
	//}
	
	
	entangle_hmm2s(thread_data_array,localhmm,param->num_threads);
	//fprintf(stderr,"local:%d	%f	%f	%f\n",t, localhmm->genome_model_t[0][0],localhmm->genome_model_t[0][1],localhmm->genome_model_t[0][2]);
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	for(t = 0;t < param->num_threads;t++) {
		free_2nd_hmm(thread_data_array[t].hmm);
		//print_hmm_param(thread_data_array[t].hmm);
	}
	
	free(thread_data_array);
	return localhmm;
}


struct hmm2* run_pHMM2_random(struct hmm2* localhmm,struct read_info** ri,struct db* db,struct parameters* param,int size)
{
	struct thread_data2* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data2)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	
	interval =  (int)((double)size /(double)param->num_threads);
	
	
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].db = db;
		thread_data_array[t].hmm = init_2nd_hmm(500, 500);
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].db = db;
	thread_data_array[param->num_threads - 1].hmm = init_2nd_hmm(500, 500);
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = size;
	thread_data_array[param->num_threads - 1].param = param;
	
	init_thread_hmm2s(thread_data_array,localhmm,param->num_threads);
	
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, do_baum_welch_thread2_random_model, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}

	entangle_hmm2s(thread_data_array,localhmm,param->num_threads);
	for(t = 0;t < param->num_threads;t++) {
		free_2nd_hmm(thread_data_array[t].hmm);
	}
	free(thread_data_array);
	return localhmm;
}



void* do_baum_welch_thread2_random_model(void *threadarg)
{
	struct thread_data2 *data;
	data = (struct thread_data2 *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm2* hmm = data->hmm;
	struct db* db = data->db;
	const int start = data->start; 
	const int end = data->end;
	int i,hit;	
	int strand = 0;
	unsigned char* seq = 0;
	
	seq = malloc(sizeof(unsigned char) * 500);

	
	for(i = start; i < end;i++){
		hit = 0;
		if(ri[i]->identity[0]  / (float) ri[i]->len  <=  data->param->expected_error){
			hmm->train_count++;
		}
		while(ri[i]->identity[hit] != -2.0f){
			
			if(ri[i]->strand[hit] !=0){
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
			}else{
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
			}
			seq = transform_genome_sequence(seq , ri[i]->len+20);
			//calc naked genome prob;
			hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),0);
			
			hit++;
		}
		//calc naked read prob;
		hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),1);
		
	}

	free(seq);
	pthread_exit((void *) 0);
}

void* do_baum_welch_thread2(void *threadarg)
{
	struct thread_data2 *data;
	data = (struct thread_data2 *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm2* hmm = data->hmm;
	struct db* db = data->db;
	const int start = data->start; 
	const int end = data->end;
	int i,j,hit;	
	int strand = 0;
	unsigned char* seq = 0;
	float pseudocount = 0.0f;
	seq = malloc(sizeof(unsigned char) * 500);
	float avg_score = 100000000;
	double sum = 0;
	//float genome_sum = 0;
	//float prior_sum = 0;
	int c;
	float* scores = malloc(sizeof(float)* 100); 
	float* random_scores = malloc(sizeof(float)* 100); 
	float* genome_scores = malloc(sizeof(float)* 100); 
	float max;
	for(i = start; i < end;i++){
		
		
		hit = 0;
		max = -SCALEINFTY;
		while(ri[i]->identity[hit] != -2.0f){
			if(ri[i]->strand[hit] !=0){
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
			}else{
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
			}
			seq = transform_genome_sequence(seq , ri[i]->len+20);
			//hmm =  score_genome(hmm,seq,ri[i]->len+20 );
			
			hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ,hit); 
			hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),0);
			//scores[hit] = hmm->score + ri[i]->priors[hit];
			//if(scores[hit] > max){
			//	max = scores[hit];
			//}
			
			scores[hit] = hmm->score;// + ri[i]->priors[hit];
			
			genome_scores[hit] = ri[i]->priors[hit];
			
			random_scores[hit] =  hmm->random_score ;
			
			hit++;
		}
		
		/*sum = scores[0];
		 for(c = 1; c < hit;c++){
		 sum = logsum(sum,scores[c]);
		 }*/
		sum = genome_scores[0];
		for(c = 1; c < hit;c++){
			sum = logsum(sum,genome_scores[c]);
		}
		for(c = 0; c < hit;c++){
			genome_scores[c] = genome_scores[c] -sum + prob2scaledprob(0.9);
		}
		
		hmm = random_model_calc(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , prob2scaledprob(0.5),1);
		sum = hmm->random_score + prob2scaledprob( 0.1);
		for(c = 0; c < hit;c++){
			sum += random_scores[c] + prob2scaledprob(1.0 - scaledprob2prob(genome_scores[c]));
		}
		//fprintf(stderr,"%f	%f\n", sum, prob2scaledprob( (1.0 / 0.9)  - 1.0 ));
		
		max = sum ;
		
		
		for(c = 0;c < hit;c++){
			sum = prob2scaledprob(1.0f);
			for(j = 0;j < hit;j++){
				if(c == j){
					//read is aligned at this seed position 
					sum += scores[j] + genome_scores[j];
				}else{
					//read not aligned at this seed
					sum += random_scores[j] + prob2scaledprob(1.0 - scaledprob2prob(genome_scores[j])) ;
					//sum += random_scores[j] + prob2scaledprob(0.1 * 1.0/ (float) hit);
				}
			}
			
			//read is aligned to genome  
			sum += prob2scaledprob(0.9);
			
			scores[c] = sum;
		}
		
		
		
		
		
		
		if (data->param->train_pair_hmm ){
			sum = prob2scaledprob(0.0);
		}else{
			sum = max;
		}
		
		for(c = 0; c < hit;c++){
			sum = logsum(sum,scores[c]);
			
		}
		
		if(i < 5){
			fprintf(stderr,"u:%f	%f	(sum:%f)\n",max,scaledprob2prob(max-sum),sum);
		}
		for(c = 0 ; c < hit;c++){	
			if(i < 5){
				fprintf(stderr,"%d:%f	%f	(sum:%f)\n",c,scores[c],scaledprob2prob(scores[c]-sum),sum);
			}
		}
		max = -SCALEINFTY;
		for(c = 0; c < hit;c++){
			if(scores[c] > max){
				max = scores[c];
			}
		}
		
		//pseudocount = prob2scaledprob( 1.0 / hit );
		
		hit = 0;
		while(ri[i]->identity[hit] != -2.0f){
			ri[i]->priors[hit] =  scores[hit] -sum;
			ri[i]->priors[hit] = logsum(ri[i]->priors[hit], pseudocount);
			hit++;
		}
		/*sum = scores[0];
		 for(c = 1 ; c < hit;c++){
		 sum = logsum(sum,scores[c]);
		 }*/
		
		
		if(avg_score == 100000000){ 
			avg_score = max ;
		}else{
			avg_score =  logsum(avg_score, max ); //// CAUSES CRASHes... 
		}
		if(ri[i]->identity[0]  / (float) ri[i]->len  <=  data->param->expected_error){
			hmm->train_count++;
			hit = 0;
			while(ri[i]->identity[hit] != -2.0f){
				
				//if(scaledprob2prob(scores[hit] -sum) >= 0.1){// heuristic....
				if(ri[i]->strand[hit] != 0){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
					
				}else{
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				}
				seq = transform_genome_sequence(seq , ri[i]->len+20);
				//hmm =  score_genome(hmm,seq,ri[i]->len+20 );
				//hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,-1 ); 
				//if(i < 10){
				//	fprintf(stderr,"%d	%f	%f	%f\n",hit, hmm->score, hmm->tfY[hit][ri[i]->len][ri[i]->len+20],  hmm->score - hmm->tfY[hit][ri[i]->len][ri[i]->len+20]);
				//}
				
				hmm = glocal_backward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20); 
				//hmm =  read_genome_posteriors(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20); 
				
				hmm = get_prob_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20, scores[hit] - sum,hit);// hmm->score + hmm->score - sum + ri[i]->priors[hit]  );
				
				//hmm = get_genome_prob(hmm,  seq,ri[i]->len+20,  prob2scaledprob(1.0));//   scores[hit] -sum );
				
				//hmm = get_prob_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,scores[hit]+ ri[i]->priors[hit] +logsum(prob2scaledprob(0.95), genome_scores[hit] - prob2scaledprob(0.05)));
				//ri[i]->priors[hit] =  scores[hit] -sum;
				//ri[i]->priors[hit] = logsum(ri[i]->priors[hit], pseudocount);
				hit++;
			}
		}
	}
	free(genome_scores);
	free(random_scores);
	free(scores);
	if(!start){
		fprintf(stderr,"SCORE:%d		%f	%f\n",start,avg_score,   scaledprob2prob(avg_score - prob2scaledprob(end-start)) );
	}
	free(seq);
	pthread_exit((void *) 0);
}

float* read_genome_posteriors(struct hmm2* hmm,unsigned char* a, unsigned char* b, int n,int m, float frac,int count_mismatches_posterior)
{
	float* posterior = malloc(sizeof(float)* (m+1) );
	int i,j;
	//float len_norm = 0;
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	
	//float test = prob2scaledprob(0.0f);
	for(i = 0; i <= m;i++){
		posterior[i] = prob2scaledprob(0.0f);
	}
	if(count_mismatches_posterior){
		for(i = 1; i <= n;i++){
			for(j = 1; j <= m;j++){
				//if((seqa[i]& 0x3) == (seqb[j]& 0x3)){
				posterior[j-1] = logsum(posterior[j-1], hmm->fM[i][j] + hmm->bM[i][j]);
				//}
				
			//	len_norm++;
			}
		}
	}else{
		for(i = 1; i <= n;i++){
			for(j = 1; j <= m;j++){
				if((seqa[i]& 0x3) == (seqb[j]& 0x3)){
					posterior[j-1] = logsum(posterior[j-1], hmm->fM[i][j] + hmm->bM[i][j]);
			//		len_norm++;
				}
			}
		}
	}
	for(i = 0; i < m;i++){
		//fprintf(stderr,"%d	%f	%f	%f\n",i,posterior[i],posterior[i] - hmm->score + frac, scaledprob2prob(posterior[i] - hmm->score + frac));
		posterior[i] = posterior[i] - hmm->score + frac;
		//test = logsum(test, posterior[i] );
	}
	
	//fprintf(stderr,"SUM:%f	%d\n",scaledprob2prob(test),n);
	//posterior[0] = posterior[1];
	return posterior;
	
	/*for(i = 1; i <= m;i++){
		//fprintf(stderr,"%0.2f ",scaledprob2prob(posterior[i]-hmm->score) - scaledprob2prob(posterior[i-1]-hmm->score) );
		fprintf(stderr,"%0.2f ",scaledprob2prob(posterior[i]-hmm->score) );
	}
	
	fprintf(stderr,"\n");
	
	for(i = 1; i <= m;i++){
		fprintf(stderr,"%0.2f ",scaledprob2prob(posterior[i]-hmm->score) - scaledprob2prob(posterior[i-1]-hmm->score) );
		//fprintf(stderr,"%0.2f ",scaledprob2prob(posterior[i]-hmm->score) );
	}
	
	fprintf(stderr,"\n");
	
	free(posterior);*/
	//return hmm;
}


unsigned char* reverse_path(unsigned char* org )
{
	unsigned char* new = 0;
	
	int i,c;
	int nuc[5] = {3,2,1,0,4};
	new = malloc(sizeof(unsigned char)* (org[0]+1));
	
	new[0] = org[0];
	c = 1;
	for (i = org[0]; i > 0; i--){
		if(( org[i] >> 4) < 5){
			if((org[i] & 0xF) < 5){
				new[c] = (nuc[org[i] >> 4] << 4) |  nuc[org[i] & 0xF];
			}else{
				new[c] = (nuc[org[i] >> 4] << 4) |  (org[i] & 0xF);
			}
		}else{
			if((org[i] & 0xF) < 5){
				new[c] = (org[i] & 0xF0) |  nuc[org[i] & 0xF];
			}else{
				new[c] =org[i];
			}
		}
		//new[c] = org[i]; 
		c++;
	}
		
	free(org);
	return new;
}

float* reverse_genome_posteriors(float* a, int len)
{
	float* new = 0;
	int i,c;
	new = malloc(sizeof(float)* (len+1));
	
	//new[0] = a[0];
	c = 0;
	for (i = len; i >= 0; i--){
		new[c] = a[i];
		c++;
	}
	
	free(a);
	return new;
}



void simple_print_path(unsigned char* alignment)
{
	
	int i,c;
	
	c = alignment[0];
	fprintf(stderr,"SeqA:");
	for(i = c; i >=1;i--){
		if((alignment[i] >> 4) == 5){
			fprintf(stderr,"-");
		}else{
			fprintf(stderr,"%d",alignment[i] >> 4);
		}
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"SeqB:");
	for(i = c; i >=1;i--){
		if((alignment[i] & 0xF) == 5){
			fprintf(stderr,"-");
		}else{
			fprintf(stderr,"%d",alignment[i] & 0xF);
		}
	}
	fprintf(stderr,"\n");
	
}



unsigned char* glocal_viterbi_2ndorder_log_Y(struct hmm2* hmm,unsigned char* a, unsigned char* b, int n,int m)
{
	//init - terminal gap penalty = 0 (p = 1)
	int i,j,c;
	
	unsigned short int tmp_tb;
	float** M = hmm->fM;
	//float** C = hmm->fC;
	float** X = hmm->fX;
	float** Y = hmm->fY;
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	float* prob = hmm->prob;	
	unsigned short int** t = hmm->traceback;
	unsigned char* alignment =  malloc(sizeof(unsigned char)* (n+m+2));
	//unsigned short int* state_path =  malloc(sizeof(unsigned short int)* (n+m+2));
	
	
	
	t[0][2] = 4 << 6;
	
	
	
	c = 0;
	M[0][0] = prob2scaledprob(0.0f);
	X[0][0] = prob2scaledprob(0.0f);
	Y[0][0] = prob2scaledprob(0.0f);
	
	M[0][1] = prob2scaledprob(0.0f);
	X[0][1] = prob2scaledprob(0.0f);
	Y[0][1] = prob2scaledprob(0.0f);
	tmp_tb = (4 << 6);
	t[0][1] = tmp_tb;
	
	M[0][2] = prob2scaledprob(0.0f);
	X[0][2] = prob2scaledprob(0.0f);
	Y[0][2] = prob[   (EGAP << 12) | (seqb[1] << 8) | (EGAP << 4) | seqb[2]];
	tmp_tb = (4 << 6);
	t[0][2] = tmp_tb;
	
	//fprintf(stderr,"%d start %d\n",i,c);
	for(j = 3; j <= m-2;j++){
		
		M[0][j] = prob2scaledprob(0.0f);
		X[0][j] = prob2scaledprob(0.0f);
		Y[0][j] = prob[   (EGAP << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j]] + Y[0][j-1];
		tmp_tb = (4 << 6);
		t[0][j] = tmp_tb;
		//Y[i][j] = logsum(Y[i][j], prob[(seqa[i] << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] );
	}
	
	M[0][m-1] = prob2scaledprob(0.0f);
	X[0][m-1] = prob2scaledprob(0.0f);
	Y[0][m-1] = prob2scaledprob(0.0f);
	t[0][m-1] = 0;
	
	M[0][m] = prob2scaledprob(0.0f);
	X[0][m] = prob2scaledprob(0.0f);
	Y[0][m] = prob2scaledprob(0.0f);
	t[0][m] = 0;
	
	i = 1;
	
	M[i][0] = prob2scaledprob(0.0f);
	X[i][0] = prob2scaledprob(0.0f);
	Y[i][0] = prob2scaledprob(0.0f);
	t[i][0] = 0;
	M[i][1] = prob2scaledprob(0.0f);
	X[i][1] = prob2scaledprob(0.0f);
	Y[i][1] = prob2scaledprob(0.0f);
	
	t[i][1] = 0;
	
	
	for(j = 2; j <= m-2;j++){
		
		
		M[i][j] = prob[              (EGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[i-1][j-1];
		tmp_tb = 4;
		t[i][j] = tmp_tb;
		
		

		X[i][j] = prob[              (EGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[i-1][j];
		tmp_tb = 4 << 3;
		t[i][j] |= tmp_tb;
		
		tmp_tb = 4 << 6;
		Y[i][j] =                            prob[           (IGAP << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + Y[i][j-1];
		
		if(prob[(seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + M[i][j-1] > Y[i][j]){
			Y[i][j] = prob[(seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + M[i][j-1];
			tmp_tb = 1 << 6;
		}
		if(prob[(seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j]] + X[i][j-1] > Y[i][j]){
			Y[i][j] = prob[(seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j]] + X[i][j-1];
			tmp_tb = 2 << 6;
			
		}
		t[i][j] |= tmp_tb;
		
	}
	M[i][m-1] = prob2scaledprob(0.0f);
	X[i][m-1] = prob2scaledprob(0.0f);
	Y[i][m-1] = prob2scaledprob(0.0f);
	t[i][m-1] = 0;
	
	M[i][m] = prob2scaledprob(0.0f);
	X[i][m] = prob2scaledprob(0.0f);
	Y[i][m] = prob2scaledprob(0.0f);
	t[i][m] = 0;
	
	
	for(i = 2; i < n;i++){//= 4;i++){
		
		
		M[i][0] = prob2scaledprob(0.0f);
		X[i][0] = prob2scaledprob(0.0f);
		Y[i][0] = prob2scaledprob(0.0f);
		t[i][0] = 0;
		M[i][1] = prob2scaledprob(0.0f);
		X[i][1] = prob2scaledprob(0.0f);
		Y[i][1] = prob2scaledprob(0.0f);
		
		t[i][1] = 0;
		
		
		for(j = 2; j <= m-2;j++){
			
			tmp_tb = 1;
			M[i][j] =                             prob[(seqa[i-1] << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + M[i-1][j-1]; 
			
			if(prob[(seqa[i-1] << 12) |               (IGAP << 8) | (seqa[i] << 4) | (seqb[j])] + X[i-1][j-1] > M[i][j]){
				M[i][j] = prob[(seqa[i-1] << 12) |               (IGAP << 8) | (seqa[i] << 4) | (seqb[j])] + X[i-1][j-1];
				tmp_tb = 2;
			}
			
			if(prob[              (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[i-1][j-1] > M[i][j]){
				M[i][j] = prob[              (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[i-1][j-1];
				tmp_tb = 4;
			}
			t[i][j] = tmp_tb;
			
			
			tmp_tb = 2 << 3;
			X[i][j] =                            prob[(seqa[i-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP] + X[i-1][j] ;
			if(prob[(seqa[i-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + M[i-1][j] > X[i][j]){
				X[i][j] = prob[(seqa[i-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + M[i-1][j];
				tmp_tb = 1 << 3;
			}
			if(prob[              (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[i-1][j] > X[i][j]){
				X[i][j] = prob[              (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[i-1][j];
				tmp_tb = 4 << 3;
			}
			t[i][j] |= tmp_tb;
			
			tmp_tb = 4 << 6;
			Y[i][j] =                            prob[           (IGAP << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + Y[i][j-1];
			
			if(prob[(seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + M[i][j-1] > Y[i][j]){
				Y[i][j] = prob[(seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + M[i][j-1];
				tmp_tb = 1 << 6;
			}
			if(prob[(seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j]] + X[i][j-1] > Y[i][j]){
				Y[i][j] = prob[(seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j]] + X[i][j-1];
				tmp_tb = 2 << 6;
			
			}
			t[i][j] |= tmp_tb;
			
		}
		M[i][m-1] = prob2scaledprob(0.0f);
		X[i][m-1] = prob2scaledprob(0.0f);
		Y[i][m-1] = prob2scaledprob(0.0f);
		t[i][m-1] = 0;
		
		M[i][m] = prob2scaledprob(0.0f);
		X[i][m] = prob2scaledprob(0.0f);
		Y[i][m] = prob2scaledprob(0.0f);
		t[i][m] = 0;
		//if(c > 5){
		//	c = 5;
		//}
		//c++;
	}
	
	
	M[n][0] = prob2scaledprob(0.0f);
	X[n][0] = prob2scaledprob(0.0f);
	Y[n][0] = prob2scaledprob(0.0f);
	t[n][0] = 0;
	
	M[n][1] = prob2scaledprob(0.0f);
	X[n][1] = prob2scaledprob(0.0f);
	Y[n][1] = prob2scaledprob(0.0f);
	t[n][1] = 0;
	
	for(j = 2; j <= m-2;j++){
		tmp_tb = 1;
		M[n][j] =                             prob[(seqa[n-1] << 12) | (seqb[j-1] << 8) | (seqa[n] << 4) | (seqb[j])] + M[n-1][j-1];
		
		if(prob[(seqa[n-1] << 12) |               (IGAP << 8) | (seqa[n] << 4) | (seqb[j])] + X[n-1][j-1] > M [n][j]){
			M[n][j] = prob[(seqa[n-1] << 12) |               (IGAP << 8) | (seqa[n] << 4) | (seqb[j])] + X[n-1][j-1];
			tmp_tb  = 2;
		}
		
		if(prob[              (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[n-1][j-1] > M[n][j]){
			M[n][j] = prob[              (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[n-1][j-1];
			tmp_tb = 4;
		}
		t[n][j] = tmp_tb;
		
		tmp_tb = 2 << 3;
		X[n][j] =                            prob[(seqa[n-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP] + X[n-1][j] ;
		if(prob[(seqa[n-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + M[n-1][j] > X[n][j]){
			X[n][j] = prob[(seqa[n-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + M[n-1][j];
			tmp_tb = 1 << 3;
		}
		if(prob[              (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[n-1][j] > X[n][j]){
			X[n][j] = prob[              (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[n-1][j];
			tmp_tb = 4 << 3;
		}
		t[n][j] |= tmp_tb;
		
		tmp_tb = 4 << 6;
		Y[n][j] =                            prob[           (EGAP << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j]] + Y[n][j-1];
		if(prob[(seqa[n] << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j]] + M[n][j-1] > Y[n][j]){
			Y[n][j] = prob[(seqa[n] << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j]] + M[n][j-1];
			tmp_tb = 1 << 6;
		}
		if(prob[(seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[j]] + X[n][j-1] > Y[n][j]){
			Y[n][j] = prob[(seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[j]] + X[n][j-1];
			tmp_tb = 2 << 6;
		
		}
		t[n][j] |= tmp_tb;
				
	}
	M[n][m-1] = prob2scaledprob(0.0f);
	X[n][m-1] = prob2scaledprob(0.0f);
	
	tmp_tb = 4 << 6;
	Y[n][m-1] =                            prob[           (EGAP << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1]] + Y[n][m-2];
	
	if(prob[(seqa[n] << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1]] + M[n][m-2] > Y[n][m-1]){
		Y[n][m-1] = prob[(seqa[n] << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1]] + M[n][m-2];
		tmp_tb = 1 << 6;
	}
	
	if(prob[(seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[m-1]] + X[n][m-2] > Y[n][m-1]){
		Y[n][m-1] = prob[(seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[m-1]] + X[n][m-2];
		tmp_tb = 2 << 6;
		
	}
	
	t[n][m-1] = tmp_tb;	
	//fprintf(stderr,"%d	%d	 = %d\n",n,m-1,tmp_tb);
	
	M[n][m] = prob2scaledprob(0.0f);
	X[n][m] = prob2scaledprob(0.0f);
	//Y[n][m] = prob2scaledprob(0.0f);
	Y[n][m] =                            prob[           (EGAP << 12) | (seqb[m-1] << 8) | (EGAP << 4) | seqb[m]] + Y[n][m-1];
	t[n][m] = 4 << 6;
	
	i = n;
	j = m;
	if(M[i][j] >= Y[i][j]){
		if(M[i][j] >= X[i][j]){
			tmp_tb = 1;
		}else{
			tmp_tb = 2;
		}
	}else{
		if(Y[i][j] > X[i][j]){
			tmp_tb = 4;
		}else{
			tmp_tb = 2;
		}
		
	}
	tmp_tb = 4;
	float id = 0.0f;
	c = 1;
	//state_path[c] = tmp_tb;
	//c++;
	
	//for(j = 0; j < n;j++){
	//	fprintf(stderr,"%d ",(int) a[j] );
	//}
	//fprintf(stderr,"\n");
	//
	//for(j = 0; j < m;j++){
	//	fprintf(stderr,"%d ",(int) b[j] );
	//}
	//fprintf(stderr,"\n");
	
	//tmp_tb = 4;
	//if(Y[i][j] > R[i][j]){
	
	//	tmp_tb = 3;
	//}else{
	//	tmp_tb = 4;
	//}
	//tmp_tb = 4;
	//fprintf(stderr,"%f	%f	%f\n",M[i][j],X[i][j],Y[i][j]);
	
	
	while(1){
		//fprintf(stderr,"%d	%d	%d	%d\n",i,j,c,tmp_tb);
		switch (tmp_tb) {
			case 1:
				alignment[c] = ((seqa[i]&0x3) << 4) | (seqb[j]&0x3);
				
				if(seqa[i] == seqb[j]){
					id += 1.0f;
				}
				
				tmp_tb = (t[i][j]  ) & 0x7;
				i--;
				j--;
				break;
			case 2:
				alignment[c] = ((seqa[i]&0x3) << 4)  | 5;
				tmp_tb = (t[i][j] >> 3 ) & 0x7;
				i--;	
				break;
			case 4:
				alignment[c] = (5 << 4) | (seqb[j]&0x3) ;
				tmp_tb = (t[i][j] >> 6 ) & 0x7;
				j--;	
				break;
				
			default:
				fprintf(stderr,"ERROR:code: %d not found\n",tmp_tb);
				exit(0);
				break;
		}
		c++;
		//state_path[c] = tmp_tb;
		
		if(i == 0 && j == 0){
			break;
		}
	}
	alignment[0] = c-1; 
	
	
	
	/*fprintf(stderr,"SeqA:");
	for(i = c-1; i >=1;i--){
		if((alignment[i] >> 4) == EGAP){
			fprintf(stderr,"-");
		}else{
			fprintf(stderr,"%d",alignment[i] >> 4);
		}
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"SeqB:");
	for(i = c-1; i >=1;i--){
		if((alignment[i] & 0xF) ==EGAP){
			fprintf(stderr,"-");
		}else{
			fprintf(stderr,"%d",alignment[i] & 0xF);
		}
	}
	fprintf(stderr,"\n");*/
	//fprintf(stderr,"Path:");
	
	
	/*for(i = c-1; i >=1;i--){
		switch (state_path[i]) {
			case 1:
				fprintf(stderr,"M");
				break;
			case 2:
				fprintf(stderr,"X");
				
				break;
			case 4:
				fprintf(stderr,"Y");
				
				break;
			default:
				break;
		}
	}
	
	
	fprintf(stderr,"	%0.1f\n\n",id / (float) n * 100.0f );
	free(state_path);*/
	return alignment;
}

struct hmm2* get_prob_2ndorder_log_Y(struct hmm2* hmm,unsigned char* a, unsigned char* b, int n,int m,float frac,int frame )
{
	int i,j,c,g;
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	float* prob = hmm->prob;	
	float* tprob = hmm->tprob;
	
	float** M = 0;
	float** X = 0;
	float** Y = 0;
	
	if(frame == -1){
		M = hmm->fM;
		X = hmm->fX;
		Y = hmm->fY;
	}else{
		M = hmm->fM;
		X = hmm->fX;
		Y = hmm->fY;
		hmm->score =  hmm->tfY[frame][n][m];
		
		hmm->fM = hmm->tfM[frame];
		hmm->fX = hmm->tfX[frame];
		hmm->fY = hmm->tfY[frame];
	}
	
	
	//for(i = 0;i <= n;i++){
	//	for(j = 0;j <=m;j++){
	//		M[i][j] =  prob2scaledprob(1.0f);
	//		X[i][j] =  prob2scaledprob(1.0f);
	//		Y[i][j] =  prob2scaledprob(1.0f);
	//	}
	//}
	c = 0;
	
	//M[0][0] = prob2scaledprob(0.0f);
	//X[0][0] = prob2scaledprob(0.0f);
	//Y[0][0] = prob2scaledprob(0.0f);
	
	//M[0][1] = prob2scaledprob(0.0f);
	//X[0][1] = prob2scaledprob(0.0f);
	//Y[0][1] = prob2scaledprob(0.0f);
	
	//M[0][2] = prob2scaledprob(0.0f);
	//X[0][2] = prob2scaledprob(0.0f);
	
	//Y[0][2] = prob[c][   (6 <<9) | (seqb[1] << 6) | (6 << 3) | seqb[2]];
	tprob[   (EGAP << 12) | (seqb[1] << 8) | (EGAP << 4) | seqb[2]] = logsum(tprob[   (EGAP << 12) | (seqb[1] << 8) | (EGAP << 4) | seqb[2]], hmm->fY[0][2] + hmm->bY[0][2]  - hmm->score );
	
	
	//fprintf(stderr,"%d start %d\n",i,c);
	for(j = 3; j <= m-2;j++){
		//M[0][j] = prob2scaledprob(0.0f);
		//X[0][j] = prob2scaledprob(0.0f);
		//Y[0][j] = prob[c][   (6 <<9) | (seqb[j-1] << 6) | (6 << 3) | seqb[j]] + Y[0][j-1];
		g =  (EGAP << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j];
		
		tprob[g]  = logsum(tprob[g],  prob[g] + hmm->fY[0][j-1] + hmm->bY[0][j]  - hmm->score );
		
		//Y[i][j] = logsum(Y[i][j], prob[(seqa[i] << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] );
	}
	
	//M[0][m-1] = prob2scaledprob(0.0f);
	//X[0][m-1] = prob2scaledprob(0.0f);
	//Y[0][m-1] = prob2scaledprob(0.0f);
	
	//M[0][m] = prob2scaledprob(0.0f);
	//X[0][m] = prob2scaledprob(0.0f);
	//Y[0][m] = prob2scaledprob(0.0f);
	
	i = 1;
	
	
	
	for(j = 2; j <= m-2;j++){
		//M[i][j] =                             prob[c][(seqa[i-1] << 9) | (seqb[j-1] << 6) | (seqa[i] << 3) | (seqb[j])] + M[i-1][j-1]; 
				//M[i][j] = logsum(M[i][j], prob[c][(seqa[i-1] << 9) |               (5 << 6) | (seqa[i] << 3) | (seqb[j])] + X[i-1][j-1]);
		
		//M[i][j] = logsum(M[i][j], prob[c][              (5 << 9) | (seqb[j-1] << 6) | (seqa[i] << 3) | (seqb[j])] + Y[i-1][j-1]);
		g =           (EGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j]);
		tprob[g] = logsum(tprob[g], prob[g] +  hmm->fY[i-1][j-1]+ hmm->bM[i][j]  - hmm->score + frac);
		
		//X[i][j] =                            prob[c][(seqa[i-1] << 9) |            (5 << 6) | (seqa[i] << 3) | 5] + X[i-1][j] ;
		
		//X[i][j] = logsum(X[i][j], prob[c][              (5 << 9)  | (seqb[j] << 6) | (seqa[i] << 3) | 5] + Y[i-1][j]);
		g =               (EGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP;
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[i-1][j] + hmm->bX[i][j] - hmm->score + frac);
		
		
		//Y[i][j] =                            prob[c][           (5 << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] + Y[i][j-1];
		g =           (IGAP << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j];
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[i][j-1] + hmm->bY[i][j] - hmm->score + frac);
		
		//Y[i][j] = logsum(Y[i][j] ,prob[c][(seqa[i] << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] + M[i][j-1]);
		g = (seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j];
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[i][j-1] + hmm->bY[i][j] - hmm->score + frac);
		
		//Y[i][j] = logsum(Y[i][j] ,prob[c][(seqa[i] << 9) |               (5 << 6)  | (5 << 3) | seqb[j]] + X[i][j-1]);
		g = (seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j];
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[i][j-1] + hmm->bY[i][j] - hmm->score + frac);
		
		
	}
	
	
	
	for(i = 2; i < n;i++){//= 4;i++){
		
		
		//M[i][0] = prob2scaledprob(0.0f);
		//X[i][0] = prob2scaledprob(0.0f);
		//Y[i][0] = prob2scaledprob(0.0f);
		
		//M[i][1] = prob2scaledprob(0.0f);
		//X[i][1] = prob2scaledprob(0.0f);
		//Y[i][1] = prob2scaledprob(0.0f);
		
		
		for(j = 2; j <= m-2;j++){
			
			//fprintf(stderr,"%f	%f	%f\n",hmm->fM[i-1][j-1] ,M[i-1][j-1] , hmm->fM[i-1][j-1] -M[i-1][j-1]);
			
			//M[i][j] =                             prob[c][(seqa[i-1] << 9) | (seqb[j-1] << 6) | (seqa[i] << 3) | (seqb[j])] + M[i-1][j-1]; 
			g = (seqa[i-1] << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j]);
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[i-1][j-1] + hmm->bM[i][j]  - hmm->score + frac);
			//M[i][j] = logsum(M[i][j], prob[c][(seqa[i-1] << 9) |               (5 << 6) | (seqa[i] << 3) | (seqb[j])] + X[i-1][j-1]);
			g = (seqa[i-1] << 12) |               (IGAP << 8) | (seqa[i] << 4) | (seqb[j]);
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[i-1][j-1] + hmm->bM[i][j]  - hmm->score + frac);
			//M[i][j] = logsum(M[i][j], prob[c][              (5 << 9) | (seqb[j-1] << 6) | (seqa[i] << 3) | (seqb[j])] + Y[i-1][j-1]);
			g =           (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j]);
			tprob[g] = logsum(tprob[g], prob[g] +  hmm->fY[i-1][j-1]+ hmm->bM[i][j]  - hmm->score + frac);

			//X[i][j] =                            prob[c][(seqa[i-1] << 9) |            (5 << 6) | (seqa[i] << 3) | 5] + X[i-1][j] ;
			g = (seqa[i-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP;
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[i-1][j] + hmm->bX[i][j] - hmm->score + frac);
			//X[i][j] = logsum(X[i][j], prob[c][(seqa[i-1] << 9) | (seqb[j] << 6) | (seqa[i] << 3) | 5] + M[i-1][j]);
			g = (seqa[i-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP;
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[i-1][j] + hmm->bX[i][j] - hmm->score + frac);
			//X[i][j] = logsum(X[i][j], prob[c][              (5 << 9)  | (seqb[j] << 6) | (seqa[i] << 3) | 5] + Y[i-1][j]);
			g =               (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP;
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[i-1][j] + hmm->bX[i][j] - hmm->score + frac);
			
			
			//Y[i][j] =                            prob[c][           (5 << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] + Y[i][j-1];
			g =           (IGAP << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j];
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[i][j-1] + hmm->bY[i][j] - hmm->score + frac);
			
			//Y[i][j] = logsum(Y[i][j] ,prob[c][(seqa[i] << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] + M[i][j-1]);
			g = (seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j];
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[i][j-1] + hmm->bY[i][j] - hmm->score + frac);
			
			//Y[i][j] = logsum(Y[i][j] ,prob[c][(seqa[i] << 9) |               (5 << 6)  | (5 << 3) | seqb[j]] + X[i][j-1]);
			g = (seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j];
			tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[i][j-1] + hmm->bY[i][j] - hmm->score + frac);
			
			
		}
		//M[i][m-1] = prob2scaledprob(0.0f);
		//X[i][m-1] = prob2scaledprob(0.0f);
		//Y[i][m-1] = prob2scaledprob(0.0f);
		
		//M[i][m] = prob2scaledprob(0.0f);
		//X[i][m] = prob2scaledprob(0.0f);
		//Y[i][m] = prob2scaledprob(0.0f);
		
		//if(c > 5){
		//	c = 5;
		//}
		//c++;
	}
	
	
	//M[n][0] = prob2scaledprob(0.0f);
	//X[n][0] = prob2scaledprob(0.0f);
	//Y[n][0] = prob2scaledprob(0.0f);
	
	//M[n][1] = prob2scaledprob(0.0f);
	//X[n][1] = prob2scaledprob(0.0f);
	//Y[n][1] = prob2scaledprob(0.0f);
	
	for(j = 2; j <= m-2;j++){
		//M[n][j] =                             prob[c][(seqa[n-1] << 9) | (seqb[j-1] << 6) | (seqa[n] << 3) | (seqb[j])] + M[n-1][j-1]; 
		g = (seqa[n-1] << 12) | (seqb[j-1] << 8) | (seqa[n] << 4) | (seqb[j]);
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[n-1][j-1] + hmm->bM[n][j]  - hmm->score + frac);
		
		//M[n][j] = logsum(M[n][j], prob[c][(seqa[n-1] << 9) |               (5 << 6) | (seqa[n] << 3) | (seqb[j])] + X[n-1][j-1]);
		g = (seqa[n-1] << 12) |               (IGAP << 8) | (seqa[n] << 4) | (seqb[j]);
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[n-1][j-1] + hmm->bM[n][j]  - hmm->score + frac);
		
		//M[n][j] = logsum(M[n][j], prob[c][              (5 << 9) | (seqb[j-1] << 6) | (seqa[i] << 3) | (seqb[j])] + Y[n-1][j-1]);
		g =               (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j]); 
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[n-1][j-1] + hmm->bM[n][j]  - hmm->score + frac);
		
		
		//X[n][j] =                            prob[c][(seqa[n-1] << 9) |            (5 << 6) | (seqa[i] << 3) | 5] + X[n-1][j] ;
		g = (seqa[n-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP;
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[n-1][j] + hmm->bX[n][j] - hmm->score + frac);
		
		//X[n][j] = logsum(X[n][j], prob[c][(seqa[n-1] << 9) | (seqb[j] << 6) | (seqa[i] << 3) | 5] + M[n-1][j]);
		g = (seqa[n-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP;
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[n-1][j] + hmm->bX[n][j] - hmm->score + frac);
		
		//X[n][j] = logsum(X[n][j], prob[c][              (5 << 9)  | (seqb[j] << 6) | (seqa[i] << 3) | 5] + Y[n-1][j]);
		g =                (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP;
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[n-1][j] + hmm->bX[n][j] - hmm->score + frac);
		
		//Y[n][j] =                            prob[c][           (6 << 9) | (seqb[j-1] << 6) | (6 << 3) | seqb[j]] + Y[n][j-1];
		g =            (EGAP << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j];
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[n][j-1] + hmm->bY[n][j] - hmm->score );
		
		//Y[n][j] = logsum(Y[n][j] ,prob[c][(seqa[n] << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] + M[n][j-1]);
		g = (seqa[n] << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j];
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[n][j-1] + hmm->bY[n][j] - hmm->score + frac);
		
		//Y[n][j] = logsum(Y[n][j] ,prob[c][(seqa[n] << 9) |               (5 << 6)  | (5 << 3) | seqb[j]] + X[n][j-1]);
		g = (seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[j];
		tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[n][j-1] + hmm->bY[n][j] - hmm->score + frac);
		
	}
	//M[n][m-1] = prob2scaledprob(0.0f);
	//X[n][m-1] = prob2scaledprob(0.0f);
	//Y[n][m-1] =                            prob[c][           (6 << 9) | (seqb[m-2] << 6) | (6 << 3) | seqb[m-1]] + Y[n][m-2];
	g =           (EGAP << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1]; 
	tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[n][m-2] + hmm->bY[n][m-1]  - hmm->score );
	
	//Y[n][m-1] = logsum(Y[n][m-1] ,prob[c][(seqa[n] << 9) | (seqb[m-2] << 6) | (5 << 3) | seqb[m-1]] + M[n][m-2]);
	g = (seqa[n] << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1];
	tprob[g] = logsum(tprob[g], prob[g] + hmm->fM[n][m-2] + hmm->bY[n][m-1]  - hmm->score + frac);

	//Y[n][m-1] = logsum(Y[n][m-1] ,prob[c][(seqa[n] << 9) |               (5 << 6)  | (5 << 3) | seqb[m-1]] + X[n][m-2]);
	g = (seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[m-1];
	tprob[g] = logsum(tprob[g], prob[g] + hmm->fX[n][m-2] + hmm->bY[n][m-1]  - hmm->score + frac);
	
	//M[n][m] = prob2scaledprob(0.0f);
	//X[n][m] = prob2scaledprob(0.0f);
	//Y[n][m] = prob2scaledprob(0.0f);
	//Y[n][m] =                            prob[c][           (6 << 9) | (seqb[m-1] << 6) | (6 << 3) | seqb[m]] + Y[n][m-1];
	g =           (EGAP << 12) | (seqb[m-1] << 8) | (EGAP << 4) | seqb[m];
	tprob[g] = logsum(tprob[g], prob[g] + hmm->fY[n][m-1] + hmm->bY[n][m] - hmm->score );
	
	
	if(frame != -1){
		//M = hmm->fM;
		//X = hmm->fX;
		//Y = hmm->fY;
		
		
		hmm->fM = M;// hmm->tfM[frame];
		hmm->fX = X;//hmm->tfX[frame];
		hmm->fY = Y;//hmm->tfY[frame];
	}
	
	return hmm;
}

struct hmm2* glocal_forward_2ndorder_log_Y(struct hmm2* hmm,unsigned char* a, unsigned char* b, int n,int m,int frame )
{
	//init - terminal gap penalty = 0 (p = 1)
	int i,j,c;
	float** M = 0;
	//float** C = hmm->fC;
	float** X = 0;
	float** Y = 0;
	
	if(frame == -1){
		M = hmm->fM;
	//float** C = hmm->fC;
		X = hmm->fX;
		Y = hmm->fY;
	}else{
		M = hmm->tfM[frame];
		//float** C = hmm->fC;
		X = hmm->tfX[frame];
		Y = hmm->tfY[frame];
	}
	
	
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	float* prob = hmm->prob;	
	
	c = 0;
	M[0][0] = prob2scaledprob(0.0f);
	X[0][0] = prob2scaledprob(0.0f);
	Y[0][0] = prob2scaledprob(0.0f);
	
	M[0][1] = prob2scaledprob(0.0f);
	X[0][1] = prob2scaledprob(0.0f);
	Y[0][1] = prob2scaledprob(0.0f);
	
	M[0][2] = prob2scaledprob(0.0f);
	X[0][2] = prob2scaledprob(0.0f);
	Y[0][2] = prob[   (EGAP << 12) | (seqb[1] << 8) | (EGAP << 4) | seqb[2]];
	//fprintf(stderr,"%d start %d\n",i,c);
	for(j = 3; j <= m-2;j++){
		M[0][j] = prob2scaledprob(0.0f);
		X[0][j] = prob2scaledprob(0.0f);
		Y[0][j] = prob[   (EGAP << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j]] + Y[0][j-1];
		//Y[i][j] = logsum(Y[i][j], prob[(seqa[i] << 9) | (seqb[j-1] << 6) | (5 << 3) | seqb[j]] );
	}
	
	
	M[0][m-1] = prob2scaledprob(0.0f);
	X[0][m-1] = prob2scaledprob(0.0f);
	Y[0][m-1] = prob2scaledprob(0.0f);
	//Y[0][m-1] = prob[   (EGAP << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1]] + Y[0][m-2];
	
	
	M[0][m] = prob2scaledprob(0.0f);
	X[0][m] = prob2scaledprob(0.0f);
	Y[0][m] = prob2scaledprob(0.0f);
	//Y[0][m] = prob[   (EGAP << 12) | (seqb[m-1] << 8) | (EGAP << 4) | seqb[m]] + Y[0][m-1];
	
	i = 1;
	M[i][0] = prob2scaledprob(0.0f);
	X[i][0] = prob2scaledprob(0.0f);
	Y[i][0] = prob2scaledprob(0.0f);
	
	M[i][1] = prob2scaledprob(0.0f);
	X[i][1] = prob2scaledprob(0.0f);
	Y[i][1] = prob2scaledprob(0.0f);
	
	
	for(j = 2; j <= m-2;j++){
		//M[i][j] =                             prob[(seqa[i-1] << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + M[i-1][j-1]; 
		//M[i][j] = logsum(M[i][j], prob[(seqa[i-1] << 12) |               (IGAP << 8) | (seqa[i] << 4) | (seqb[j])] + X[i-1][j-1]);
		M[i][j] = prob[              (EGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[i-1][j-1];
		
		//X[i][j] =                            prob[(seqa[i-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP] + X[i-1][j] ;
		//X[i][j] = logsum(X[i][j], prob[(seqa[i-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + M[i-1][j]);
		X[i][j] = prob[              (EGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[i-1][j];
		
		Y[i][j] =                            prob[           (IGAP << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + Y[i][j-1];
		Y[i][j] = logsum(Y[i][j] ,prob[(seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + M[i][j-1]);
		Y[i][j] = logsum(Y[i][j] ,prob[(seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j]] + X[i][j-1]);
		
	}
	M[i][m-1] = prob2scaledprob(0.0f);
	X[i][m-1] = prob2scaledprob(0.0f);
	Y[i][m-1] = prob2scaledprob(0.0f);
	
	M[i][m] = prob2scaledprob(0.0f);
	X[i][m] = prob2scaledprob(0.0f);
	//X[i][m] = prob[              (IGAP << 12)  | (seqb[m] << 8) | (seqa[i] << 4) | IGAP] + Y[i-1][m];
	Y[i][m] = prob2scaledprob(0.0f);
	
	
	for(i = 2; i < n;i++){//= 4;i++){
		
		
		M[i][0] = prob2scaledprob(0.0f);
		X[i][0] = prob2scaledprob(0.0f);
		Y[i][0] = prob2scaledprob(0.0f);
		
		M[i][1] = prob2scaledprob(0.0f);
		X[i][1] = prob2scaledprob(0.0f);
		Y[i][1] = prob2scaledprob(0.0f);
		

		for(j = 2; j <= m-2;j++){
			M[i][j] =                             prob[(seqa[i-1] << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + M[i-1][j-1]; 
			M[i][j] = logsum(M[i][j], prob[(seqa[i-1] << 12) |               (IGAP << 8) | (seqa[i] << 4) | (seqb[j])] + X[i-1][j-1]);
			M[i][j] = logsum(M[i][j], prob[              (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[i-1][j-1]);
			
			X[i][j] =                            prob[(seqa[i-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP] + X[i-1][j] ;
			X[i][j] = logsum(X[i][j], prob[(seqa[i-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + M[i-1][j]);
			X[i][j] = logsum(X[i][j], prob[              (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[i-1][j]);
			
			Y[i][j] =                            prob[           (IGAP << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + Y[i][j-1];
			Y[i][j] = logsum(Y[i][j] ,prob[(seqa[i] << 12) | (seqb[j-1] << 8) | (IGAP << 4) | seqb[j]] + M[i][j-1]);
			Y[i][j] = logsum(Y[i][j] ,prob[(seqa[i] << 12) |               (IGAP << 8)  | (IGAP << 4) | seqb[j]] + X[i][j-1]);
			
		}
		M[i][m-1] = prob2scaledprob(0.0f);
		X[i][m-1] = prob2scaledprob(0.0f);
		Y[i][m-1] = prob2scaledprob(0.0f);
		
		M[i][m] = prob2scaledprob(0.0f);
		X[i][m] = prob2scaledprob(0.0f);
		//X[i][m] =                            prob[(seqa[i-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP] + X[i-1][m] ;
		Y[i][m] = prob2scaledprob(0.0f);
		
		//if(c > 5){
		//	c = 5;
		//}
		//c++;
	}
	
	
	M[n][0] = prob2scaledprob(0.0f);
	X[n][0] = prob2scaledprob(0.0f);
	Y[n][0] = prob2scaledprob(0.0f);
	
	M[n][1] = prob2scaledprob(0.0f);
	X[n][1] = prob2scaledprob(0.0f);
	Y[n][1] = prob2scaledprob(0.0f);
	
	for(j = 2; j <= m-2;j++){
		M[n][j] =                             prob[(seqa[n-1] << 12) | (seqb[j-1] << 8) | (seqa[n] << 4) | (seqb[j])] + M[n-1][j-1]; 
		M[n][j] = logsum(M[n][j], prob[(seqa[n-1] << 12) |               (IGAP << 8) | (seqa[n] << 4) | (seqb[j])] + X[n-1][j-1]);
		M[n][j] = logsum(M[n][j], prob[              (IGAP << 12) | (seqb[j-1] << 8) | (seqa[i] << 4) | (seqb[j])] + Y[n-1][j-1]);
		
		X[n][j] =                            prob[(seqa[n-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP] + X[n-1][j] ;
		X[n][j] = logsum(X[n][j], prob[(seqa[n-1] << 12) | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + M[n-1][j]);
		X[n][j] = logsum(X[n][j], prob[              (IGAP << 12)  | (seqb[j] << 8) | (seqa[i] << 4) | IGAP] + Y[n-1][j]);
		
		Y[n][j] =                            prob[           (EGAP << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j]] + Y[n][j-1];
		Y[n][j] = logsum(Y[n][j] ,prob[(seqa[n] << 12) | (seqb[j-1] << 8) | (EGAP << 4) | seqb[j]] + M[n][j-1]);
		Y[n][j] = logsum(Y[n][j] ,prob[(seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[j]] + X[n][j-1]);
		
	}
	M[n][m-1] = prob2scaledprob(0.0f);
	X[n][m-1] = prob2scaledprob(0.0f);
	Y[n][m-1] =                            prob[           (EGAP << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1]] + Y[n][m-2];
	Y[n][m-1] = logsum(Y[n][m-1] ,prob[(seqa[n] << 12) | (seqb[m-2] << 8) | (EGAP << 4) | seqb[m-1]] + M[n][m-2]);
	Y[n][m-1] = logsum(Y[n][m-1] ,prob[(seqa[n] << 12) |               (IGAP << 8)  | (EGAP << 4) | seqb[m-1]] + X[n][m-2]);
	
	
	M[n][m] = prob2scaledprob(0.0f);
	X[n][m] = prob2scaledprob(0.0f);
	//X[n][m] =                            prob[(seqa[n-1] << 12) |            (IGAP << 8) | (seqa[i] << 4) | IGAP] + X[n-1][m] ;
	//Y[n][m] = prob2scaledprob(0.0f);
	Y[n][m] =                            prob[           (EGAP << 12) | (seqb[m-1] << 8) | (EGAP << 4) | seqb[m]] + Y[n][m-1];
	//Y[n][m] = logsum(Y[n][m] ,prob[c][(seqa[n] << 9) | (seqb[m-1] << 6) | (6 << 3) | seqb[m]] + M[n][m-1]);
	//Y[n][m] = logsum(Y[n][m] ,prob[c][(seqa[n] << 9) |               (6 << 6)  | (6 << 3) | seqb[m]] + X[n][m-1]);
	hmm->score =Y[n][m];
	
	/*fprintf(stderr,"FM\n");
	for(j = 0; j <= m;j++){
		for(i = 0; i <= n;i++){
			
			fprintf(stderr,"%0.2f ",M[i][j]);
		}
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"\n");
	
	fprintf(stderr,"FX\n");
	
	for(j = 0; j <= m;j++){
		for(i = 0; i <= n;i++){
			fprintf(stderr,"%0.2f ",X[i][j]);
		}
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"\n");
	
	fprintf(stderr,"FY\n");
	for(j = 0; j <= m;j++){
		for(i = 0; i <= n;i++){
			
			fprintf(stderr,"%0.2f ",Y[i][j]);
		}
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"\n");*/
	
	//exit(0);
	//fprintf(stderr,"Forward: %f	%f	%f	%f	%f\n\n",hmm->score,  scaledprob2prob(hmm->score)  ,M[n][m],X[n][m],Y[n][m]);
	return hmm;
}

struct hmm2* glocal_backward_2ndorder_log_Y(struct hmm2* hmm,unsigned char* a, unsigned char* b, int n,int m)
{
	//init - terminal gap penalty = 0 (p = 1)
	int i,j;//c;
	
	
	float** M = hmm->bM;
	//float** C = hmm->fC;
	float** X = hmm->bX;
	float** Y = hmm->bY;
	float* prob = hmm->prob;
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	//float score;
	
	//c = 0;//n;
	//c = 5;
	//fprintf(stderr,"%d stop %d\n",n,c);
	
	M[n][m] = prob2scaledprob(0.0f);
	X[n][m] = prob2scaledprob(0.0f);
	Y[n][m] = prob2scaledprob(1.0f);
	
	
	M[n][m-1] = prob2scaledprob(0.0f);
	X[n][m-1] = prob2scaledprob(0.0f);
	Y[n][m-1] = prob[            (EGAP << 12) | (seqb[m-1] << 8) | (EGAP << 4) | seqb[m]] + Y[n][m];
	
	
	
	for(j = m-2; j > 1;j--){
		M[n][j] = prob[(seqa[n] << 12) | (seqb[j] << 8) | (EGAP << 4) | seqb[j+1]] + Y[n][j+1];
		X[n][j] =  prob[(seqa[n] << 12) |            (IGAP << 8) | (EGAP << 4) | seqb[j+1]] + Y[n][j+1];
		Y[n][j] =  prob[            (EGAP << 12) | (seqb[j] << 8) | (EGAP << 4) | seqb[j+1]] + Y[n][j+1];
	}
	
	M[n][1] =  prob2scaledprob(0.0f);
	X[n][1] =  prob2scaledprob(0.0f);
	Y[n][1] =  prob2scaledprob(0.0f);
	
	M[n][0] = prob2scaledprob(0.0f);
	X[n][0] = prob2scaledprob(0.0f);
	Y[n][0] = prob2scaledprob(0.0f);
	//was here ...... 
	
	for(i = n-1; i >= 1;i--){
		M[i][m] = prob2scaledprob(0.0f);
		X[i][m] = prob2scaledprob(0.0f);
		Y[i][m] = prob2scaledprob(0.0f);
		
		
		M[i][m-1] = prob2scaledprob(0.0f);
		X[i][m-1] = prob2scaledprob(0.0f);
		Y[i][m-1] = prob2scaledprob(0.0f);

		for(j = m-2; j > 1;j--){
			M[i][j] =                             prob[(seqa[i] << 12) | (seqb[j] << 8) | (seqa[i+1] << 4) | seqb[j+1]] + M[i+1][j+1];
			M[i][j] = logsum(M[i][j], prob[(seqa[i] << 12) | (seqb[j] << 8) |                 (IGAP << 4) | seqb[j+1]] + Y[i][j+1]);
			M[i][j] = logsum(M[i][j], prob[(seqa[i] << 12) | (seqb[j] << 8) | (seqa[i+1] << 4)  |                IGAP] + X[i+1][j]);
						
			X[i][j] =                             prob[(seqa[i] << 12) | (IGAP << 8) | (seqa[i+1] << 4) |               IGAP ] + X[i+1][j];
			X[i][j] =  logsum(X[i][j], prob[(seqa[i] << 12) | (IGAP << 8) | (seqa[i+1] << 4) | seqb[j+1]] + M[i+1][j+1]);
			X[i][j] =  logsum(X[i][j], prob[(seqa[i] << 12) | (IGAP << 8) |                (IGAP << 4) | seqb[j+1]] + Y[i][j+1]);
			
			Y[i][j] =                             prob[(IGAP << 12) | (seqb[j] << 8) |                (IGAP << 4) | seqb[j+1]] + Y[i][j+1];
			Y[i][j] = logsum(Y[i][j], prob[(IGAP << 12) | (seqb[j] << 8) | (seqa[i+1] << 4) | seqb[j+1]] + M[i+1][j+1]);
			Y[i][j] = logsum(Y[i][j], prob[(IGAP << 12) | (seqb[j] << 8) | (seqa[i+1] << 4) | IGAP  ] + X[i+1][j]);
			
			
			
		}
		
		M[i][1] =  prob2scaledprob(0.0f);
		X[i][1] =  prob2scaledprob(0.0f);
		Y[i][1] =  prob2scaledprob(0.0f);
		
		M[i][0] = prob2scaledprob(0.0f);
		X[i][0] = prob2scaledprob(0.0f);
		Y[i][0] = prob2scaledprob(0.0f);
		
	}
	
		
	
	M[0][m] = prob2scaledprob(0.0f);
	X[0][m] = prob2scaledprob(0.0f);
	Y[0][m] = prob2scaledprob(0.0f);
	
	
	M[0][m-1] = prob2scaledprob(0.0f);
	X[0][m-1] = prob2scaledprob(0.0f);
	Y[0][m-1] = prob2scaledprob(0.0f);
	
	
	
	for(j = m-2; j > 1;j--){
		//M[0][j] =                             prob[c][(seqa[0] << 9) | (seqb[j] << 6) | (seqa[1] << 3) | seqb[j+1]] + M[1][j+1];
		//M[0][j] = logsum(M[0][j], prob[c][     (seqa[0] << 9) | (seqb[j] << 6) |                 (5 << 3) | seqb[j+1]] + Y[0][j+1]);
		//M[0][j] = logsum(M[0][j], prob[c][(seqa[0] << 9) | (seqb[j] << 6) | (seqa[1] << 3)  | 5] + X[1][j]);
		
		//X[0][j] =                             prob[c][(seqa[0] << 9) | (5 << 6) | (seqa[1] << 3) | 5 ] + X[1][j];
		//X[0][j] =  logsum(X[0][j], prob[c][(seqa[0] << 9) | (5 << 6) | (seqa[1] << 3) | seqb[j+1]] + M[1][j+1]);
		//X[0][j] =  logsum(X[0][j], prob[c][     (seqa[0] << 9) | (5 << 6) |                (5 << 3) | seqb[j+1]] + Y[0][j+1]);
		M[0][j] =  prob2scaledprob(0.0f);
		X[0][j] = prob2scaledprob(0.0f);
		Y[0][j] =                             prob[ (EGAP << 12) | (seqb[j] << 8) |                (EGAP << 4) | seqb[j+1]] + Y[0][j+1];
		Y[0][j] = logsum(Y[0][j], prob[(EGAP << 12) | (seqb[j] << 8) | (seqa[1] << 4) | seqb[j+1]] + M[1][j+1]);
		Y[0][j] = logsum(Y[0][j], prob[(EGAP << 12) | (seqb[j] << 8) | (seqa[1] << 4) | IGAP  ] + X[1][j]);
		
		
	}
	
	M[0][1] =  prob2scaledprob(0.0f);
	X[0][1] =  prob2scaledprob(0.0f);
	Y[0][1] =                               prob[(EGAP << 12) | (seqb[1] << 8) |             (EGAP << 4) | seqb[2]] + Y[0][2];
	//Y[0][1] = logsum(Y[0][1], prob[c][(5 << 9) | (seqb[1] << 6) | (seqa[1] << 3) | seqb[2]] + M[1][2]);
	//Y[0][1] = logsum(Y[0][1], prob[c][(5 << 9) | (seqb[1] << 6) | (seqa[1] << 3) | 5  ] + X[1][1]);
	
	
	
	M[0][0] = prob2scaledprob(0.0f);
	X[0][0] = prob2scaledprob(0.0f);
	Y[0][0] = prob[(EGAP << 12) | (seqb[1] << 8) |             (EGAP << 4) | seqb[2]] + Y[0][1];
	
		//fprintf(stderr,"Y[n][m]	%f\n",prob[c][           (6 << 9) | (seqb[j-1] << 6) | (6 << 3) | seqb[j]] );
	//score =Y[0][1];
	//exit(0);
	
	/*fprintf(stderr,"BM\n");
	for(j = 0; j <= m;j++){
		for(i = 0; i <= n;i++){
			
			fprintf(stderr,"%0.2f ",M[i][j]);
		}
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"\n");
	
	fprintf(stderr,"BX\n");
	
	for(j = 0; j <= m;j++){
		for(i = 0; i <= n;i++){
			fprintf(stderr,"%0.2f ",X[i][j]);
		}
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"\n");
	
	fprintf(stderr,"BY\n");
	for(j = 0; j <= m;j++){
		for(i = 0; i <= n;i++){
			
			fprintf(stderr,"%0.2f ",Y[i][j]);
		}
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	fprintf(stderr,"\n");
	*/
	//fprintf(stderr,"F:%f	B:%f	diff:%f	significance%f\n", hmm->score,score, hmm->score -score,hmm->fY[n][m] - hmm->fX[n][m]);
	//
	//if((hmm->score - score) > 0.002f){
		//	fprintf(stderr,"backward: %f	diff:%f	%f	%f	%f	%f	\n\n",score, hmm->score -score, scaledprob2prob(hmm->score)  ,M[1][1],X[1][0],Y[0][1]);
		//	exit(0);
	//}
	//if((hmm->score - score) < -0.002f){
		//	fprintf(stderr,"backward: %f	diff:%f	%f	%f	%f	%f	\n\n",score, hmm->score -score, scaledprob2prob(hmm->score)  ,M[1][1],X[1][0],Y[0][1]);
		//	exit(0);
	//}
	return hmm;
}


struct hmm2* re_estimate_2ndorder_random(struct hmm2* hmm)
{
	int j;
	float* sum_prob =  0;
	
	sum_prob = malloc(sizeof(float)*256);
	
	int r1;
	
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			sum_prob[r1] = prob2scaledprob(0.0f);
		}
		sum_prob[r1] = logsum(sum_prob[r1], hmm->random_genome_model_t[j]);
		
	}
	for(j = 0; j < 5;j++){
		if(sum_prob[j] == prob2scaledprob(0.0f)){
			sum_prob[j] = prob2scaledprob(1.0f);
		}
	}
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			//sum_prob[r1] = prob2scaledprob(0.0f);
		}
		hmm->random_genome_model[j]  = hmm->random_genome_model_t[j]  - sum_prob[r1];		
	}
	
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			sum_prob[r1] = prob2scaledprob(0.0f);
		}
		sum_prob[r1] = logsum(sum_prob[r1], hmm->random_read_model_t[j]);
		
	}
	for(j = 0; j < 5;j++){
		if(sum_prob[j] == prob2scaledprob(0.0f)){
			sum_prob[j] = prob2scaledprob(1.0f);
		}
	}
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			//sum_prob[r1] = prob2scaledprob(0.0f);
		}
		hmm->random_read_model[j]  = hmm->random_read_model_t[j]  - sum_prob[r1];		
	}
	free(sum_prob);
	return hmm;
}

struct hmm2* re_estimate_2ndorder(struct hmm2* hmm)
{
	int i;//,j;
	float* sum_prob =  0;
	//float sum = 0.0f;
	sum_prob = malloc(sizeof(float)*256);
	
	int r1,g1;
	
	//for(i = 0;i < 1;i++){
	//	for(j = 0;j < 64;j++){
	//sum_prob = prob2scaledprob(0.000f);
	//	}
	
	for(i = 0; i < 256;i++){
		sum_prob[i] = prob2scaledprob(0.0f);
	}
	
	for(i = 0; i < 65536;i++){
		r1 = (i >> 4) & 15;
		g1= (i ) & 15;
		sum_prob[(r1 << 4) | g1] = logsum(sum_prob[(r1 << 4) | g1] , hmm->prob[i]);
		
		
	}
	
	for(i = 0; i < 65536;i++){
		r1 = (i >> 4) & 15;
		g1= (i ) & 15;
		//sum_prob[(r1 << 4) | g1] = logsum(sum_prob[(r1 << 4) | g1] , hmm->prob[i]);
		
		hmm->prob[i] = logsum(hmm->prob[i], sum_prob[(r1 << 4) | g1]);
	}
	
	for(i = 0; i < 256;i++){
		sum_prob[i] = prob2scaledprob(0.0f);
	}
	
	for(i = 0; i < 65536;i++){
		r1 = (i >> 12) & 15;
		g1= (i >> 8) & 15;
		
	//	if(r1 == EGAP && r2 == EGAP){
	//		hmm->tprob[i] = logsum(-1.0,hmm->tprob[i]);
	//	}
		hmm->tprob[i] = logsum(hmm->tprob[i], hmm->prob[i] +  prob2scaledprob(PSEUDO_CONSTANT));
		//r2 = (i >> 4) & 15;
		//g2 = i & 15; 
		//hmm->prob[i] = logsum(-10,hmm->tprob[i]);
		hmm->prob[i] = hmm->tprob[i];
		hmm->tprob[i] =  prob2scaledprob(0.0f);
		/*if(r1 == 6 || r2 == 6){
		 sum_prob[1] = logsum(sum_prob[1], hmm->prob[i][j]);
		 }else{*/
		sum_prob[(r1 << 4) | g1] = logsum(sum_prob[(r1 << 4) | g1] , hmm->prob[i]);
		/*}*/
		//sum_prob[(r1 <<3) | g1] = logsum(sum_prob[(r1 <<3) | g1], hmm->prob[i][j]);
	}
	for(i = 0; i < 256;i++){
		if(sum_prob[i] == prob2scaledprob(0.0f)){
			sum_prob[i] = prob2scaledprob(1.0f);
		}
	}
	for(i = 0; i < 65536;i++){
		r1 = (i >> 12) & 15;
		g1= (i >> 8) & 15;
		//r2 = (j >> 4) & 15;
		//g2 = j & 15; 
		
		/*if(r1 == 6 || r2 == 6){
		 hmm->prob[i][j]  = hmm->prob[i][j]  - sum_prob[1];
		 
		 //sum_prob[1] = logsum(sum_prob[1], hmm->prob[i][j]);
		 }else{*/
		hmm->prob[i]  = hmm->prob[i]  - sum_prob[(r1 << 4) | g1];//,prob2scaledprob(1.0));
		
		//sum_prob[0] = logsum(sum_prob[0], hmm->prob[i][j]);
		/*}*/
		
		//hmm->prob[i][j]  = hmm->prob[i][j]  - sum_prob[(r1 <<3) | g1];
	}
	
	/*float sanity = 0.0;
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			sum_prob[r1] = prob2scaledprob(0.0f);
		}
		sum_prob[r1] = logsum(sum_prob[r1], hmm->random_genome_model_t[j]);
		
	}
	for(j = 0; j < 5;j++){
		if(sum_prob[j] == prob2scaledprob(0.0f)){
			sum_prob[j] = prob2scaledprob(1.0f);
		}
	}
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			//sum_prob[r1] = prob2scaledprob(0.0f);
		}
		hmm->random_genome_model[j]  = hmm->random_genome_model_t[j]  - sum_prob[r1];		
	}
	
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			sum_prob[r1] = prob2scaledprob(0.0f);
		}
		sum_prob[r1] = logsum(sum_prob[r1], hmm->random_read_model_t[j]);
		
	}
	for(j = 0; j < 5;j++){
		if(sum_prob[j] == prob2scaledprob(0.0f)){
			sum_prob[j] = prob2scaledprob(1.0f);
		}
	}
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
			//sum_prob[r1] = prob2scaledprob(0.0f);
		}
		hmm->random_read_model[j]  = hmm->random_read_model_t[j]  - sum_prob[r1];		
	}
	
	*/
	/*for(i = 0; i < 31;i++){
		r1 = -1;
		
		sum = prob2scaledprob(0.0f);
		sanity = 0.0;
		
		for(j = 0; j < 25;j++){
			hmm->genome_model_t[i][j] = logsum(hmm->genome_model_t[i][j],  hmm->genome_model[i][j] +  prob2scaledprob(PSEUDO_CONSTANT));
		}
		
		for(j = 0; j < 25;j++){
			if(j % 5 == 0){
				r1++;
				sum_prob[r1] = prob2scaledprob(0.0f);
			}
			sum_prob[r1] = logsum(sum_prob[r1], hmm->genome_model_t[i][j]);
			sum = logsum(sum, hmm->genome_model_t[i][j]);
		}
		for(j = 0; j < 256;j++){
			if(sum_prob[j] == prob2scaledprob(0.0f)){
				sum_prob[j] = prob2scaledprob(1.0f);
			}
		}
		//for(j = 0; j < 25;j++){
		//	hmm->genome_model_t[i][j]  = logsum(prob2scaledprob(0.1) ,hmm->genome_model_t[i][j]  - sum);
		//}
		sanity = 0.0;
		//for(j = 0; j < 25;j++){
		//	sum = logsum(sum, hmm->genome_model_t[i][j]);
		//}
		//fprintf(stderr,"%d	SUM:%f\n",i,sum);
		r1 = -1;
		for(j = 0; j < 25;j++){
			if(j % 5 == 0){
				r1++;
				//sum_prob[r1] = prob2scaledprob(0.0f);
			}
			hmm->genome_model[i][j]  = hmm->genome_model_t[i][j]  - sum_prob[r1];
			//hmm->genome_model[i][j]  = hmm->genome_model_t[i][j]  - sum;
		//	fprintf(stderr,"%0.2f ",scaledprob2prob( hmm->genome_model[i][j])*100.0f);
			sanity += scaledprob2prob( hmm->genome_model[i][j]);
			hmm->genome_model_t[i][j] = prob2scaledprob(0.0f);
			
		}
		//for(j = 0; j < 25;j++){
		//	fprintf(stderr,"%0.1f ",scaledprob2prob(hmm->genome_model[i][j]) * 100.0f);
		//}
		
		
		//fprintf(stderr," %f\n",sanity);
	}*/
	
	//}
	free(sum_prob);
	return hmm;
}

/*struct hmm2* score_genome(struct hmm2* hmm, unsigned char* seq,int len)
{
	int i,c,code;
	hmm->genome_score = prob2scaledprob(1.0);
	c = 0;
	for(i = 0; i < len-1;i++){
		
		if(i > 30){
			c = 30;
		}
		code = (seq[i] & 0x3) *5 + (seq[i+1] & 0x3);
		
		//hmm->genome_model_t[c][code] =  logsum(hmm->genome_model_t[c][code], hmm->genome_model[c][code]);
		hmm->genome_score += hmm->genome_model[c][code];
		c++;
	}
	
	//hmm->genome_score = logsum(prob2scaledprob(0.5),  hmm->genome_score - prob2scaledprob(0.5));
	
	return hmm;
}


struct hmm2* get_genome_prob(struct hmm2* hmm, unsigned char* seq,int len, float frac)
{
	int i,c,code;
	c = 0;
	for(i = 0; i < len-1;i++){
		
		if(i > 30){
			c = 30;
		}
		code = (seq[i] & 0x3) *5 + (seq[i+1] & 0x3);
		//fprintf(stderr,"%d	%f:	%f	%f	%f\n",i, hmm->genome_model[c][code] - hmm->genome_score + frac, hmm->genome_model[c][code],hmm->genome_score , frac);
		hmm->genome_model_t[c][code] =  logsum(hmm->genome_model_t[c][code],  frac);
		//fprintf(stderr,"%d %f	%f	%f	:%f\n",i,f[i-1]+ hmm->genome_model[c][code] + b[i+1],hmm->genome_score ,frac,f[i]+ hmm->genome_model[c][code] + b[i+1] - hmm->genome_score + frac);
		//hmm->genome_score += hmm->genome_model[c][code];
		c++;
	}
	//free(f);
	//free(b);
	//hmm->genome_score = logsum(prob2scaledprob(0.7),  hmm->genome_score - prob2scaledprob(0.3));
	
	return hmm;
}*/

struct hmm2* random_model_calc(struct hmm2* hmm, unsigned char* rseq, unsigned char* gseq,int rlen, int glen,float frac,int mode)
{
	int i,code;
	hmm->random_score = prob2scaledprob(1.0);
	if(mode == 0){
		for(i = 0; i < glen-1;i++){
			if(gseq[i] == NCODE){
				code = 20;
				
			}else{
				code = (gseq[i] & 0x3) *5;
			}
			if(gseq[i+1] == NCODE){
				code += 4;
			}else{
				code+= (gseq[i+1] & 0x3);
			}
			hmm->random_score += hmm->random_genome_model[code];
			hmm->random_genome_model_t[code] =  logsum(hmm->random_genome_model_t[code],  frac);
		}
	}
	if(mode == 1){
		for(i = 0; i < rlen-1;i++){
			if(rseq[i] == NCODE){
				code = 20;
			
			}else{
				code = (rseq[i] & 0x3) *5;
			}
			if(rseq[i+1] == NCODE){
				code += 4;
			}else{
				code+= (rseq[i+1] & 0x3);
			}
			hmm->random_score += hmm->random_read_model[code];
			hmm->random_read_model_t[code] =  logsum(hmm->random_read_model_t[code], frac );
		}
	}
	
	return hmm;
}


struct hmm2* init_2nd_hmm(int x, int y)
{
	struct hmm2* hmm = 0;
	int i,j,c;
	int r1,r2,g1,g2;
	//int gaps = 0;
	float* sum_prob = 0;
	//float sum;
	x = x + 2;
	y = y + 2;
	hmm = malloc(sizeof(struct hmm));
	hmm->fM = malloc(sizeof(float*)*x);
	hmm->fX = malloc(sizeof(float*)*x);
	hmm->fY = malloc(sizeof(float*)*x);
	
	
	hmm->tfM = malloc(sizeof(float**)*LIST_STORE_SIZE);
	hmm->tfX = malloc(sizeof(float**)*LIST_STORE_SIZE);
	hmm->tfY = malloc(sizeof(float**)*LIST_STORE_SIZE);
	
	hmm->bM = malloc(sizeof(float*)*x);
	hmm->bX = malloc(sizeof(float*)*x);
	hmm->bY = malloc(sizeof(float*)*x);
	
	hmm->train_count = 0;
	hmm->traceback = malloc(sizeof(unsigned short int*)*x);
	
	hmm->prob = malloc(sizeof(float) * 65536);
	hmm->tprob = malloc(sizeof(float) * 65536);
	
	hmm->x = x;
	hmm->y = y;
	
	for(i = 0; i < LIST_STORE_SIZE;i++){
		hmm->tfM[i] = malloc(sizeof(float*)*x);
		hmm->tfX[i] = malloc(sizeof(float*)*x);
		hmm->tfY[i] = malloc(sizeof(float*)*x);
		for(j = 0;j < x;j++){
			hmm->tfM[i][j] = malloc(sizeof(float*)*y);
			hmm->tfX[i][j] = malloc(sizeof(float*)*y);
			hmm->tfY[i][j] = malloc(sizeof(float*)*y);
			
			for(c = 0; c < y;c++){
				hmm->tfM[i][j][c] = 0;
				hmm->tfX[i][j][c] = 0;
				hmm->tfY[i][j][c] = 0;
			}
		}
	}
	
	for(i = 0;i < x;i++){
		hmm->fM[i] = malloc(sizeof(float*)*y);
		hmm->fX[i] = malloc(sizeof(float*)*y);
		hmm->fY[i] = malloc(sizeof(float*)*y);
		
		hmm->bM[i] = malloc(sizeof(float*)*y);
		hmm->bX[i] = malloc(sizeof(float*)*y);
		hmm->bY[i] = malloc(sizeof(float*)*y);
		
		hmm->traceback[i] = malloc(sizeof(unsigned short int*)*y);
		for(j = 0; j < y;j++){
			hmm->fM[i][j] = 0;
			hmm->fX[i][j] = 0;
			hmm->fY[i][j] = 0;
			
			hmm->bM[i][j] = 0;
			hmm->bX[i][j] = 0;
			hmm->bY[i][j] = 0;
			hmm->traceback[i][j] = 0;
		}
	}
	
	sum_prob = malloc(sizeof(float) * 256 );
	
	
	hmm->random_genome_model = malloc(sizeof(float)* 25);
	hmm->random_read_model = malloc(sizeof(float)* 25);
	hmm->random_genome_model_t = malloc(sizeof(float)* 25); 
	hmm->random_read_model_t= malloc(sizeof(float)* 25);
	
	
	
	for(i = 0; i < 256;i++){
		sum_prob[i] = 0;
	}
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
		}
		hmm->random_genome_model[j] = prob2scaledprob(0.00001f);
		hmm->random_read_model[j] = prob2scaledprob(0.00001f);
		hmm->random_genome_model_t[j] = prob2scaledprob(0.0f);
		hmm->random_read_model_t[j] = prob2scaledprob(0.0f);
		
		sum_prob[r1] = logsum(sum_prob[r1], hmm->random_genome_model[j]);
	}
	
	r1 = -1;
	for(j = 0; j < 25;j++){
		if(j % 5 == 0){
			r1++;
		}
		hmm->random_genome_model[j]  = hmm->random_genome_model[j]  - sum_prob[r1] ;
		hmm->random_read_model[j]  = hmm->random_read_model[j]  - sum_prob[r1] ;
		//hmm->genome_model[i][j]  = hmm->genome_model[i][j]  - sum;
		//fprintf(stderr,"%f	%d	%d\n",genome_model[i][j],i,j);
	}
	/*for(i = 0; i < 256;i++){
		sum_prob[i] = 0;
	}
	hmm->genome_model = malloc(sizeof(float*)*31);
	hmm->genome_model_t = malloc(sizeof(float*)*31);
	for(i = 0; i < 31;i++){
		hmm->genome_model[i] = malloc(sizeof(float)* 25);
		hmm->genome_model_t[i] = malloc(sizeof(float)* 25);
		sum = prob2scaledprob(0.0f);	
		r1 = -1;
		for(j = 0; j < 25;j++){
			if(j % 5 == 0){
				r1++;
			}
			hmm->genome_model[i][j] = prob2scaledprob(0.00001f);
			hmm->genome_model_t[i][j] = prob2scaledprob(0.0f);
			sum = logsum(sum, hmm->genome_model[i][j]);
			sum_prob[r1] = logsum(sum_prob[r1], hmm->genome_model[i][j]);
		}
		r1 = -1;
		for(j = 0; j < 25;j++){
			if(j % 5 == 0){
				r1++;
			}
			hmm->genome_model[i][j]  = hmm->genome_model[i][j]  - sum_prob[r1] ;
			//hmm->genome_model[i][j]  = hmm->genome_model[i][j]  - sum;
			//fprintf(stderr,"%f	%d	%d\n",genome_model[i][j],i,j);
		}
	}
	
	*/
	
	
	for(i = 0; i < 256;i++){
		sum_prob[i] = 0;
	}
	for(i = 0; i < 65536;i++){
		r1 = (i >> 12) & 15;
		g1= (i >> 8) & 15;
		r2 = (i >> 4) & 15;
		g2 = i &15; 
		hmm->prob[i] = 1.0f;
		
		
		/*if(r1 == EGAP &&  g1 <= NCODE){
			hmm->prob[i] = 1.0f;
		}else if(r1 == IGAP &&  g1 <= NCODE) {
			hmm->prob[i] = 0.01f;
		}else if(r1 <= NCODE &&  g1 == IGAP) {
			hmm->prob[i] = 0.01f;
		}else if(r1 <= NCODE && g1 <= NCODE) {
			if(r1 == NCODE || g1 == NCODE){
				hmm->prob[i] = 0.01f;
			}else if((r1 & 0x3) == (g1 & 0x3) ){
				hmm->prob[i] = 0.98f;
			}else{
				hmm->prob[i] = 0.01f;
			}
			
		}

		if(r2 == EGAP &&  g2 <= NCODE){
			hmm->prob[i] *= 1.0f;
		}else if(r2 == IGAP &&  g2 <= NCODE) {
			hmm->prob[i] *= 0.01f;
		}else if(r2 <= NCODE &&  g2 == IGAP) {
			hmm->prob[i] *= 0.01f;
		}else if(r2 <= NCODE && g2 <= NCODE) {
			if(r2 == NCODE || g2 == NCODE){
				hmm->prob[i] *= 0.01f;
			}else if((r2 & 0x3) == (g2 & 0x3) ){
				hmm->prob[i] *= 0.98f;
			}else{
				hmm->prob[i] *= 0.01f;
			}
			
		}*/
		
		
		//if(r1 == EGAP && r2 == EGAP){
		//	hmm->prob[i] += 1.0f;
		//}
		/*else{
			if(r2 == NCODE || g2 == NCODE){
				hmm->prob[i] = 4.24E-03;
			}else if(r2 > NCODE && g2 > NCODE){
				hmm->prob[i] = 0.0;
			}else if(r2 > NCODE || g2 > NCODE){
				hmm->prob[i] = 4.24E-03;
			}else if(r2 == g2){
				hmm->prob[i] = 2.12E-01;
			}
		}*/
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		if(r1 < 4){
			if(r2 == IGAP){
				hmm->prob[i] = 0.0f;
			}
			if(g2 == IGAP){
				hmm->prob[i] = 0.0f;
			}
		}
		
		if(r1 >= 8 && r1 < 12){
			if(r2 == IGAP){
				hmm->prob[i] = 0.0f;
			}
			if(g2 == IGAP){
				hmm->prob[i] = 0.0f;
			}
		}
		
		
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		

		
		/*if((r2 < 12) && (g2 < 12)){
			if((r2 & 0x3)  == (g2 & 0x3)){
				hmm->prob[i] += 9.0f;
			}//else{
			//	hmm->prob[i] += 9.0f;
			//}
		}*/
		sum_prob[(r1  << 4) | g1 ]  += hmm->prob[i];
		/*if(r2 <= EGAP && g2 <= IGAP){
			if(r2 > NCODE && g2 > NCODE){
				hmm->prob[i] = 0.0f;
			}else{
				
				if(r2 < NCODE && g2 < NCODE){
					if((r2 & 3 ) == (g2 & 3  )){
						hmm->prob[i] = 1.0f;
					}else{
						hmm->prob[i] = 0.02f;
					}
				}else if(r2 == NCODE || g2 == NCODE){
					hmm->prob[i] = 0.02f;
				}else{
					if(r2 == EGAP){
						hmm->prob[i] = 1.0f; 
					}else{
						hmm->prob[i] = 0.01f;
					}
				}
			}
			sum_prob[(r1  << 4) | g1 ]  += hmm->prob[i];
		}	*/	
		
		
		/*if(r1>= 15 || g1 >= 15  || r2 >= 15 || g2 >= 15){
			hmm->prob[i] = 0.0f;
		}else{
			//gaps = 0;
		
			if(r2 <= NCODE && g2 <= NCODE){
				if((r2 == NCODE) ||  (g2 == NCODE)){
					hmm->prob[i] =  0.02f;
				}else{
					if((r2 & 3 ) == (g2 & 3  )){
						hmm->prob[i] = 1.0f;
					}else{
						hmm->prob[i] =  0.02f;
					}
				}
				//if((r2 & 3 ) == (g2 & 3  )){
				//	hmm->prob[i] = 1.0f;
					//if((r1 & 0x3) == (g1 & 0x3)){
					//	hmm->prob[i] += 1.0f;
					//}
					
				//}else{
				//	hmm->prob[i] =  0.02f;	
				//}
			}else if(r1 == EGAP && r2 == EGAP  ){
				hmm->prob[i] = 1.0f;
			}else if(r2 > NCODE && g2 > NCODE){ // gaps in both read and genome of any type...
				hmm->prob[i] = 0.0f;
			}else if(r2 == IGAP && g2 < 13 ){
				
				//if(r1 < 4){
				//	hmm->prob[i] = 0.01f;
				//}else if(r1 < 8){
				//	hmm->prob[i] = 0.01f;
				//}else{
				//	hmm->prob[i] = 0.005f;
				//}
				
				hmm->prob[i] = 0.01f;
				//if(r1 == IGAP && g2 < 13 ){
				//	hmm->prob[i] += 1.0f;
				//}
			}else if(r2 == EGAP && g2 < 13){
				
				if(r1 < 4){
					hmm->prob[i] = 0.0f;
				}else if(r1 < 8){
					hmm->prob[i] = 0.0f;
				}else{
					hmm->prob[i] = 0.005f;
				}
				
				hmm->prob[i] = 0.01f;
				//if(r1 == EGAP && g1 < 13){
				//	hmm->prob[i] += 1.0f;
				//}
			}else if(r2 < 13 && g2 == IGAP ){
				hmm->prob[i] = 0.01f;
				//if(r1 < 13 && g1 == IGAP ){
				//	 hmm->prob[i] += 1.0f;
				//}
			
			}else if(g2 == EGAP ){
				hmm->prob[i] = 0.0f;
			}
			
			if(r1 < 4){
				if(r2 == IGAP){
					hmm->prob[i] = 0.0f;
				}
				if(g2 == IGAP){
					hmm->prob[i] = 0.0f;
				}
			}
			
			if(r1 >= 8 && r1 < 12){
				if(r2 == IGAP){
					hmm->prob[i] = 0.0f;
				}
				if(g2 == IGAP){
					hmm->prob[i] = 0.0f;
				}
			}
			
			sum_prob[(r1  << 4) | g1 ]  += hmm->prob[i];
		}*/
	}
	
	for(i = 0; i < 65536;i++){
		r1 = (i >> 12) & 15;
		g1= (i >> 8) & 15;
		r2 = (i >> 4) & 15;
		g2 = i & 15; 
		
		//if(r2 <= EGAP && g2 <= IGAP){
		//	hmm->prob[i] = prob2scaledprob( hmm->prob[i] /  sum_prob[(r1 << 4) | g1 ] );
		//}
		//hmm->prob[i] = prob2scaledprob( hmm->prob[i] );
		if(r1>= 15 || g1 >= 15  || r2 >= 15 || g2 >= 15){
			//prob[i] = sum_prob;
			//fprintf(stderr,"%d	%d	%d	%d	%f	%f	%f\n",r1,g1,r2,g2,prob[j][i], prob2scaledprob(1.0f) ,sum_prob);
			//prob[i] = prob2scaledprob( prob[i] / (sum_prob[(r1 << 3) | g1] )  );
			//prob[i] = prob[i] / (sum_prob[(r1 << 3) | g1] ) ;
			//for(j = 0;j < 11;j++){
			hmm->prob[i] = prob2scaledprob(0.0f);
			//}
		}else{
			//prob[i] = prob[i] / (sum_prob[(r1<<3) | g1] );
			//fprintf(stderr,"%d	%d	%d	%d	%f	%f	%f\n",r1,g1,r2,g2,prob[j][i], prob2scaledprob(prob[j][i] / sum_prob ) ,sum_prob );
			hmm->prob[i] = prob2scaledprob( hmm->prob[i] /  sum_prob[(r1 << 4) | g1 ] );
			
			//for(j = 1;j < 11;j++){
			//prob[j][i] = prob[0][i];
			//}
			//prob[i] = prob[i] / (sum_prob[(r1 << 3) | g1] ) ;
		}
		//if(r1 == 0 && g1 == 0){
		//	fprintf(stderr,"%d	%d	%d	%d	%f	%f	%f\n",r1,g1,r2,g2,hmm->prob[i], prob2scaledprob(hmm->prob[i]) ,sum_prob);
		//}
		//hmm->tprob[i] = prob2scaledprob(0.0f);
	}
	//exit(0);
	free(sum_prob);
	return hmm;
}




void init_thread_hmm2s(struct thread_data2* thread_data_array,struct hmm2* localhmm,int numthreads)
{
	int i,j,c;
	for(c = 0; c < numthreads;c++){
		thread_data_array[c].hmm->train_count = 0;
		for(i = 0; i < 65536;i++){
			thread_data_array[c].hmm->prob[i] = localhmm->prob[i];
			thread_data_array[c].hmm->tprob[i] = prob2scaledprob(0.0f);
		}
		for(j = 0; j < 25;j++){
			thread_data_array[c].hmm->random_genome_model[j] = localhmm->random_genome_model[j];
			thread_data_array[c].hmm->random_genome_model_t[j] = prob2scaledprob(0.0f);
			thread_data_array[c].hmm->random_read_model[j] = localhmm->random_read_model[j];
			thread_data_array[c].hmm->random_read_model_t[j] = prob2scaledprob(0.0f);

		}
		/*for(i = 0; i < 31;i++){
			for(j = 0; j < 25;j++){
				thread_data_array[c].hmm->genome_model[i][j] = localhmm->genome_model[i][j];
				thread_data_array[c].hmm->genome_model_t[i][j] = prob2scaledprob(0.0f);
			}
		}*/
	}
		
}

void entangle_hmm2s(struct thread_data2* thread_data_array,struct hmm2* localhmm,int numthreads)
{
	int i,j,c;
	
	for(i = 0; i < 65536;i++){
		localhmm->tprob[i] = thread_data_array[0].hmm->tprob[i];
	}
	
	/*for(i = 0; i < 31;i++){
		for(j = 0; j < 25;j++){
			localhmm->genome_model_t[i][j] = thread_data_array[0].hmm->genome_model_t[i][j];
			//thread_data_array[c].hmm->genome_model[i][j] = localhmm->genome_model[i][j];
			//thread_data_array[c].hmm->genome_model_t[i][j] = prob2scaledprob(0.0f);
		}
	}*/
	for(j = 0; j < 25;j++){
		localhmm->random_genome_model_t[j] = logsum(localhmm->random_genome_model_t[j] ,thread_data_array[0].hmm->random_genome_model_t[j]);
		localhmm->random_read_model_t[j] = logsum(localhmm->random_read_model_t[j] ,thread_data_array[0].hmm->random_read_model_t[j]);
	}
	
	
	for(c = 1; c < numthreads;c++){
		localhmm->train_count += thread_data_array[c].hmm->train_count;
		for(i = 0; i < 65536;i++){
			localhmm->tprob[i] = logsum(localhmm->tprob[i], thread_data_array[c].hmm->tprob[i]);
		}
		/*for(i = 0; i < 31;i++){
			for(j = 0; j < 25;j++){
				localhmm->genome_model_t[i][j] = logsum(localhmm->genome_model_t[i][j] ,thread_data_array[c].hmm->genome_model_t[i][j]);
			}
		}*/
		for(j = 0; j < 25;j++){
			localhmm->random_genome_model_t[j] = logsum(localhmm->random_genome_model_t[j] ,thread_data_array[c].hmm->random_genome_model_t[j]);
			localhmm->random_read_model_t[j] = logsum(localhmm->random_read_model_t[j] ,thread_data_array[c].hmm->random_read_model_t[j]);
		}
	}
}


void write_rescue_probabilities(struct rescue_probabilities* rp, unsigned int len,unsigned int block_len,struct parameters * param)
{
	FILE* hmmfile;
	char* outfile;
	int i,j;
	j = (int)strlen(param->hmm_file);
	outfile = malloc(sizeof(char)*(j+4));
	for(i = 0; i < j ;i++){
		outfile[i] = param->hmm_file[i];
	}
	outfile[j] = '.';
	outfile[j+1] = 'g';
	outfile[j+2] = 'p';
	outfile[j+3] = 0;
	if ((hmmfile = fopen(outfile, "wb")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	fwrite(rp->plus,sizeof(float),(len >> block_len) + 1u,hmmfile );
	fwrite(rp->minus,sizeof(float),(len >> block_len) + 1u,hmmfile );
	//fwrite(rp->minus = malloc(sizeof(float)  *((len >> block_len) + 1u));

	fclose(hmmfile);
}

void read_rescue_probabilities(struct rescue_probabilities* rp, unsigned int len,unsigned int block_len,struct parameters * param)
{
	FILE* hmmfile;
	char* outfile;
	int i,j;
	j = (int)strlen(param->hmm_file);
	outfile = malloc(sizeof(char)*(j+4));
	for(i = 0; i < j ;i++){
		outfile[i] = param->hmm_file[i];
	}
	outfile[j] = '.';
	outfile[j+1] = 'g';
	outfile[j+2] = 'p';
	outfile[j+3] = 0;
	if ((hmmfile = fopen(outfile, "rb")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	fread(rp->plus,sizeof(float),(len >> block_len) + 1u,hmmfile );
	fread(rp->minus,sizeof(float),(len >> block_len) + 1u,hmmfile );
	fclose(hmmfile);
}

void write_hmm2_to_file(struct hmm2* hmm,struct parameters * param)
{
	FILE* hmmfile; 
	if ((hmmfile = fopen(param->hmm_file, "wb")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	fwrite(&param->local_weighting, sizeof(int),1,hmmfile);
	//fwrite(&param->random_model_prior, sizeof(float),1,hmmfile);
	
	fwrite(hmm->prob,sizeof(float),65536,hmmfile);
	fwrite(hmm->random_genome_model,sizeof(float),25,hmmfile);
	fwrite(hmm->random_read_model ,sizeof(float),25,hmmfile);
	
	fclose(hmmfile);
}

void read_hmm2_from_file(struct hmm2* hmm,struct parameters * param)
{
	FILE* hmmfile; 
	//fread(seed_parameters , sizeof(int), 4, file);
	/*file = fopen(param->infile[1], "rb");*/
	if ((hmmfile = fopen(param->hmm_file, "rb")) == NULL){
		fprintf(stderr,"can't open hmm file:%s \n",param->hmm_file);
		exit(-1);
	}
	fread(&param->local_weighting, sizeof(int),1,hmmfile);
	//fread(&param->random_model_prior, sizeof(float),1,hmmfile);
	fread(hmm->prob, sizeof(float),65536,hmmfile);
	fread(hmm->random_genome_model,sizeof(float),25,hmmfile);
	fread(hmm->random_read_model ,sizeof(float),25,hmmfile);
	fclose(hmmfile);
}

unsigned char* transform_genome_sequence(unsigned char* seq,unsigned int len)
{
	int i; 
	for(i = 0;i < len;i++){
		if(seq[i] == 4){
			seq[i] = NCODE;
		}/*else if(i  < 10){
			seq[i] += 8;
		}else if(i >= len-10){
			seq[i] += 4;
		}*/
	}
	
	
	return seq;
}


unsigned char* convert_16_to_5_nuc(unsigned char*  seq, int len)
{
	int j;
	for(j = 0;j < len;j++){
		if(seq[j] == NCODE){
			seq[j] = 4;
		}else {
			seq[j]  = seq[j]  & 3;
		}
	}
	
	
	return seq;
}


unsigned char* reverse_complement_16(unsigned char* p,int len)
{
	unsigned char* tmp = malloc(sizeof(unsigned char)*MAX_LINE);
	int i,c;
	c = 0;
	
	int rev_nuc_code16[16];
	rev_nuc_code16[0] = 3;//A Adenine
	rev_nuc_code16[1] = 2;//C	Cytosine
	rev_nuc_code16[2] = 1;//G	Guanine
	rev_nuc_code16[3] = 0;//T	Thymine
	rev_nuc_code16[4] = 7;//A Adenine
	rev_nuc_code16[5] = 6;//C	Cytosine
	rev_nuc_code16[6] = 5;//G	Guanine
	rev_nuc_code16[7] = 4;//T	Thymine
	rev_nuc_code16[8] = 11;//A Adenine
	rev_nuc_code16[9] = 10;//C	Cytosine
	rev_nuc_code16[10] = 9;//G	Guanine
	rev_nuc_code16[11] = 8;//T	Thymine
	rev_nuc_code16[12] = 12;//A Adenine
	rev_nuc_code16[13] = 13;//C	Cytosine
	rev_nuc_code16[14] = 14;//G	Guanine
	rev_nuc_code16[15] = 15;//T	Thymine
	
	
	for(i =len-1; i >= 0;i--){
		tmp[c] = rev_nuc_code16[(int)p[i]];
		c++;
	}
	tmp[c] = 0;
	for(i= 0; i < len;i++){
		p[i] = tmp[i];
	}
	free(tmp);
	return p;
}



void free_2nd_hmm(struct hmm2* hmm)
{
	int i,j;
	int x;//,y; 
	
	x = hmm->x;
	//y = hmm->y;
	
	for(i = 0; i < LIST_STORE_SIZE;i++){
		
		for(j = 0;j < x;j++){
			free(hmm->tfM[i][j]);// = malloc(sizeof(float*)*y);
			free(hmm->tfX[i][j]);// = malloc(sizeof(float*)*y);
			free(hmm->tfY[i][j]);// = malloc(sizeof(float*)*y);
			
			
		}
		free(hmm->tfM[i]);
		free(hmm->tfX[i]);
		free(hmm->tfY[i]);
	}
	free(hmm->tfM);
	free(hmm->tfX);
	free(hmm->tfY);
	
	for(i = 0;i < x;i++){
		free(hmm->traceback[i]);
		free(hmm->fM[i]);
		free(hmm->fX[i]);
		free(hmm->fY[i]);
		
		free(hmm->bM[i]);
		free(hmm->bX[i]);
		free(hmm->bY[i]);
	}
	
	
	/*for(i = 0; i < 31;i++){
		free(hmm->genome_model[i]);
		free(hmm->genome_model_t[i]);
	}
	free(hmm->genome_model );
	free(hmm->genome_model_t);*/
	
	free(hmm->random_genome_model);
	free(hmm->random_read_model);
	free(hmm->random_genome_model_t);
	free(hmm->random_read_model_t);
	
	free(hmm->fM);
	free(hmm->fX);
	free(hmm->fY);
	
	free(hmm->bM);
	free(hmm->bX);
	free(hmm->bY);
	free(hmm->prob);
	free(hmm->tprob);
	
	free(hmm->traceback);
	
	free(hmm);
}


