/*
 
 Copyright (C) 2010,2011 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>

#define OPT_VITERBI 1
#define OPT_ASSEMBLY 2

#define OPT_READLEN 3
#define OPT_AT_START_ERROR 4
#define OPT_AT_END_ERROR 5
#define OPT_INDEL 6
#define OPT_WINDOW_START 7

#define OPT_WIGGLE 8


#define OPT_LOCAL_WEIGHT 9
#define OPT_TRAINING_SEQUENCES 10
#define OPT_RANP 11
#define OPT_COUNT_MISMATCHES 12
#define OPT_FDR 13

struct parameters {
	FILE* foutfile;
	char** infile;
	char* outfile;
	char* hmm_file;
	char* command_line;
	char* assembly_version;
	char* control_file;
	int infiles;
	int quiet_flag;
	int num_query;
	int upper_limit;
	int fullsam;
	int genome_window;
	int alignment_window;
	//int reported_seed_hits;
	int num_threads;
	int iterations;
	int r;
	int step;
	int viterbi;
	int wiggle;
	int local_weighting;
	int training_sequences;
	float quality_cutoff;
	float fdr;
	int error_cutoff;
	unsigned int sim_length;
	float sim_start_error;
	float sim_end_error;
	float sim_mismatch_prob;
	float random_model_prior;
	float unaligned_probability_cutoff;
	float unaligned_prior;
	float train_pair_hmm;
	int count_mismatches_posterior;
	int sim_numseq;
	int window_start;
	float expected_error;
	int total_len; 
	int seed_errors; 
	float mean_error;
	float e_cutoff;
	
	int lightDB;
	
};

struct parameters* interface(struct parameters* param,int argc, char *argv[]);

void free_param(struct parameters* param);

void usage(void);
void usage_index(void);
void usage_seed(void);
void usage_align(void);
void usage_sim(void);
void usage_post(void);
void usage_realign(void);
void usage_show(void);
void usage_bedsim(void);
void usage_eval(void);
void usage_filter(void);








