/*
 
 Copyright (C) 2010,2011 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */


#include "delve.h"
#include "interface.h"
#include "dbio.h"
#include "align_to_sam.h"
#include "pHMM.h"

struct hmm* init_hmm(int x, int y)
{
	struct hmm* hmm = 0;
	int i,j,c;
	//float sum;
	x = x + 2;
	y = y + 2;
	hmm = malloc(sizeof(struct hmm));
	hmm->fM = malloc(sizeof(float*)*x);
	hmm->fC = malloc(sizeof(float*)*x);
	hmm->fX = malloc(sizeof(float*)*x);
	hmm->fY = malloc(sizeof(float*)*x);
	hmm->fR = malloc(sizeof(float*)*x);
	hmm->bM = malloc(sizeof(float*)*x);
	hmm->bC = malloc(sizeof(float*)*x);
	hmm->bX = malloc(sizeof(float*)*x);
	hmm->bY = malloc(sizeof(float*)*x);
	hmm->bR = malloc(sizeof(float*)*x);
	
	hmm->traceback = malloc(sizeof(unsigned short int*)*x);
	
	hmm->x = x;
	hmm->y = y;
	for(i = 0;i < x;i++){
		hmm->fM[i] = malloc(sizeof(float*)*y);
		hmm->fC[i] = malloc(sizeof(float*)*y);
		hmm->fX[i] = malloc(sizeof(float*)*y);
		hmm->fY[i] = malloc(sizeof(float*)*y);
		hmm->fR[i] = malloc(sizeof(float*)*y);
		hmm->bM[i] = malloc(sizeof(float*)*y);
		hmm->bC[i] = malloc(sizeof(float*)*y);
		hmm->bX[i] = malloc(sizeof(float*)*y);
		hmm->bY[i] = malloc(sizeof(float*)*y);
		hmm->bR[i] = malloc(sizeof(float*)*y);
		
		hmm->traceback[i] = malloc(sizeof(unsigned short int*)*y);
		for(j = 0; j < y;j++){
			hmm->fM[i][j] = 0;
			hmm->fC[i][j] = 0;
			hmm->fX[i][j] = 0;
			hmm->fY[i][j] = 0;
			hmm->fR[i][j] = 0;
			hmm->bM[i][j] = 0;
			hmm->bC[i][j] = 0;
			hmm->bX[i][j] = 0;
			hmm->bY[i][j] = 0;
			hmm->bR[i][j] = 0;
			hmm->traceback[i][j] = 0;
		}
	}
	
	hmm->rey = malloc(sizeof(float)*5);
	for(i =0; i < 5;i++){
		hmm->rey[i] = 0;
	}
	
	hmm->em = malloc(sizeof(float**) * REGIONS);
	hmm->ec = malloc(sizeof(float**) * REGIONS);
	hmm->tem = malloc(sizeof(float**) * REGIONS);
	hmm->tec = malloc(sizeof(float**) * REGIONS);
	
	
	for(c = 0;c < REGIONS;c++){
		hmm->em[c] = malloc(sizeof(float*)*5);
		hmm->tem[c] = malloc(sizeof(float*)*5);
		hmm->ec[c] = malloc(sizeof(float*)*5);
		hmm->tec[c] = malloc(sizeof(float*)*5);
		for(i = 0; i < 5;i++){
			hmm->em[c][i] = malloc(sizeof(float)*5);
			hmm->tem[c][i] = malloc(sizeof(float)*5);
			hmm->ec[c][i] = malloc(sizeof(float)*5);
			hmm->tec[c][i] = malloc(sizeof(float)*5);
			for(j = 0; j < 5;j++){
				hmm->em[c][i][j] = FLT_MIN;
				hmm->tem[c][i][j] = FLT_MIN;
				hmm->ec[c][i][j] = FLT_MIN;
				hmm->tec[c][i][j] = FLT_MIN;
			}
		}
	}		
	
	hmm->ex = malloc(sizeof(float*) * REGIONS);
	hmm->ey = malloc(sizeof(float*) * REGIONS);
	hmm->tex = malloc(sizeof(float*) * REGIONS);
	hmm->tey = malloc(sizeof(float*) * REGIONS);
	
	
	for(c = 0; c < REGIONS;c++){
		hmm->ex[c] = malloc(sizeof(float) * 5);
		hmm->ey[c] = malloc(sizeof(float) * 5);
		hmm->tex[c] = malloc(sizeof(float) * 5);
		hmm->tey[c] = malloc(sizeof(float) * 5);
	}
	
	
	hmm->mm = malloc(sizeof(float) * REGIONS);
	hmm->mc = malloc(sizeof(float) * REGIONS);
	hmm->mgx = malloc(sizeof(float) * REGIONS);
	hmm->mgy = malloc(sizeof(float) * REGIONS);
	hmm->mr = malloc(sizeof(float) * REGIONS);
	
	hmm->cc = malloc(sizeof(float) * REGIONS);
	hmm->cm = malloc(sizeof(float) * REGIONS);
	hmm->cgx = malloc(sizeof(float) * REGIONS);
	hmm->cgy = malloc(sizeof(float) * REGIONS);
	hmm->cr = malloc(sizeof(float) * REGIONS);
	
	hmm->gxgx = malloc(sizeof(float) * REGIONS);
	hmm->gxm = malloc(sizeof(float) * REGIONS);
	hmm->gxc = malloc(sizeof(float) * REGIONS);
	
	hmm->gygy = malloc(sizeof(float) * REGIONS);
	hmm->gym = malloc(sizeof(float) * REGIONS);
	hmm->gyc = malloc(sizeof(float) * REGIONS);
	
	hmm->rr = malloc(sizeof(float) * REGIONS);	
	hmm->rmc = malloc(sizeof(float) * REGIONS);
	
	hmm->tmm = malloc(sizeof(float) * REGIONS);
	hmm->tmc = malloc(sizeof(float) * REGIONS);
	hmm->tmr = malloc(sizeof(float) * REGIONS);
	hmm->tmgx = malloc(sizeof(float) * REGIONS);
	hmm->tmgy = malloc(sizeof(float) * REGIONS);
	
	hmm->tcc = malloc(sizeof(float) * REGIONS);
	hmm->tcm = malloc(sizeof(float) * REGIONS);
	hmm->tcr = malloc(sizeof(float) * REGIONS);
	hmm->tcgx = malloc(sizeof(float) * REGIONS);
	hmm->tcgy = malloc(sizeof(float) * REGIONS);
	
	hmm->tgxgx = malloc(sizeof(float) * REGIONS);
	hmm->tgxm = malloc(sizeof(float) * REGIONS);
	hmm->tgxc = malloc(sizeof(float) * REGIONS);
	
	hmm->tgygy = malloc(sizeof(float) * REGIONS);
	hmm->tgym = malloc(sizeof(float) * REGIONS);
	hmm->tgyc = malloc(sizeof(float) * REGIONS);
	hmm->tgyr = malloc(sizeof(float) * REGIONS);
	hmm->trr = malloc(sizeof(float) * REGIONS);
	hmm->trgy = malloc(sizeof(float) * REGIONS);
	
	for(c = 0;c < REGIONS;c++){
		

		hmm->mm[c] = prob2scaledprob( 0.47f);
		hmm->mc[c] = prob2scaledprob( 0.47f);
		hmm->mgx[c] = prob2scaledprob( 0.01f);
		hmm->mgy[c] = prob2scaledprob( 0.01f);
		hmm->mr[c] = prob2scaledprob(0.04f);
		
		hmm->cc[c] = prob2scaledprob( 0.47f);
		hmm->cm[c] = prob2scaledprob(0.47f);
		hmm->cgx[c] = prob2scaledprob(0.01f);
		hmm->cgy[c] = prob2scaledprob(0.01f);
		hmm->cr[c] = prob2scaledprob(0.04f);
		
			
		
		hmm->gxgx[c] = prob2scaledprob( 0.02f);
		hmm->gxm[c] = prob2scaledprob( 0.49f);
		hmm->gxc[c] = prob2scaledprob(0.49f);
		
		
		hmm->gygy[c] = prob2scaledprob(0.02f);
		hmm->gym[c] = prob2scaledprob(0.49f);
		hmm->gyc[c] = prob2scaledprob(0.49f); 
		
		//hmm->rr[c] = prob2scaledprob((1.0f - TAU) * ((float) 100.0f / (float) 101.0f)); 
		//hmm->rmc[c] = prob2scaledprob((1.0f - TAU) * ((float) 0.5f / (float) 101.0f)); //needs to be 0.5 because goes to both R and M
		hmm->rr[c] = prob2scaledprob((1.0f - TAU) * ((float) 30.0f / (float) 31.0f)); 
		hmm->rmc[c] = prob2scaledprob((1.0f - TAU) * ((float) 0.5f / (float) 31.0f)); //needs to be 0.5 because goes to both R and M

		
		hmm->tmm[c] = -SCALEINFTY;
		hmm->tmc[c] = -SCALEINFTY;
		hmm->tmr[c] = -SCALEINFTY;
		hmm->tmgx[c] = -SCALEINFTY;
		hmm->tmgy[c] = -SCALEINFTY;
		
		hmm->tcc[c] = -SCALEINFTY;
		hmm->tcm[c] = -SCALEINFTY;
		hmm->tcr[c] = -SCALEINFTY;
		hmm->tcgx[c] = -SCALEINFTY;
		hmm->tcgy[c] = -SCALEINFTY;
		
		hmm->tgxgx[c] = -SCALEINFTY;
		hmm->tgxm[c] =  -SCALEINFTY;
		hmm->tgxc[c] =  -SCALEINFTY;
		
		
		hmm->tgygy[c] = -SCALEINFTY;
		hmm->tgym[c] = -SCALEINFTY;
		hmm->tgyc[c] = -SCALEINFTY;
		hmm->tgyr[c] = -SCALEINFTY;
		hmm->trr[c] = -SCALEINFTY;
		hmm->trgy[c] = -SCALEINFTY;
				
		for(i = 0; i < 5;i++){
			hmm->ex[c][i] = prob2scaledprob(0.2f);
			hmm->ey[c][i] = prob2scaledprob(0.2f);
			hmm->tex[c][i] =  -SCALEINFTY;
			hmm->tey[c][i] =  -SCALEINFTY;
			// match should be more likely than mismatch (0.2 !!!!!!)
			for(j = 0; j < 5;j++){
				
				if(i !=j){
					hmm->ec[c][i][j] = prob2scaledprob(1.0f/50.0f);
					hmm->em[c][i][j] = prob2scaledprob(1.0f/30.0f);
				}else{
					hmm->ec[c][i][j] = prob2scaledprob(6.0f/50.0f);
					hmm->em[c][i][j] = prob2scaledprob(2.0f/30.0f);
				}
				
				hmm->tem[c][i][j] = -SCALEINFTY;
				hmm->tec[c][i][j] = -SCALEINFTY;
			}
		}
		//// MOD 4x4
		
		
		/*for(i = 0; i < 5;i++){
			sum = prob2scaledprob(0.0f);
			for(j = 0; j < 5;j++){
				sum = logsum(sum , hmm->em[c][i][j]);
			}
			for(j = 0; j < 5;j++){
				hmm->em[c][i][j] = (hmm->em[c][i][j] - sum );
				//hmm->em[c][i][j] =  -SCALEINFTY;//prob2scaledprob(FLT_MIN);
			}
			sum = prob2scaledprob(0.0f);
			for(j = 0; j < 5;j++){
				sum = logsum(sum , hmm->ec[c][i][j]);
			}
			for(j = 0; j < 5;j++){
				hmm->ec[c][i][j] = (hmm->ec[c][i][j] - sum );
				//hmm->ec[c][i][j] =  -SCALEINFTY;//prob2scaledprob(FLT_MIN);
			}
		}*/
		

	
		
	}
	return hmm;
}

void free_hmm(struct hmm* hmm)
{
	int i,j;
	int x;//,y; 
	
	x = hmm->x;
	//y = hmm->y;
	for(i = 0;i < x;i++){
		free(hmm->traceback[i]);
		free(hmm->fM[i]);
		free(hmm->fC[i]);
		free(hmm->fX[i]);
		free(hmm->fY[i]);
		free(hmm->fR[i]);
		free(hmm->bM[i]);
		free(hmm->bC[i]);
		free(hmm->bX[i]);
		free(hmm->bY[i]);
		free(hmm->bR[i]);
	}
	
	free(hmm->traceback);
	
	free(hmm->rey);
	
	free(hmm->fM);
	free(hmm->fC);
	free(hmm->fX);
	free(hmm->fY);
	free(hmm->fR);
	free(hmm->bM);
	free(hmm->bC);
	free(hmm->bX);
	free(hmm->bY);
	free(hmm->bR);
	
	for(j = 0; j < REGIONS;j++){
		free(hmm->ex[j]);
		free(hmm->ey[j]);
		free(hmm->tex[j]);
		free(hmm->tey[j]);
		
		for(i = 0; i < 5;i++){
			free(hmm->em[j][i]);
			free(hmm->tem[j][i]);
			free(hmm->ec[j][i]);
			free(hmm->tec[j][i]);
		}
		free(hmm->tem[j]);
		free(hmm->em[j]);
		free(hmm->tec[j]);
		free(hmm->ec[j]);
	}
	
	free(hmm->ex);
	free(hmm->ey);
	free(hmm->tex);
	free(hmm->tey);
	
	
	free(hmm->tem);
	free(hmm->em);
	free(hmm->tec);
	free(hmm->ec);
	
	
	free(hmm->mm);
	free(hmm->mc);
	free(hmm->mgx);
	free(hmm->mgy);
	free(hmm->mr);
	
	free(hmm->cc);
	free(hmm->cm);
	free(hmm->cgx);
	free(hmm->cgy);
	free(hmm->cr);
	
	free(hmm->gxgx);
	free(hmm->gxm);
	free(hmm->gxc);
	
	free(hmm->gygy);
	free(hmm->gym);
	free(hmm->gyc);
	
	free(hmm->rr);
	free(hmm->rmc);
	
	free(hmm->tmm);
	free(hmm->tmc);
	free(hmm->tmr);
	free(hmm->tmgx);
	free(hmm->tmgy);
	
	free(hmm->tcc);
	free(hmm->tcm);
	free(hmm->tcr);
	free(hmm->tcgx);
	free(hmm->tcgy);
	
	free(hmm->tgxgx);
	free(hmm->tgxm);
	free(hmm->tgxc);
	
	free(hmm->tgygy);
	free(hmm->tgym);
	free(hmm->tgyc);
	free(hmm->tgyr);
	free(hmm->trr);
	free(hmm->trgy);
	
	
	free(hmm);
}



void* do_baum_welch_95ID_thread(void *threadarg)
{
	struct thread_data *data;
	data = (struct thread_data *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm* hmm = data->hmm;
	struct db* db = data->db;
	const int start = data->start; 
	const int end = data->end;
	int i,j,hit;	
	int strand = 0;
	unsigned char* seq = 0;
	//float viterbi_score = 0.0;
	seq = malloc(sizeof(unsigned char) * 500);
	float avg_score = 100000000;// = prob2scaledprob(0.0);
	float* random_emission = 0;
	random_emission = malloc(sizeof(float)*5);
	random_emission[0] = prob2scaledprob(0.2475f);
	random_emission[1] = prob2scaledprob(0.2475f);
	random_emission[2] = prob2scaledprob(0.2475f);
	random_emission[3] = prob2scaledprob(0.2475f);
	random_emission[4] = prob2scaledprob(0.01f);
	for(i = start; i < end;i++){
		hit = 0;
		j = 0;
		while(ri[i]->identity[hit] != -2.0f){
			if(ri[i]->identity[hit] >= 0.95){
				if(hit == 0){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				}else{
					if(ri[i]->strand[hit] != ri[i]->strand[0]){
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
					}else{
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
					}
				}
				
				hmm = global_local_forward_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]); 
				j++;
				if(avg_score == 100000000){ 
					avg_score = hmm->score;
				}else{
					avg_score =  logsum(avg_score, hmm->score);
				}
				//fprintf(stderr,"%lud	%d	%f	%f\n",seeds[i][j],i, hmm->score, avg_score);
				hmm = global_local_backward_log(hmm,random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]);
				hmm = get_prob_log(hmm, random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20, 0,ri[i]->strand[hit]) ;
			}
			hit++;
		}
		

		
	}
	free(random_emission);
	
	if(start == 0 ){
		fprintf(stderr,"SCORE:%d	%f\n",start,avg_score);
	}
	free(seq);
	pthread_exit((void *) 0);
}


void* do_baum_welch_thread(void *threadarg)
{
	struct thread_data *data;
	data = (struct thread_data *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm* hmm = data->hmm;
	struct db* db = data->db;
	const int start = data->start; 
	const int end = data->end;
	int i,j,hit;	
	int strand = 0;
	unsigned char* seq = 0;
	
	float pseudocount = 0.0f;
	//float viterbi_score = 0.0;
	seq = malloc(sizeof(unsigned char) * 500);
	float avg_score = 100000000;// = prob2scaledprob(0.0);
	float* random_emission = 0;
	random_emission = malloc(sizeof(float)*5);
	random_emission[0] = prob2scaledprob(0.2475f);
	random_emission[1] = prob2scaledprob(0.2475f);
	random_emission[2] = prob2scaledprob(0.2475f);
	random_emission[3] = prob2scaledprob(0.2475f);
	random_emission[4] = prob2scaledprob(0.01f);	
	float sum = 0;
	int c;
	float* scores = malloc(sizeof(float)* 100); 
	
	for(i = start; i < end;i++){
		hit = 0;
		j = 0;
		while(ri[i]->identity[hit] != -2.0f){
			//if(ri[i]->identity[hit] >= 0.95){
				if(hit == 0){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				}else{
					if(ri[i]->strand[hit] != ri[i]->strand[0]){
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);

					}else{
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
					}
				}
				
				hmm = global_local_forward_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]  ); 
				scores[j] = hmm->score + ri[i]->priors[hit];
				j++;
				//if(avg_score == 100000000){ 
				//	avg_score = hmm->score;
				//}else{
				//	avg_score =  logsum(avg_score, hmm->score);
				//}
				//fprintf(stderr,"%lud	%d	%f	%f\n",seeds[i][j],i, hmm->score, avg_score);
				//hmm = global_local_backward_log(hmm,random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20);
				//hmm = get_prob_log(hmm, random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20);
			//}
			hit++;
		}
		pseudocount = prob2scaledprob( 1.0 / hit );
		sum = scores[0];
		for(c = 1; c < j;c++){
			sum = logsum(sum,scores[c]);
		}
		/*max = -1.0;
		for(c = 0;c < j;c++){
			scores[c] = scaledprob2prob(scores[c] - sum);
			if(scores[c] > max){
				max= scores[c];
			}
		//	fprintf(stderr,"mq(%d):%f\n",c,scores[c] );
		}*/
		//fprintf(stderr,"\n");
		//sum = 0;
		//for(c = 0;c < j;c++){
		//	scores[c]  = ((1.0f - max)/(float) hit ) + max * scores[c];
			//sum += scores[c] ;
		//	fprintf(stderr,"prior(%d):%f\n",c,scores[c] );
		//}
		//fprintf(stderr,"SUM: %f\n",sum);
		
		
		
		hit = 0;
		//j = 0;
		while(ri[i]->identity[hit] != -2.0f){
			//if(ri[i]->identity[hit] >= 0.95){
				if(hit == 0){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				}else{
					if(ri[i]->strand[hit] != ri[i]->strand[0]){
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
					}else{
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
					}
				}
				
				
				
				hmm = global_local_forward_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]); 
				
				if(avg_score == 100000000){ 
					avg_score = hmm->score;
				}else{
					avg_score =  logsum(avg_score, hmm->score);
				}
				//fprintf(stderr,"%lud	%d	%f	%f\n",seeds[i][j],i, hmm->score, avg_score);
				hmm = global_local_backward_log(hmm,random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]);
				hmm = get_prob_log(hmm, random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 , hmm->score - sum + ri[i]->priors[hit],ri[i]->strand[hit]);
				ri[i]->priors[hit] =  hmm->score - sum + ri[i]->priors[hit] ;
				ri[i]->priors[hit] = logsum(ri[i]->priors[hit], pseudocount);
			//}
			
			hit++;
		}
		
		
	}
	free(scores);
	free(random_emission);
	
	//if(start == 0 ){
		fprintf(stderr,"SCORE:%d		%d%f\n",start,start,avg_score);
	//}
	free(seq);
	pthread_exit((void *) 0);
}


void init_thread_hmms(struct thread_data* thread_data_array,struct hmm* localhmm,int numthreads)
{
	int i,j,c,g;
	for(c = 0; c < numthreads;c++){
		for(i = 0; i < 5;i++){
			thread_data_array[c].hmm->rey[i] = localhmm->rey[i];// = logsum(localhmm->tex[i],);
		}
	}
	
	
	for(g = 0; g < REGIONS;g++){
	
		for(c = 0; c < numthreads;c++){
			for(i = 0; i < 5;i++){
				thread_data_array[c].hmm->tex[g][i] = localhmm->tex[g][i];// = logsum(localhmm->tex[i],);
				thread_data_array[c].hmm->tey[g][i] = localhmm->tey[g][i];// = logsum(localhmm->tey[i],);
				thread_data_array[c].hmm->ex[g][i] = localhmm->ex[g][i];// = logsum(localhmm->tex[i],);
				thread_data_array[c].hmm->ey[g][i] = localhmm->ey[g][i];// = logsum(localhmm->tey[i],);
				for(j = 0; j < 5;j++){
					thread_data_array[c].hmm->tem[g][i][j] = localhmm->tem[g][i][j];// = logsum(localhmm->tem[i][j],);
					thread_data_array[c].hmm->tec[g][i][j] = localhmm->tec[g][i][j];// =  logsum(localhmm->tec[i][j],);
					thread_data_array[c].hmm->em[g][i][j] = localhmm->em[g][i][j];// = logsum(localhmm->tem[i][j],);
					thread_data_array[c].hmm->ec[g][i][j] = localhmm->ec[g][i][j];// =  logsum(localhmm->tec[i][j],);
					
				}
			}
			thread_data_array[c].hmm->tmm[g] = localhmm->tmm[g];//  = logsum(localhmm->tmm,);
			thread_data_array[c].hmm->tmc[g] = localhmm->tmc[g];
			thread_data_array[c].hmm->tmr[g] = localhmm->tmr[g];//  = logsum(localhmm->tmc,);
			thread_data_array[c].hmm->tmgx[g] = localhmm->tmgx[g];//  = logsum(localhmm->tmgx,);
			thread_data_array[c].hmm->tmgy[g] = localhmm->tmgy[g];//  = logsum(localhmm->tmgy,);
			
			thread_data_array[c].hmm->tcc[g] = localhmm->tcc[g];//  = logsum(localhmm->tcc,);
			thread_data_array[c].hmm->tcm[g] = localhmm->tcm[g];//   = logsum(localhmm->tcm,);
			thread_data_array[c].hmm->tcr[g] = localhmm->tcr[g];//   = logsum(localhmm->tcm,);
			thread_data_array[c].hmm->tcgx[g] = localhmm->tcgx[g];//  = logsum(localhmm->tcgx,);
			thread_data_array[c].hmm->tcgy[g] = localhmm->tcgy[g];//  = logsum(localhmm->tcgy,);
			
			thread_data_array[c].hmm->tgxgx[g] = localhmm->tgxgx[g];//  = logsum(localhmm->tgxgx,);
			thread_data_array[c].hmm->tgxm[g] = localhmm->tgxm[g];//  = logsum(localhmm->tgxm,);
			thread_data_array[c].hmm->tgxc[g] = localhmm->tgxc[g];//  = logsum(localhmm->tgxc,);
			
			thread_data_array[c].hmm->tgygy[g] = localhmm->tgygy[g];//  = logsum(localhmm->tgygy,);
			thread_data_array[c].hmm->tgym[g] = localhmm->tgym[g];//  = logsum(localhmm->tgym,);
			thread_data_array[c].hmm->tgyc[g] = localhmm->tgyc[g];//  = logsum(localhmm->tgyc,);
			
			thread_data_array[c].hmm->mm[g] = localhmm->mm[g];//  = logsum(localhmm->tmm,);
			thread_data_array[c].hmm->mc[g] = localhmm->mc[g];//  = logsum(localhmm->tmc,);
			thread_data_array[c].hmm->mr[g] = localhmm->mr[g];//  = logsum(localhmm->tmc,);
			thread_data_array[c].hmm->mgx[g] = localhmm->mgx[g];//  = logsum(localhmm->tmgx,);
			thread_data_array[c].hmm->mgy[g] = localhmm->mgy[g];//  = logsum(localhmm->tmgy,);
			
			thread_data_array[c].hmm->cc[g] = localhmm->cc[g];//  = logsum(localhmm->tcc,);
			thread_data_array[c].hmm->cm[g] = localhmm->cm[g];//   = logsum(localhmm->tcm,);
			thread_data_array[c].hmm->cr[g] = localhmm->cr[g];//   = logsum(localhmm->tcm,);
			thread_data_array[c].hmm->cgx[g] = localhmm->cgx[g];//  = logsum(localhmm->tcgx,);
			thread_data_array[c].hmm->cgy[g] = localhmm->cgy[g];//  = logsum(localhmm->tcgy,);
			
			thread_data_array[c].hmm->gxgx[g] = localhmm->gxgx[g];//  = logsum(localhmm->tgxgx,);
			thread_data_array[c].hmm->gxm[g] = localhmm->gxm[g];//  = logsum(localhmm->tgxm,);
			thread_data_array[c].hmm->gxc[g] = localhmm->gxc[g];//  = logsum(localhmm->tgxc,);
			
			thread_data_array[c].hmm->gygy[g] = localhmm->gygy[g];//  = logsum(localhmm->tgygy,);
			thread_data_array[c].hmm->gym[g] = localhmm->gym[g];//  = logsum(localhmm->tgym,);
			thread_data_array[c].hmm->gyc[g] = localhmm->gyc[g];//  = logsum(localhmm->tgyc,);
			
			thread_data_array[c].hmm->rr[g] = localhmm->rr[g];
			thread_data_array[c].hmm->rmc[g] = localhmm->rmc[g];
		}
	}
}

void entangle_hmms(struct thread_data* thread_data_array,struct hmm* localhmm,int numthreads)
{
	int i,j,c,g;
	
	//set local to data in thread 1 
	for(g = 0;g < REGIONS;g++){
		for(i = 0; i < 5;i++){
			localhmm->tex[g][i] = thread_data_array[0].hmm->tex[g][i];
			localhmm->tey[g][i] = thread_data_array[0].hmm->tey[g][i];
			for(j = 0; j < 5;j++){
				localhmm->tem[g][i][j] = thread_data_array[0].hmm->tem[g][i][j];
				localhmm->tec[g][i][j] =  thread_data_array[0].hmm->tec[g][i][j];
				
			}
		}
		
		localhmm->tmm[g] = thread_data_array[0].hmm->tmm[g];
		localhmm->tmc[g] = thread_data_array[0].hmm->tmc[g];
		localhmm->tmr[g] = thread_data_array[0].hmm->tmr[g];
		localhmm->tmgx[g] = thread_data_array[0].hmm->tmgx[g];
		localhmm->tmgy[g] = thread_data_array[0].hmm->tmgy[g];
		
		localhmm->tcc[g] = thread_data_array[0].hmm->tcc[g];
		localhmm->tcm[g] = thread_data_array[0].hmm->tcm[g];
		localhmm->tcr[g] = thread_data_array[0].hmm->tcr[g];
		localhmm->tcgx[g] = thread_data_array[0].hmm->tcgx[g];
		localhmm->tcgy[g] = thread_data_array[0].hmm->tcgy[g];
		
		localhmm->tgxgx[g] = thread_data_array[0].hmm->tgxgx[g];
		localhmm->tgxm[g] = thread_data_array[0].hmm->tgxm[g];
		localhmm->tgxc[g] = thread_data_array[0].hmm->tgxc[g];
		
		localhmm->tgygy[g] = thread_data_array[0].hmm->tgygy[g];
		localhmm->tgym[g] = thread_data_array[0].hmm->tgym[g];
		localhmm->tgyc[g] = thread_data_array[0].hmm->tgyc[g];
	}
	
	for(g = 0;g < REGIONS;g++){
		for(c = 1; c < numthreads;c++){
			for(i = 0; i < 5;i++){
				localhmm->tex[g][i] = logsum(localhmm->tex[g][i],thread_data_array[c].hmm->tex[g][i]);
				localhmm->tey[g][i] = logsum(localhmm->tey[g][i],thread_data_array[c].hmm->tey[g][i]);
				for(j = 0; j < 5;j++){
					localhmm->tem[g][i][j] = logsum(localhmm->tem[g][i][j],thread_data_array[c].hmm->tem[g][i][j]);
					localhmm->tec[g][i][j] =  logsum(localhmm->tec[g][i][j],thread_data_array[c].hmm->tec[g][i][j]);
					
				}
			}
			localhmm->tmm[g] = logsum(localhmm->tmm[g],thread_data_array[c].hmm->tmm[g]);
			localhmm->tmc[g] = logsum(localhmm->tmc[g],thread_data_array[c].hmm->tmc[g]);
			localhmm->tmr[g] = logsum(localhmm->tmr[g],thread_data_array[c].hmm->tmr[g]);
			localhmm->tmgx[g] = logsum(localhmm->tmgx[g],thread_data_array[c].hmm->tmgx[g]);
			localhmm->tmgy[g] = logsum(localhmm->tmgy[g],thread_data_array[c].hmm->tmgy[g]);
			
			localhmm->tcc[g] = logsum(localhmm->tcc[g],thread_data_array[c].hmm->tcc[g]);
			localhmm->tcm[g] = logsum(localhmm->tcm[g],thread_data_array[c].hmm->tcm[g]);
			localhmm->tcr[g] = logsum(localhmm->tcr[g],thread_data_array[c].hmm->tcr[g]);
			localhmm->tcgx[g] = logsum(localhmm->tcgx[g],thread_data_array[c].hmm->tcgx[g]);
			localhmm->tcgy[g] = logsum(localhmm->tcgy[g],thread_data_array[c].hmm->tcgy[g]);
			
			localhmm->tgxgx[g] = logsum(localhmm->tgxgx[g],thread_data_array[c].hmm->tgxgx[g]);
			localhmm->tgxm[g] = logsum(localhmm->tgxm[g],thread_data_array[c].hmm->tgxm[g]);
			localhmm->tgxc[g] = logsum(localhmm->tgxc[g],thread_data_array[c].hmm->tgxc[g]);
			
			localhmm->tgygy[g] = logsum(localhmm->tgygy[g],thread_data_array[c].hmm->tgygy[g]);
			localhmm->tgym[g] = logsum(localhmm->tgym[g],thread_data_array[c].hmm->tgym[g]);
			localhmm->tgyc[g] = logsum(localhmm->tgyc[g],thread_data_array[c].hmm->tgyc[g]);
		}
	}
}

unsigned char* global_local_viterbi_log(struct hmm* hmm,float* rey,unsigned char* a, unsigned char* b, int n,int m,int strand)
{
	int i,j,c,g;
	unsigned short int tmp_tb;
	
	float mm[REGIONS];
	float mc[REGIONS];
	float mr[REGIONS];
	float mgx[REGIONS];
	float mgy[REGIONS];
	
	float gxgx[REGIONS];
	float gxm[REGIONS];
	float gxc[REGIONS];
	
	float gygy[REGIONS];
	float gym[REGIONS];
	float gyc[REGIONS];
	
	float cc[REGIONS];
	float cm[REGIONS];
	float cr[REGIONS];
	float cgx[REGIONS];
	float cgy[REGIONS];
	
	float rr[REGIONS];
	
	float rmc[REGIONS];
	
	//float max = 0;
	for(g = 0; g < REGIONS;g++){
		mm[g] = hmm->mm[g];
		mc[g] = hmm->mc[g];
		mr[g] = hmm->mr[g];
		mgx[g] = hmm->mgx[g];
		mgy[g] = hmm->mgy[g];
		
		gxgx[g] = hmm->gxgx[g];
		gxm[g] = hmm->gxm[g];
		gxc[g] = hmm->gxc[g];
		
		gygy[g] = hmm->gygy[g];
		gym[g] = hmm->gym[g];
		gyc[g] = hmm->gyc[g];
		
		cc[g] = hmm->cc[g];
		cm[g] = hmm->cm[g];
		cr[g] = hmm->cr[g];
		cgx[g] = hmm->cgx[g];
		cgy[g] = hmm->cgy[g];
		
		//rr[g]  = hmm->rr[g];
		//rmc[g] = hmm->rmc[g];
		rr[g] = prob2scaledprob((float)m / (float)(m+1));
		rmc[g] = prob2scaledprob((float)0.5f / (float)(m+1));

	}
	
		
	/*for(i = 0; i < 5;i++){
		rey[i] =0;
	}
	for(i = 0; i < m;i++){
		rey[b[i]]++;
	}
	for(i = 0; i < 5;i++){
		rey[i] = prob2scaledprob( (float) rey[i] / (float)m  );
	}*/
	
	
	float** M = hmm->fM;
	float** C = hmm->fC;
	float** X = hmm->fX;
	float** Y = hmm->fY;
	float** R = hmm->fR;
	unsigned short int** t = hmm->traceback;
	
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	unsigned char* alignment =  malloc(sizeof(unsigned char)* (n+m+2));
	unsigned short int* state_path =  malloc(sizeof(unsigned short int)* (n+m+2));
	
	
	M[0][0] = -SCALEINFTY;//prob2scaledprob(0.25f); 
	C[0][0] = -SCALEINFTY;//prob2scaledprob(0.25f); 
	X[0][0] = -SCALEINFTY;//prob2scaledprob(0.25f);
	Y[0][0] = -SCALEINFTY;//prob2scaledprob(0.01f);
	R[0][0] = prob2scaledprob(1.0f);
	
	for(i = 1; i <= n;i++){
		t[i][0] = 0;
		M[i][0] = -SCALEINFTY;
		C[i][0] = -SCALEINFTY;
		X[i][0] = -SCALEINFTY;
		Y[i][0] = -SCALEINFTY;
		R[i][0] = -SCALEINFTY;
		//t[i][0] = 2 << 6; 
	}
	
	if(!strand){
		g = 0;
	}else{
		g = REGIONS -1;
	}
	
	for(j = 1; j <= m;j++){
		tmp_tb = 0; 
		M[0][j] = -SCALEINFTY;
		C[0][j] = -SCALEINFTY;
		X[0][j] = -SCALEINFTY;
		
		Y[0][j] = -SCALEINFTY;
		
		R[0][j] = R[0][j-1] + rr[g];
		tmp_tb |= (4 << 12);
		
		R[0][j] += rey[seqb[j]];
		t[0][j] = tmp_tb;
	}
	
	
	for(i = 1; i <= n;i++){
		
		g = (int) floor((float)(i-1) / (float)n * (float)REGIONS); 
		if(strand){
			g = REGIONS - (1+g); 
		}
		for(j = 1; j <= m;j++){
			tmp_tb = 0;
			/*M[i][j] = hmm->em[seqa[i]][seqb[j]] + MAX3(mm + M[i-1][j-1] , gm + X[i-1][j-1], gm + Y[i-1][j-1]);
			 X[i][j] = hmm->ex[seqa[i]] + MAX (mg + M[i-1][j],gg + X[i-1][j]);
			 Y[i][j] = hmm->ey[seqb[j]] + MAX (mg + M[i][j-1],gg + Y[i][j-1]);
			 */
			
			M[i][j] = mm[g] + M[i-1][j-1];
			if(cm[g] + C[i-1][j-1] > M[i][j]){
				M[i][j] = cm[g] + C[i-1][j-1];
				tmp_tb = 1;
			}
			if(gxm[g] + X[i-1][j-1]  > M[i][j]){
				M[i][j] = gxm[g] + X[i-1][j-1];
				tmp_tb = 2;
			}
			if(gym[g] + Y[i-1][j-1] > M[i][j]){
				M[i][j] = gym[g] + Y[i-1][j-1];
				tmp_tb = 3;
			}
			
			if(rmc[g] + R[i-1][j-1] > M[i][j]){
				M[i][j] = rmc[g] + R[i-1][j-1];
				tmp_tb = 4;
			}
			
			
			M[i][j] += hmm->em[g][seqa[i]][seqb[j]];
			
			t[i][j] =  tmp_tb;
			//M[i][j] = hmm->em[seqa[i]][seqb[j]] + MAX4(mm + M[i-1][j-1], cm + C[i-1][j-1], gmx + X[i-1][j-1] , gmy + Y[i-1][j-1] );
			//tmp_tb = 0;
			C[i][j] = cc[g] + C[i-1][j-1];
			tmp_tb = (1 << 3);
			if( mc[g] + M[i-1][j-1] > C[i][j]){
				C[i][j] =  mc[g] + M[i-1][j-1];
				tmp_tb =0;
			}
			
			if(gxc[g] + X[i-1][j-1]  > C[i][j]){
				C[i][j] = gxc[g] + X[i-1][j-1];
				tmp_tb = (2 << 3);
			}
			if(gyc[g] + Y[i-1][j-1] > C[i][j]){
				C[i][j] = gyc[g] + Y[i-1][j-1];
				tmp_tb = (3 << 3);
			}
			
			if(rmc[g] + R[i-1][j-1] > C[i][j]){
				C[i][j] = rmc[g] + R[i-1][j-1];
				tmp_tb = (4 << 3);
			}
			
			C[i][j] += hmm->ec[g][seqa[i]][seqb[j]];
			
			t[i][j] |=  tmp_tb;
			//tmp_tb = 0;
			
			X[i][j] = gxgx[g] + X[i-1][j] ;
			tmp_tb = (2 << 6);
			
			if(cgx[g] + C[i-1][j] > X[i][j]){
				X[i][j] = cgx[g] + C[i-1][j];
				tmp_tb = (1 << 6);
			}
			
			if(mgx[g] + M[i-1][j] > X[i][j]){
				X[i][j] = mgx[g] + M[i-1][j];
				tmp_tb = 0;
			}
			X[i][j] += hmm->ex[g][seqa[i]];
			
			t[i][j] |=  tmp_tb;
			//tmp_tb = 0;
			
			
			
			//X[i][j] = hmm->ex[seqa[i]] + MAX(mgx + M[i-1][j] , ggx + X[i-1][j]);
			Y[i][j]= gygy[g] + Y[i][j-1] ;
			tmp_tb = (3 << 9);
			
			if(cgy[g] + C[i][j-1] > Y[i][j]){
				Y[i][j] = cgy[g] + C[i][j-1];
				tmp_tb = (1 << 9);
			}
			
			
			if(mgy[g] + M[i][j-1] > Y[i][j]){
				Y[i][j] = mgy[g] + M[i][j-1];
				tmp_tb = 0;
			}
			/*if(rgy + R[i][j-1] > Y[i][j]){
			 Y[i][j] = rgy + Y[i][j-1];
			 tmp_tb = (4 << 9);
			 }*/
			
			Y[i][j] +=  hmm->ey[g][seqb[j]];
			
			t[i][j] |=  tmp_tb;
			//tmp_tb = 0;
			
			//Y[i][j] = hmm->ey[seqb[j]]  + logsum(logsum(mgy + M[i][j-1],cgy + C[i][j-1]) , logsum(gygy + Y[i][j-1],rgy + R[i][j-1]) );
			//Y[i][j] = hmm->ey[seqb[j]]  + MAX(mgy + M[i][j-1] , ggy + Y[i][j-1]);
			R[i][j] = R[i][j-1] + rr[g];
			tmp_tb = (4 << 12);
			
			if(mr[g] + M[i][j-1]  > R[i][j]){
				R[i][j] = mr[g] + M[i][j-1];
				tmp_tb = (0 << 12);
			}
			if(cr[g] + C[i][j-1] > R[i][j]){
				R[i][j] = cr[g] + C[i][j-1];
				tmp_tb = (1 << 12);
			}
			
			
			/*if(R[i][j-1] + rr > Y[i][j-1] + gyr ){
			 
			 
			 }else{
			 R[i][j] = Y[i][j-1] + gyr;
			 tmp_tb = (3 << 12);
			 }*/
			
			R[i][j] += rey[seqb[j]]; 
			t[i][j] |=  tmp_tb;
		}
	}
	hmm->score = TAU_LOG + R[n][m];
	i = n;
	j = m;
	
	//if(Y[i][j] > R[i][j]){
		
	//	tmp_tb = 3;
	//}else{
	//	tmp_tb = 4;
	//}
	tmp_tb = 4;
	//fprintf(stderr,"%f	%f	%f	%f\n",M[i][j],C[i][j],X[i][j],Y[i][j]);
	
	float id = 0.0f;
	c = 1;
	state_path[c] = tmp_tb;
	
	while(1){
		switch (tmp_tb) {
			case 0:
				alignment[c] = (seqa[i] << 4) | seqb[j];
				
				if(seqa[i] == seqb[j]){
					id += 1.0f;
				}
				
				tmp_tb = t[i][j] & 0x7;
				i--;
				j--;
				break;
			case 1:
				alignment[c] = (seqa[i] << 4) | seqb[j];
				
				if(seqa[i] == seqb[j]){
					id += 1.0f;
				}
				
				tmp_tb = (t[i][j] >> 3 ) & 0x7;
				i--;
				j--;
				break;
			case 2:
				alignment[c] = (seqa[i] << 4)  | 5;
				tmp_tb = (t[i][j] >> 6 ) & 0x7;
				i--;	
				break;
			case 3:
				alignment[c] = (5 << 4) | seqb[j] ;
				tmp_tb = (t[i][j] >> 9 ) & 0x7;
				j--;	
				break;
			case 4:
				alignment[c] = (5 << 4) | seqb[j] ;
				tmp_tb = (t[i][j] >> 12 ) & 0x7;
				j--;	
				break;
				
			default:
				fprintf(stderr,"ERROR:code: %d not found\n",tmp_tb);
				break;
		}
		c++;
		state_path[c] = tmp_tb;
		if(i == 0 && j == 0){
			break;
		}
	}
	alignment[0] = c-1; 
	
	
	
	/*fprintf(stderr,"SeqA:");
	 for(i = c-1; i >=1;i--){
	 if((alignment[i] >> 4) == 5){
	 fprintf(stderr,"-");
	 }else{
	 fprintf(stderr,"%d",alignment[i] >> 4);
	 }
	 }
	 fprintf(stderr,"\n");
	 fprintf(stderr,"SeqB:");
	 for(i = c-1; i >=1;i--){
	 if((alignment[i] & 0xF) ==5){
	 fprintf(stderr,"-");
	 }else{
	 fprintf(stderr,"%d",alignment[i] & 0xF);
	 }
	 }
	 fprintf(stderr,"\n");
	 fprintf(stderr,"Path:");
	 
	 
	 for(i = c-1; i >=1;i--){
	 switch (state_path[i]) {
	 case 0:
	 fprintf(stderr,"M");
	 break;
	 case 1:
	 fprintf(stderr,"C");
	 break;
	 case 2:
	 fprintf(stderr,"X");
	 
	 break;
	 case 3:
	 fprintf(stderr,"Y");
	 
	 break;
	 case 4:
	 fprintf(stderr,"R");
	 
	 break;
	 default:
	 break;
	 }
	 }
	
	 fprintf(stderr,"	%0.1f\n\n",id / (float) n * 100.0f );*/
	free(state_path);
	return alignment;
}



struct hmm* global_local_forward_log(struct hmm* hmm,float* rey,unsigned char* a, unsigned char* b, int n,int m,int strand)
{
	//init - terminal gap penalty = 0 (p = 1)
	int i,j,g;
	
	float mm[REGIONS];
	float mc[REGIONS];
	float mr[REGIONS];
	float mgx[REGIONS];
	float mgy[REGIONS];
	
	float gxgx[REGIONS];
	float gxm[REGIONS];
	float gxc[REGIONS];
	
	float gygy[REGIONS];
	float gym[REGIONS];
	float gyc[REGIONS];
	
	float cc[REGIONS];
	float cm[REGIONS];
	float cr[REGIONS];
	float cgx[REGIONS];
	float cgy[REGIONS];
	
	float rr[REGIONS];
	
	float rmc[REGIONS];
	
	//float max = 0;
	for(g = 0; g < REGIONS;g++){
		mm[g] = hmm->mm[g];
		mc[g] = hmm->mc[g];
		mr[g] = hmm->mr[g];
		mgx[g] = hmm->mgx[g];
		mgy[g] = hmm->mgy[g];
		
		gxgx[g] = hmm->gxgx[g];
		gxm[g] = hmm->gxm[g];
		gxc[g] = hmm->gxc[g];
		
		gygy[g] = hmm->gygy[g];
		gym[g] = hmm->gym[g];
		gyc[g] = hmm->gyc[g];
		
		cc[g] = hmm->cc[g];
		cm[g] = hmm->cm[g];
		cr[g] = hmm->cr[g];
		cgx[g] = hmm->cgx[g];
		cgy[g] = hmm->cgy[g];
		//rr[g]  = hmm->rr[g];
		//rmc[g] = hmm->rmc[g];
		rr[g] = prob2scaledprob((float)m / (float)(m+1));
		rmc[g] = prob2scaledprob((float)0.5f / (float)(m+1));
	}
	/*for(i = 0; i < 5;i++){
		rey[i] =0;
	}
	for(i = 0; i < m;i++){
		rey[b[i]]++;
	}
	for(i = 0; i < 5;i++){
		rey[i] = prob2scaledprob( (float) rey[i] / (float)m  );
	}*/
	
	
	float** M = hmm->fM;
	float** C = hmm->fC;
	float** X = hmm->fX;
	float** Y = hmm->fY;
	float** R = hmm->fR;
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	
	
	if(!strand){
		g = 0;
	}else{
		g = REGIONS -1;
	}
	
	M[0][0] = -SCALEINFTY;//prob2scaledprob(0.25f); 
	C[0][0] = -SCALEINFTY;//prob2scaledprob(0.25f); 
	X[0][0] = -SCALEINFTY;//prob2scaledprob(0.25f);
	Y[0][0] = -SCALEINFTY;//prob2scaledprob(0.01f);
	R[0][0] = prob2scaledprob(1.0f);
	
	for(i = 1; i <= n;i++){
		M[i][0] = -SCALEINFTY;
		C[i][0] = -SCALEINFTY;
		X[i][0] = -SCALEINFTY;//hmm->ex[seqa[i]] + logsum (logsum(mgx + M[i-1][0],cgx + C[i-1][0]),  gxgx + X[i-1][0]);
		Y[i][0] = -SCALEINFTY;
		R[i][0] = -SCALEINFTY;
	}
	
	for(j = 1; j <= m;j++){
		M[0][j] = -SCALEINFTY;
		C[0][j] = -SCALEINFTY;
		X[0][j] = -SCALEINFTY;
		Y[0][j] = -SCALEINFTY;//hmm->ey[seqb[j]] + logsum(logsum(mgy + M[0][j-1] , cgy + C[0][j-1]) , logsum(gygy + Y[0][j-1],rgy + R[0][j-1] ));
		R[0][j] = rey[seqb[j]] + R[0][j-1] + rr[g];//logsum(R[0][j-1] + rr ,Y[0][j-1] + gyr);
	}
	
	for(i = 1; i <= n;i++){
		g = (int) floor((float)(i-1) / (float)n * (float)REGIONS); 
		if(strand){
			g = REGIONS -(1+ g);
		}
		for(j = 1; j <= m;j++){
			M[i][j] = hmm->em[g][seqa[i]][seqb[j]] + logsum(logsum(logsum(mm[g] + M[i-1][j-1], cm[g] + C[i-1][j-1]), rmc[g] + R[i-1][j-1]),logsum(gxm[g] + X[i-1][j-1],gym[g] + Y[i-1][j-1] ));
			C[i][j] = hmm->ec[g][seqa[i]][seqb[j]] + logsum(logsum(logsum(mc[g] + M[i-1][j-1], cc[g] + C[i-1][j-1]), rmc[g] + R[i-1][j-1]),logsum(gxc[g] + X[i-1][j-1],gyc[g] + Y[i-1][j-1] )) ;
			X[i][j] = hmm->ex[g][seqa[i]] + logsum(logsum(mgx[g] + M[i-1][j],cgx[g] + C[i-1][j]) , gxgx[g] + X[i-1][j]);
			Y[i][j] = hmm->ey[g][seqb[j]] + logsum(logsum(mgy[g] + M[i][j-1],cgy[g] + C[i][j-1]) , gygy[g] + Y[i][j-1]);
			R[i][j] =  rey[seqb[j]] + logsum(R[i][j-1] + rr[g],logsum(mr[g] + M[i][j-1] ,cr[g] + C[i][j-1]));
		}
	}
	hmm->score = TAU_LOG + R[n][m];
	return hmm;
}

struct hmm* global_local_backward_log(struct hmm* hmm,float* rey,unsigned char* a, unsigned char* b, int n,int m,int strand)
{
	int i,j,g;
	
	float mm[REGIONS];
	float mc[REGIONS];
	float mr[REGIONS];
	float mgx[REGIONS];
	float mgy[REGIONS];
	
	float gxgx[REGIONS];
	float gxm[REGIONS];
	float gxc[REGIONS];
	
	float gygy[REGIONS];
	float gym[REGIONS];
	float gyc[REGIONS];
	
	float cc[REGIONS];
	float cm[REGIONS];
	float cr[REGIONS];
	float cgx[REGIONS];
	float cgy[REGIONS];
	
	float rr[REGIONS];
	
	float rmc[REGIONS];
	
	//float max = 0;
	for(g = 0; g < REGIONS;g++){
		mm[g] = hmm->mm[g];
		mc[g] = hmm->mc[g];
		mr[g] = hmm->mr[g];
		mgx[g] = hmm->mgx[g];
		mgy[g] = hmm->mgy[g];
		
		gxgx[g] = hmm->gxgx[g];
		gxm[g] = hmm->gxm[g];
		gxc[g] = hmm->gxc[g];
		
		gygy[g] = hmm->gygy[g];
		gym[g] = hmm->gym[g];
		gyc[g] = hmm->gyc[g];
		
		cc[g] = hmm->cc[g];
		cm[g] = hmm->cm[g];
		cr[g] = hmm->cr[g];
		cgx[g] = hmm->cgx[g];
		cgy[g] = hmm->cgy[g];
		
		//rr[g]  = hmm->rr[g];
		//rmc[g] = hmm->rmc[g];
		rr[g] = prob2scaledprob((float)m / (float)(m+1));
		rmc[g] = prob2scaledprob((float)0.5f / (float)(m+1));
	}
	/*for(i = 0; i < 5;i++){
		rey[i] =0;
	}
	for(i = 0; i < m;i++){
		rey[b[i]]++;
	}
	for(i = 0; i < 5;i++){
		rey[i] = prob2scaledprob( (float) rey[i] / (float)m  );
	}*/
	
	float** M = hmm->bM;
	
	float** C = hmm->bC;
	
	float** X = hmm->bX;
	float** Y = hmm->bY;
	float** R = hmm->bR;
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	
	M[n][m] = -SCALEINFTY;//TAU_LOG;
	C[n][m] = -SCALEINFTY;//TAU_LOG;
	X[n][m] = -SCALEINFTY;//TAU_LOG; 
	Y[n][m] = -SCALEINFTY;//TAU_LOG;
	R[n][m] = TAU_LOG; 
	
	for(i = n-1; i >= 1;i--){
		M[i][m] = -SCALEINFTY;//	mgx + hmm->ex[seqa[i+1]] + X[i+1][m] ;
		C[i][m] = -SCALEINFTY;//	cgx + hmm->ex[seqa[i+1]] + X[i+1][m] ;
		X[i][m] = -SCALEINFTY;	//gxgx + hmm->ex[seqa[i+1]] + X[i+1][m];
		Y[i][m] = -SCALEINFTY;	
		R[i][m] = -SCALEINFTY;	
		
	}
	if(!strand){
		g  = REGIONS -1;
	}else{
		g = 0;
	}
	for(j = m-1; j >= 1;j--){
		M[n][j] = mr[g] + rey[seqb[j+1]] + R[n][j+1];//-9999999;	//mgy + hmm->ey[seqb[j+1]] + Y[n][j+1] ;
		C[n][j] = cr[g] +  rey[seqb[j+1]] + R[n][j+1];//-9999999;//cgy + hmm->ey[seqb[j+1]] + Y[n][j+1] ;
		X[n][j] = -SCALEINFTY;
		Y[n][j] = -SCALEINFTY;//logsum(gygy + hmm->ey[seqb[j+1]] + Y[n][j+1], rgy + rey[seqb[j+1]] + R[n][j+1]);
		R[n][j] = rr[g] + rey[seqb[j+1]] + R[n][j+1];//,gyr + hmm->ey[seqb[j+1]]  + Y[n][j+1]) ;
	}
	
	for(i = n-1; i >= 1;i--){
		g = (int) floor((float)(i-1) / (float)n * (float)REGIONS); 
		if(strand){
			g = REGIONS -(1+ g);
		}
		for(j = m-1; j >= 1;j--){
			M[i][j] = logsum(
					 logsum (
						 logsum(
							mm[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + M[i+1][j+1],
							mc[g] +  hmm->ec[g][seqa[i+1]][seqb[j+1]] + C[i+1][j+1]
							),
						 mr[g] + rey[seqb[j+1]] + R[i][j+1]
						 
						 ) , 
					 logsum(
						mgx[g] + hmm->ex[g][seqa[i+1]] + X[i+1][j] , 
						mgy[g] + hmm->ey[g][seqb[j+1]] + Y[i][j+1]
						)
					 );
			
			C[i][j] = logsum(
					 logsum (
						 logsum(
							cm[g] +  hmm->em[g][seqa[i+1]][seqb[j+1]] + M[i+1][j+1],
							cc[g] + hmm->ec[g][seqa[i+1]][seqb[j+1]] + C[i+1][j+1]
							),
						 cr[g] +  rey[seqb[j+1]] + R[i][j+1]
						 ),
					 logsum(
						cgx[g] + hmm->ex[g][seqa[i+1]] + X[i+1][j],
						cgy[g] + hmm->ey[g][seqb[j+1]] + Y[i][j+1]
						)
					 );
			
			X[i][j] = logsum(
					 logsum(
						gxm[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + M[i+1][j+1],
						gxc[g] + hmm->ec[g][seqa[i+1]][seqb[j+1]] + C[i+1][j+1]
						
						),
					 gxgx[g] + hmm->ex[g][seqa[i+1]] + X[i+1][j]
					 ); 
			
			Y[i][j] = logsum(
					 logsum(
						gym[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + M[i+1][j+1] ,
						gyc[g] + hmm->ec[g][seqa[i+1]][seqb[j+1]] + C[i+1][j+1]
						),
					 // logsum(
					 gygy[g] + hmm->ey[g][seqb[j+1]] + Y[i][j+1]
					 //		rgy + rey[seqb[j+1]] + R[i][j+1]
					 //)
					 
					 );
			
			R[i][j] = logsum(
					 rr[g] +  rey[seqb[j+1]] + R[i][j+1],
					 logsum(
						rmc[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + M[i+1][j+1],
						rmc[g] + hmm->ec[g][seqa[i+1]][seqb[j+1]] + C[i+1][j+1]
						)
					 );
			
		}
	}
	return hmm;
}



struct hmm* get_prob_log(struct hmm* hmm, float* rey,unsigned char* a, unsigned char* b, int n,int m,float frac,int strand)
{
	int i,j,g;
	unsigned char* seqa = a -1;
	unsigned char* seqb = b -1;
	
	for(i = 1; i <= n;i++){
		
		g = (int) floor((float)(i-1) / (float)n * (float)REGIONS);
		if(strand){
			g = REGIONS - (1+ g);
		}
		for (j = 1; j <= m;j++){
			hmm->tem[g][seqa[i]][seqb[j]] = logsum(hmm->tem[g][seqa[i]][seqb[j]] , hmm->fM[i][j] + hmm->bM[i][j]   -hmm->score +frac );// * 1000000; 
			hmm->tec[g][seqa[i]][seqb[j]] = logsum(hmm->tec[g][seqa[i]][seqb[j]] , hmm->fC[i][j] + hmm->bC[i][j]-hmm->score +frac);
			hmm->tex[g][seqa[i]] = logsum(hmm->tex[g][seqa[i]], hmm->fX[i][j] + hmm->bX[i][j]-hmm->score +frac);// / hmm->score; 
			hmm->tey[g][seqb[j]] = logsum(hmm->tey[g][seqb[j]], hmm->fY[i][j] + hmm->bY[i][j]-hmm->score +frac);// / hmm->score; 
		}
	}
	
	for(i = 1; i < n;i++){
		g = (int) floor((float)(i-1) / (float)n * (float)REGIONS);
		if(strand){
			g = REGIONS - (1+ g);
		}
		for (j = 1; j < m;j++){
			hmm->tmm[g] = logsum(hmm->tmm[g],hmm->fM[i][j] + hmm->mm[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + hmm->bM[i+1][j+1] - hmm->score + frac);
			hmm->tmc[g] = logsum(hmm->tmc[g],hmm->fM[i][j] + hmm->mc[g] + hmm->ec[g][seqa[i+1]][seqb[j+1]] + hmm->bC[i+1][j+1] - hmm->score + frac );
			
			//hmm->tmr[g] = logsum(hmm->tmr[g],hmm->fM[i][j] + hmm->mr[g] + rey[seqb[j+1]] + hmm->bR[i][j+1] - hmm->score+ frac );
			
			hmm->tmgx[g] = logsum(hmm->tmgx[g],hmm->fM[i][j] + hmm->mgx[g] + hmm->ex[g][seqa[i+1]] + hmm->bX[i+1][j] - hmm->score +frac);
			hmm->tmgy[g] = logsum(hmm->tmgy[g],hmm->fM[i][j] + hmm->mgy[g] + hmm->ey[g][seqb[j+1]] + hmm->bY[i][j+1] - hmm->score +frac);
			
			hmm->tcc[g] = logsum(hmm->tcc[g],hmm->fC[i][j] + hmm->cc[g] + hmm->ec[g][seqa[i+1]][seqb[j+1]] + hmm->bC[i+1][j+1] - hmm->score +frac );
			hmm->tcm[g] = logsum(hmm->tcm[g],hmm->fC[i][j] + hmm->cm[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + hmm->bM[i+1][j+1] - hmm->score +frac);
			
			//hmm->tcr[g] = logsum(hmm->tcr[g],hmm->fC[i][j] + hmm->cr[g] + rey[seqb[j+1]] + hmm->bR[i][j+1] - hmm->score+frac);
			
			hmm->tcgx[g] = logsum(hmm->tcgx[g],hmm->fC[i][j] + hmm->cgx[g] + hmm->ex[g][seqa[i+1]] + hmm->bX[i+1][j] - hmm->score + frac);
			hmm->tcgy[g] = logsum(hmm->tcgy[g],hmm->fC[i][j] + hmm->cgy[g] + hmm->ey[g][seqb[j+1]] + hmm->bY[i][j+1] - hmm->score + frac);
			
			
			
			hmm->tgxgx[g] = logsum(hmm->tgxgx[g],hmm->fX[i][j] + hmm->gxgx[g] + hmm->ex[g][seqa[i+1]] + hmm->bX[i+1][j] - hmm->score +frac );
			hmm->tgxm[g] = logsum(hmm->tgxm[g],hmm->fX[i][j] + hmm->gxm[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + hmm->bM[i+1][j+1] - hmm->score +frac);
			hmm->tgxc[g] = logsum(hmm->tgxc[g],hmm->fX[i][j] + hmm->gxc[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + hmm->bC[i+1][j+1] - hmm->score +frac );
			
			hmm->tgygy[g] = logsum(hmm->tgygy[g],hmm->fY[i][j] + hmm->gygy[g] + hmm->ey[g][seqb[j+1]] + hmm->bY[i][j+1] - hmm->score);
			hmm->tgym[g] = logsum(hmm->tgym[g],hmm->fY[i][j] + hmm->gym[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + hmm->bM[i+1][j+1] - hmm->score +frac );
			hmm->tgyc[g] = logsum(hmm->tgyc[g],hmm->fY[i][j] + hmm->gyc[g] + hmm->em[g][seqa[i+1]][seqb[j+1]] + hmm->bC[i+1][j+1] - hmm->score +frac );
			
			//hmm->tgyr = logsum(hmm->tgyr,hmm->fY[i][j] +  hmm->gyr +  rey[seqb[j+1]] + hmm->bR[i][j+1]- hmm->score); 
			//hmm->trr = logsum(hmm->trr,  hmm->fR[i][j] + hmm->rr + rey[seqb[j+1]] + hmm->bR[i][j+1]- hmm->score);
			//hmm->trgy = logsum(hmm->trgy,hmm->fR[i][j] + hmm->rgy + hmm->ey[seqb[j+1]] + hmm->bY[i][j+1] - hmm->score);
		}
	}
	return hmm;
}

struct hmm* re_estimate(struct hmm* hmm)
{
	float sum = 0.0f;
	int i,j,g;
	
	for(g = 0; g < REGIONS;g++){
		
		hmm->mm[g] =  prob2scaledprob(0.96f) + (hmm->tmm[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], hmm->tmgy[g])));// + NOT_TAU_LOG;
		hmm->mgx[g] = prob2scaledprob(0.96f) + (hmm->tmgx[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], hmm->tmgy[g])));// + NOT_TAU_LOG;
		hmm->mgy[g] = prob2scaledprob(0.96f) + (hmm->tmgy[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], hmm->tmgy[g])));// + NOT_TAU_LOG;
		hmm->mc[g] = prob2scaledprob(0.96f) + (hmm->tmc[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], hmm->tmgy[g])));// + NOT_TAU_LOG;
		
		hmm->cc[g] = prob2scaledprob(0.96f) +(hmm->tcc[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],hmm->tcm[g])));// + NOT_TAU_LOG;
		hmm->cm[g] = prob2scaledprob(0.96f) +(hmm->tcm[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],hmm->tcm[g])));//+ NOT_TAU_LOG;
		hmm->cgx[g] = prob2scaledprob(0.96f) +(hmm->tcgx[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],hmm->tcm[g])));// + NOT_TAU_LOG;
		hmm->cgy[g] = prob2scaledprob(0.96f) +(hmm->tcgy[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],hmm->tcm[g])));// + NOT_TAU_LOG;
		
		
		
		
		/*hmm->mm[g] =  hmm->tmm[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], logsum(hmm->tmgy[g], hmm->tmr[g]) ));// + NOT_TAU_LOG;
		 hmm->mgx[g] = hmm->tmgx[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], logsum(hmm->tmgy[g], hmm->tmr[g])));// + NOT_TAU_LOG;
		 hmm->mgy[g] = hmm->tmgy[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], logsum(hmm->tmgy[g], hmm->tmr[g])));// + NOT_TAU_LOG;
		 hmm->mc[g] = hmm->tmc[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], logsum(hmm->tmgy[g], hmm->tmr[g])));// + NOT_TAU_LOG;
		hmm->mr[g] = hmm->tmr[g] - logsum(logsum( hmm->tmm[g],hmm->tmc[g]) ,logsum( hmm->tmgx[g], logsum(hmm->tmgy[g], hmm->tmr[g])));// + NOT_TAU_LOG;
		 
		 hmm->cc[g] = hmm->tcc[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],logsum(hmm->tcm[g],hmm->tcr[g])   ));// + NOT_TAU_LOG;
		 hmm->cm[g] = hmm->tcm[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],logsum(hmm->tcm[g],hmm->tcr[g]) ));//+ NOT_TAU_LOG;
		 hmm->cgx[g] = hmm->tcgx[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],logsum(hmm->tcm[g],hmm->tcr[g]) ));// + NOT_TAU_LOG;
		 hmm->cgy[g] = hmm->tcgy[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],logsum(hmm->tcm[g],hmm->tcr[g]) ));// + NOT_TAU_LOG;
		 hmm->cr[g] = hmm->tcr[g] - logsum(logsum(hmm->tcgx[g],hmm->tcgy[g]),logsum(hmm->tcc[g],logsum(hmm->tcm[g],hmm->tcr[g]) ));// + NOT_TAU_LOG;
		*/ 
		
		hmm->gxgx[g] = (hmm->tgxgx[g] - logsum(logsum(hmm->tgxgx[g],hmm->tgxm[g]),hmm->tgxc[g]));// + NOT_TAU_LOG;
		hmm->gxm[g] = (hmm->tgxm[g] -  logsum(logsum(hmm->tgxgx[g],hmm->tgxm[g]),hmm->tgxc[g]));//+ NOT_TAU_LOG;
		hmm->gxc[g] = (hmm->tgxc[g] -  logsum(logsum(hmm->tgxgx[g],hmm->tgxm[g]),hmm->tgxc[g])); //+ NOT_TAU_LOG;
		
		hmm->gygy[g] = (hmm->tgygy[g] -  logsum(logsum(hmm->tgygy[g],hmm->tgym[g]),hmm->tgyc[g]));// + NOT_TAU_LOG;
		hmm->gym[g] = (hmm->tgym[g] -  logsum(logsum(hmm->tgygy[g],hmm->tgym[g]),hmm->tgyc[g]));// + NOT_TAU_LOG;
		hmm->gyc[g] = (hmm->tgyc[g] -  logsum(logsum(hmm->tgygy[g],hmm->tgym[g]),hmm->tgyc[g]));//+ NOT_TAU_LOG;
		
		hmm->tmm[g] = -SCALEINFTY;
		hmm->tmc[g] = -SCALEINFTY;
		hmm->tmr[g] = -SCALEINFTY;
		hmm->tmgx[g] = -SCALEINFTY;
		hmm->tmgy[g] = -SCALEINFTY;
		
		hmm->tcc[g] = -SCALEINFTY;
		hmm->tcm[g] = -SCALEINFTY;
		hmm->tcr[g] = -SCALEINFTY;
		hmm->tcgx[g] = -SCALEINFTY;
		hmm->tcgy[g] = -SCALEINFTY;
		
		hmm->tgxgx[g] = -SCALEINFTY;
		hmm->tgxm[g] = -SCALEINFTY;
		hmm->tgxc[g] = -SCALEINFTY;
		
		hmm->tgygy[g] = -SCALEINFTY;
		hmm->tgym[g] = -SCALEINFTY;
		hmm->tgyc[g] = -SCALEINFTY;
		
		hmm->tgyr[g] = -SCALEINFTY;
		
		hmm->trr[g] = -SCALEINFTY;
		hmm->trgy[g] = -SCALEINFTY;
	}
	
	
	for(g = 0; g < REGIONS;g++){
		
		sum = hmm->tex[g][0];//-99999999;
		
		for(i = 1; i < 5;i++){
			sum =  logsum(sum , hmm->tex[g][i]);
		}
		//fprintf(stderr,"tex:%f\n",sum);
		for(i = 0; i < 5;i++){
			//	fprintf(stderr,"%d	%f	%f\n",i,tex[i],log(tex[i]/ sum));
			hmm->ex[g][i] = (hmm->tex[g][i] - sum);
			hmm->tex[g][i] = -SCALEINFTY;
		}
	}
	
	for(g = 0; g <REGIONS;g++){
		sum = hmm->tey[g][0];//-99999999;
		
		for(i = 1; i < 5;i++){
			sum =  logsum(sum ,hmm->tey[g][i]);
		}
		//fprintf(stderr,"tey:%f\n",sum);
		for(i = 0; i < 5;i++){
			hmm->ey[g][i] = (hmm->tey[g][i] - sum);
			hmm->tey[g][i] = -SCALEINFTY;
		}
	}
	for(g = 0; g <REGIONS;g++){
		sum = prob2scaledprob(0.0f);
		for(i = 0; i < 5;i++){
			for(j = 0; j < 5;j++){
				sum = logsum(sum , hmm->tem[g][i][j]);
			}
		}
		for(i = 0; i < 5;i++){
			for(j = 0; j < 5;j++){
				hmm->em[g][i][j] = (hmm->tem[g][i][j] - sum );
				hmm->tem[g][i][j] =  -SCALEINFTY;//prob2scaledprob(FLT_MIN);
			}
		}
	}
	for(g = 0; g <REGIONS;g++){
		sum = prob2scaledprob(0.0f);
		for(i = 0; i < 5;i++){
			for(j = 0; j < 5;j++){
				sum = logsum(sum , hmm->tec[g][i][j]);
			}
		}
		for(i = 0; i < 5;i++){
			for(j = 0; j < 5;j++){
				hmm->ec[g][i][j] = (hmm->tec[g][i][j] - sum ) ;
				hmm->tec[g][i][j] =  -SCALEINFTY;//prob2scaledprob(FLT_MIN);
			}
		}
	}
	
	//// MOD 4x4
	
	/*
	for(g = 0; g <REGIONS;g++){
		
		for(i = 0; i < 5;i++){
			sum = prob2scaledprob(0.0f);
			for(j = 0; j < 5;j++){
				sum = logsum(sum , hmm->tem[g][i][j]);
			}
			for(j = 0; j < 5;j++){
				hmm->em[g][i][j] = (hmm->tem[g][i][j] - sum );
				hmm->tem[g][i][j] =  -SCALEINFTY;//prob2scaledprob(FLT_MIN);
			}
			sum = prob2scaledprob(0.0f);
			for(j = 0; j < 5;j++){
				sum = logsum(sum , hmm->tec[g][i][j]);
			}
			for(j = 0; j < 5;j++){
				hmm->ec[g][i][j] = (hmm->tec[g][i][j] - sum ) ;
				hmm->tec[g][i][j] =  -SCALEINFTY;//prob2scaledprob(FLT_MIN);
			}
		}
	}*/
	
	
	return hmm;
}



void write_hmm_to_file(struct hmm* hmm,struct parameters * param)
{
	FILE* hmmfile; 
	int i,j,g;
	if ((hmmfile = fopen(param->hmm_file, "wb")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	for(g = 0;g < REGIONS;g++){
	
		fwrite(&hmm->mm[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->mc[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->mr[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->mgx[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->mgy[g], sizeof(float),1,hmmfile);
		
		fwrite(&hmm->cc[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->cm[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->cr[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->cgx[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->cgy[g], sizeof(float),1,hmmfile);
		
		fwrite(&hmm->gxgx[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->gxm[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->gxc[g], sizeof(float),1,hmmfile);
		
		fwrite(&hmm->gygy[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->gym[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->gyc[g], sizeof(float),1,hmmfile);
		
		fwrite(&hmm->rr[g], sizeof(float),1,hmmfile);
		fwrite(&hmm->rmc[g], sizeof(float),1,hmmfile);
		
		for(i = 0; i < 5;i++){
			for(j = 0;j < 5;j++){
				fwrite(&hmm->em[g][i][j], sizeof(float),1,hmmfile);
			}
		}
		
		for(i = 0; i < 5;i++){
			for(j = 0;j < 5;j++){
				fwrite(&hmm->ec[g][i][j], sizeof(float),1,hmmfile);
			}
		}
		
		for(i = 0; i < 5;i++){
			fwrite(&hmm->ex[g][i], sizeof(float),1,hmmfile);
		}
		
		for(i = 0; i < 5;i++){
			fwrite(&hmm->ey[g][i], sizeof(float),1,hmmfile);
		}
	}
		
	fclose(hmmfile);
}

void read_hmm_from_file(struct hmm* hmm,struct parameters * param)
{
	FILE* hmmfile; 
	int i,j,g;
	//fread(seed_parameters , sizeof(int), 4, file);
	/*file = fopen(param->infile[1], "rb");*/
	if ((hmmfile = fopen(param->hmm_file, "rb")) == NULL){
		fprintf(stderr,"can't open hmm file:%s \n",param->hmm_file);
		exit(-1);
	}
	for(g = 0;g < REGIONS;g++){

		fread(&hmm->mm[g], sizeof(float),1,hmmfile);
		fread(&hmm->mc[g], sizeof(float),1,hmmfile);
		fread(&hmm->mr[g], sizeof(float),1,hmmfile);
		fread(&hmm->mgx[g], sizeof(float),1,hmmfile);
		fread(&hmm->mgy[g], sizeof(float),1,hmmfile);
		
		fread(&hmm->cc[g], sizeof(float),1,hmmfile);
		fread(&hmm->cm[g], sizeof(float),1,hmmfile);
		fread(&hmm->cr[g], sizeof(float),1,hmmfile);
		fread(&hmm->cgx[g], sizeof(float),1,hmmfile);
		fread(&hmm->cgy[g], sizeof(float),1,hmmfile);
		
		fread(&hmm->gxgx[g], sizeof(float),1,hmmfile);
		fread(&hmm->gxm[g], sizeof(float),1,hmmfile);
		fread(&hmm->gxc[g], sizeof(float),1,hmmfile);
		
		fread(&hmm->gygy[g], sizeof(float),1,hmmfile);
		fread(&hmm->gym[g], sizeof(float),1,hmmfile);
		fread(&hmm->gyc[g], sizeof(float),1,hmmfile);
		
		fread(&hmm->rr[g], sizeof(float),1,hmmfile);
		fread(&hmm->rmc[g], sizeof(float),1,hmmfile);
		
		for(i = 0; i < 5;i++){
			for(j = 0;j < 5;j++){
				fread(&hmm->em[g][i][j], sizeof(float),1,hmmfile);
			}
		}
		
		for(i = 0; i < 5;i++){
			for(j = 0;j < 5;j++){
				fread(&hmm->ec[g][i][j], sizeof(float),1,hmmfile);
			}
		}
		
		for(i = 0; i < 5;i++){
			fread(&hmm->ex[g][i], sizeof(float),1,hmmfile);
		}
		
		for(i = 0; i < 5;i++){
			fread(&hmm->ey[g][i], sizeof(float),1,hmmfile);
		}
	}
	fclose(hmmfile);
}

void print_hmm_param(struct hmm* hmm)
{
	int i,j,g;
	
	char nuc[5] = "ACGTN";
	float sanity1 = 0.0f; 
	//float sanity2 = 0.0f; 

	
	fprintf(stdout,"\nTransition probabilities:\n\n");
	
	fprintf(stdout,"        ");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"R%d	",g+1);
	}
	fprintf(stdout,"\n");
	
	fprintf(stdout,"m->m   :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->mm[g]));
	}
	fprintf(stdout,"\n");
	
	fprintf(stdout,"m->c   :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->mc[g]));
	}
	fprintf(stdout,"\n");
	
	fprintf(stdout,"m->r   :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->mr[g]));
	}
	fprintf(stdout,"\n");
	
	fprintf(stdout,"m->gx  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->mgx[g]));
	}
	fprintf(stdout,"\n");
	
	fprintf(stdout,"m->gy  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->mgy[g]));
	}
	fprintf(stdout,"\n\n");
	//fprintf(stderr,"Sanity:%f\n",scaledprob2prob(hmm->mm)+scaledprob2prob(hmm->mc)+scaledprob2prob(hmm->mr)+scaledprob2prob(hmm->mgx)+scaledprob2prob(hmm->mgy));
	//fprintf(stderr,"Sanity:%f\n\n",scaledprob2prob( logsum(logsum(logsum(hmm->mm,hmm->mc),logsum(hmm->mr,hmm->mgx)), hmm->mgy)));
		
	fprintf(stdout,"c->c   :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->cc[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"c->m   :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->cm[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"c->r   :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->cr[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"c->gx  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->cgx[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"c->gy  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->cgy[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"\n");
	//fprintf(stderr,"Sanity:%f\n\n",scaledprob2prob(hmm->cc)+scaledprob2prob(hmm->cm)+scaledprob2prob(hmm->cr)+scaledprob2prob(hmm->cgx)+scaledprob2prob(hmm->cgy));
	//fprintf(stderr,"Sanity:%f\n\n",scaledprob2prob( logsum(logsum(logsum(hmm->cc,hmm->cm),logsum(hmm->cr,hmm->cgx)), hmm->cgy)));
	
	fprintf(stdout,"gx->gx :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->gxgx[g]));
	}
	fprintf(stdout,"\n");

	fprintf(stdout,"gx->m  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->gxm[g]));
	}
	fprintf(stdout,"\n");

	fprintf(stdout,"gx->c  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->gxc[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"\n");
	//fprintf(stderr,"Sanity :%f\n\n",scaledprob2prob(hmm->gxgx)+scaledprob2prob(hmm->gxm)+scaledprob2prob(hmm->gxc));
	
	fprintf(stdout,"gy->gy :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->gygy[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"gy->m  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->gym[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"gy->c  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->gyc[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"\n");
	//fprintf(stderr,"gy->r  :%f	(%f)\n",scaledprob2prob(hmm->gyr)*100.0f,hmm->gyr);
	//fprintf(stderr,"Sanity :%f\n\n",scaledprob2prob(hmm->gygy)+scaledprob2prob(hmm->gym)+scaledprob2prob(hmm->gyc));
	
	fprintf(stdout,"r->r   :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->rr[g]));
	}
	fprintf(stdout,"\n");
	//fprintf(stderr,"r->gy  :%f	(%f)\n",scaledprob2prob(hmm->rgy)*100.0f,hmm->rgy);
	fprintf(stdout,"r->mc  :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(hmm->rmc[g]));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"r->end :");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"%0.4f\t",scaledprob2prob(TAU_LOG));
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"\n");
	
	//fprintf(stderr,"Sanity :%f\n\n",scaledprob2prob(hmm->rr)+scaledprob2prob(hmm->rmc)+scaledprob2prob(hmm->rmc) + scaledprob2prob(TAU_LOG));
	fprintf(stdout,"Emission probabilities:\n\n");
	fprintf(stdout,"M State:\n");
	fprintf(stdout,"R/G	");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"R%d	",g+1);
	}
	fprintf(stdout,"\n");
	for(i = 0; i < 5;i++){
		for(j = 0;j < 5;j++){
			fprintf(stdout,"%c%c	",nuc[i],nuc[j]);
			for(g = 0; g < REGIONS;g++){
				fprintf(stdout,"%0.4f	",scaledprob2prob (hmm->em[g][i][j]));
			}
			fprintf(stdout,"\n");
			//sanity1 += scaledprob2prob(hmm->em[i][j]);
			//sanity2 += scaledprob2prob(hmm->ec[i][j]);
		}
		//fprintf(stderr,"\n");
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"C State:\n");
	fprintf(stdout,"R/G	");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"R%d	",g+1);
	}
	fprintf(stdout,"\n");
	
	for(i = 0; i < 5;i++){
		for(j = 0;j < 5;j++){
			fprintf(stdout,"%c%c	",nuc[i],nuc[j]);
			for(g = 0; g < REGIONS;g++){
				fprintf(stdout,"%0.4f	",scaledprob2prob (hmm->ec[g][i][j]));
			}
			fprintf(stdout,"\n");
			//sanity1 += scaledprob2prob(hmm->em[i][j]);
			//sanity2 += scaledprob2prob(hmm->ec[i][j]);
		}
		//fprintf(stderr,"\n");
	}
	fprintf(stdout,"\n");
	fprintf(stdout,"\n");
	//fprintf(stderr,"Sanity:%f	%f\n",sanity1,sanity2);
		
	
	sanity1 = 0.0f;
	fprintf(stdout,"X State:\n");
	fprintf(stdout,"R/G	");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"R%d	",g+1);
	}
	fprintf(stdout,"\n");
	
	for(i = 0; i < 5;i++){
		fprintf(stdout,"%c-	",nuc[i]);
		for(g = 0; g < REGIONS;g++){
			fprintf(stdout,"%0.4f  ",scaledprob2prob(hmm->ex[g][i]));
		}
		fprintf(stdout,"\n");
		//sanity1 += scaledprob2prob(hmm->ex[i]);
	}
	fprintf(stdout,"\n");
	//fprintf(stderr,"Sanity:%f\n",sanity1);
	sanity1 = 0.0f;
	fprintf(stdout,"Y State:\n");
	fprintf(stdout,"R/G	");
	for(g = 0; g < REGIONS;g++){
		fprintf(stdout,"R%d	",g+1);
	}
	fprintf(stdout,"\n");
	for(i = 0; i < 5;i++){
		fprintf(stdout,"-%c	",nuc[i]);
		for(g = 0; g < REGIONS;g++){
			fprintf(stdout,"%0.4f  ",scaledprob2prob(hmm->ey[g][i]));
		}
		fprintf(stdout,"\n");
		//sanity1 += scaledprob2prob(hmm->ey[i]);
	}
	fprintf(stdout,"\n");
	//fprintf(stderr,"Sanity:%f\n",sanity1);

}



