/*
 
 Copyright (C) 2010,2011 Timo Lassmann <timolassmann@gmail.com>
 
 This file is part of delve.
 
 Delve is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 Delve is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with Delve.  If not, see <http://www.gnu.org/licenses/>.
 
 */

#include "delve.h"
#include "interface.h"
#include "dbio.h"

#include "pHMM.h"
#include "nuc_code.h"
#include "seed.h"
#include "core.h"
#include "scoreboard.h"
#include "read_sam.h"
#include "align_to_sam.h"
#include <ctype.h>
#include "align.h"

int phmm_controller(struct db* db,struct parameters* param)
{
	struct read_info** ri = 0;
	FILE* file = 0;
	int i,numseq,iter = 0;
	int total_numseq = 0;
	init_logsum();
	struct hmm* hmm = init_hmm(500, 500);
	
	for(i = 0;i < 5;i++){
		hmm->rey[i] = prob2scaledprob(db->seq_info->nuc_counts[i] );
	}
	
	struct rescue_probabilities* rp = 0;
	
	if(param->local_weighting){
		rp =  init_rp(db->seq_info->total_len ,param->local_weighting);
	}
	
	ri = malloc(sizeof(struct read_info*) * param->num_query);
	
	for(i = 0; i < param->num_query;i++){
		ri[i] = malloc(sizeof(struct read_info));
		ri[i]->seq = 0;
		ri[i]->name = 0;
		ri[i]->qual = 0;
		ri[i]->len = 0;
		ri[i]->priors = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->strand = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->hits = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->identity = malloc(sizeof(float)* (LIST_STORE_SIZE+1));
	}
	
	/*if(!isatty(0)){
		file = stdin;
	}else{
	*/
		if (!(file = fopen(param->infile[1] , "r" ))){
			fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[1]);
			exit(-1);
		}
	//}
	
	if(param->hmm_file &&  access( param->hmm_file, R_OK) != -1 ) {
		fprintf(stderr,"Reading Model\n");
		read_hmm_from_file(hmm,param);
	}else {
		if(param->local_weighting){
			for(iter = 0; iter < param->iterations;iter++){ 
				total_numseq = 0;
				while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
					fprintf(stderr,"Reading\n");
					//ri = add_k(ri, db,numseq);
					if(iter > 2){
						ri = set_local_prob(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq);
					}
					hmm = run_pHMM(hmm,ri, db,param,numseq);
					if(iter > 2){
						rp = update_rp(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq,0, 1000000.0f);
					}
					total_numseq+= numseq;
					if(total_numseq > param->training_sequences){
						break;
					}
				}
				rewind(file);
				hmm = re_estimate(hmm);
				if(iter > 2){
					rp = update_rp(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq,1,1000000.0f);
				}
				fprintf(stderr,"Iter:%d\n",iter);
			}
		}else{
			for(iter = 0; iter < param->iterations;iter++){ 
				total_numseq = 0;
				while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
					fprintf(stderr,"Reading\n");
					
					ri = add_k(ri, db,numseq);
					hmm = run_pHMM(hmm,ri, db,param,numseq);
					total_numseq+= numseq;
					if(total_numseq > param->training_sequences){
						break;
					}
				}
				
				rewind(file);
				hmm = re_estimate(hmm);
				fprintf(stderr,"Iter:%d\n",iter);
			}
		}
		
		if(param->hmm_file){
			write_hmm_to_file(hmm,param);
		}
		rewind(file);
	}
	
	print_samheader(db,param);
	
	
	while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
		ri = add_k(ri, db,numseq);
		if(param->local_weighting){
			ri = set_local_prob(rp, ri,db->seq_info->total_len ,param->local_weighting,numseq);
		}
		
		
		ri = run_score_alignments(hmm,ri,db,param,numseq);
	}
	
	
	if(param->local_weighting){
		free_rp(rp);
	}
	for(i = 0; i < param->num_query;i++){
		
		free(ri[i]->strand); 
		free(ri[i]->hits);
		free(ri[i]->identity);
		free(ri[i]);
	}
	free_hmm(hmm);
	free(ri);

	//if(param->infile[1]){
		fclose(file);
	//}
	return 1;
}

struct rescue_probabilities* init_rp(unsigned int len,unsigned int block_len)
{
	int i;
	struct rescue_probabilities* rp = malloc(sizeof(struct rescue_probabilities));
	rp->plus = malloc(sizeof(float) *((len >> block_len) + 1u));
	rp->minus = malloc(sizeof(float)  *((len >> block_len) + 1u));
	
	rp->t_plus = malloc(sizeof(float) *((len >> block_len) + 1u));
	rp->t_minus = malloc(sizeof(float)  *((len >> block_len) + 1u));
	
	float add = prob2scaledprob((0.5f / (float)((len >> block_len )+ 1u))); 

	for(i = 0; i < (len >> block_len) + 1u;i++){
		rp->plus[i] = add;
		rp->minus[i] = add;
		rp->t_plus[i] = add;
		rp->t_minus[i] = add;
		
	}
	rp->total = prob2scaledprob(1.0f);
	rp->block_size = block_len;
	return rp;
}

void free_rp(struct rescue_probabilities* rp)
{	
	free(rp->plus);
	free(rp->minus);
	free(rp->t_plus);
	free(rp->t_minus);
	free(rp); 
}

struct read_info** set_local_prob(struct rescue_probabilities* rp, struct read_info** ri,unsigned int len,unsigned int block_len,int n)
{
	int i,j,hit;
	unsigned int pos1 = 0;
	unsigned int pos2 = 0;

	for(i = 0;i < n;i++){
		hit = 0;
		while(ri[i]->identity[hit] != -2.0f){ // option
			pos1 = ri[i]->hits[hit] >> block_len;
			
			pos2 = (ri[i]->hits[hit]+ri[i]->len) >> block_len;
			if(!ri[i]->strand[hit]){
				ri[i]->priors[hit] = rp->plus[pos1];
				for(j = pos1+1; j <= pos2;j++){
					ri[i]->priors[hit]  = logsum(ri[i]->priors[hit] ,rp->plus[j]);
				}
				ri[i]->priors[hit]  = ri[i]->priors[hit]  - prob2scaledprob((pos2 - pos1) +1);
				
			}else{
				
				ri[i]->priors[hit] = rp->minus[pos1];
				for(j = pos1+1; j <= pos2;j++){
					ri[i]->priors[hit]  = logsum(ri[i]->priors[hit] ,rp->minus[j]);
				}
				ri[i]->priors[hit]  = ri[i]->priors[hit]  - prob2scaledprob((pos2 - pos1) +1);
				
				//ri[i]->priors[hit] = rp->minus[pos];
			}
			hit++;
		}
	}
	return ri;
}

struct rescue_probabilities* update_rp(struct rescue_probabilities* rp, struct read_info** ri,unsigned int len,unsigned int block_len,int n,int up, float lib_size)
{
	int i,hit;
	unsigned int pos = 0;
	//unsigned int pos2 = 0;
	//float add = prob2scaledprob((lib_size / (float)((len >> block_len) + 1u))); 
	float add = prob2scaledprob((0.5f / (float)((len >> block_len) + 1u))); 
	
	if(!up){
		for(i = 0;i < n;i++){
			hit = 0;
			while(ri[i]->identity[hit] != -2.0f){ // option
				//if(ri[i]->identity[hit] >= 0.95){
				//pos1 = ri[i]->hits[hit] >> block_len;
				
				//pos2 = (ri[i]->hits[hit]+ri[i]->len) >> block_len;
				pos = ri[i]->hits[hit] >> block_len;
				if(!ri[i]->strand[hit]){
					//for(j = pos1+1; j <= pos2;j++){
					//	rp->t_plus[j] = logsum(rp->t_plus[j], ri[i]->priors[hit] - prob2scaledprob(pos2 - pos1 + 1  ) );
					//}
					
					rp->t_plus[pos] = logsum(rp->t_plus[pos], ri[i]->priors[hit]);
				}else{
					//for(j = pos1+1; j <= pos2;j++){
					//	rp->t_minus[j] = logsum(rp->t_minus[j], ri[i]->priors[hit] - prob2scaledprob(pos2 - pos1 + 1  ) );
					//}
					
					rp->t_minus[pos] = logsum(rp->t_minus[pos], ri[i]->priors[hit]);
				}
				rp->total = logsum(rp->total, ri[i]->priors[hit] );
				//}
				hit++;
			}
		}
	}else{
		//rp->total = logsum(rp->total,  prob2scaledprob(PSEUDO_CONSTANT_GENOME));
		for(i = 0; i < (len >> block_len) + 1u;i++){
			//rp->t_plus[i]  = logsum(rp->t_plus[i] ,rp->plus[i] +  prob2scaledprob(PSEUDO_CONSTANT_GENOME));
			//rp->t_minus[i]  = logsum(rp->t_minus[i] ,rp->minus[i] +  prob2scaledprob(PSEUDO_CONSTANT_GENOME));
			
			rp->plus[i] =  rp->t_plus[i]  - rp->total ;
			rp->minus[i] = rp->t_minus[i]  - rp->total ;
			rp->t_plus[i] = add;
			rp->t_minus[i] = add;
		}
		rp->total = prob2scaledprob(1.0f);
	}
	return rp;
}


void print_samheader(struct db* db,struct parameters* param)
{
	int i;
	fprintf(outfile, "@HD\tVN:1.0\n");
	for(i = 0; i < db->seq_info->num_seq-1;i++){
		fprintf(outfile, "@SQ\tSN:%s\tLN:%u",db->seq_info->names[i], (unsigned int) db->seq_info->len[i+1] -db->seq_info->len[i]);
		if(param->assembly_version){
			fprintf(outfile, "\tAS:%s\n",param->assembly_version);
		}else{
			fprintf(outfile,"\n");
		}
	}
	
	fprintf(outfile, "@SQ\tSN:%s\tLN:%u",db->seq_info->names[db->seq_info->num_seq-1], (unsigned int) db->seq_info->total_len - db->seq_info->len[db->seq_info->num_seq-1]);
	if(param->assembly_version){
		fprintf(outfile, "\tAS:%s\n",param->assembly_version);
	}else{
		fprintf(outfile,"\n");
	}
	fprintf(outfile, "@PG\tID:Delve\tVN:%s\tCL:%s\n",VERSION,param->command_line) ;
}

int post_phmm_controller(struct db* db,struct parameters* param)
{
	struct read_info** ri = 0;
	FILE* file = 0;
	int i,size,numseq = 0;
	struct genome_probabilities* gp = 0;	
	struct hmm* hmm = init_hmm(500, 500);
	for(i =0;i < 5;i++){
		hmm->rey[i] = prob2scaledprob(db->seq_info->nuc_counts[i] );
	}
	read_hmm_from_file(hmm,param);
	
	ri = malloc(sizeof(struct read_info*) * param->num_query);
	
	for(i = 0; i < param->num_query;i++){
		ri[i] = malloc(sizeof(struct read_info));
		ri[i]->seq = 0;
		ri[i]->name = 0;
		ri[i]->qual = 0;
		ri[i]->len = 0;
		ri[i]->priors = 0;
		ri[i]->strand = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->hits = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->identity = malloc(sizeof(float)* (LIST_STORE_SIZE+1));
	}
	
	gp = malloc(sizeof(struct genome_probabilities));
	
	size = (db->seq_info->total_len/ 1024) + 1;

	gp->plus = malloc(sizeof(float*) * size);
	gp->minus = malloc(sizeof(float*) * size);
	gp->num_reads = 0;
	for(i = 0;i < size;i++){
		gp->plus[i] = 0;
		gp->minus[i] = 0;
	}

	if (!(file = fopen(param->infile[1] , "r" ))){
		fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[1]);
		exit(-1);
	}
		
	while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
		ri = get_posteriors(ri, hmm, db, param, numseq,gp);
	}
	
	print_bedgraphs(db,param,gp,size);
	
	for(i = 0;i < size;i++){
		if(gp->plus[i]){
			free(gp->plus[i]);
		}
		if(gp->minus[i]){
			free(gp->minus[i]);
		}
	}
	
	free_hmm(hmm);
	free(gp->plus);
	free(gp->minus);
	free(gp);
	fclose(file);
	return 1;
}



int realign_sam(struct db* db,struct parameters* param)
{
	struct read_info** ri = 0;
	
	FILE* file = 0;
	int i,numseq = 0;
	ri = malloc(sizeof(struct read_info*) * param->num_query);
	
	for(i = 0; i < param->num_query;i++){
		ri[i] = malloc(sizeof(struct read_info));
		ri[i]->seq = 0;
		ri[i]->name = 0;
		ri[i]->qual = 0;
		ri[i]->len = 0;
		ri[i]->priors = 0;
		ri[i]->strand = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->hits = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1));
		ri[i]->identity = malloc(sizeof(float)* (LIST_STORE_SIZE+1));
	}
	
	/*if(!isatty(0)){
		file = stdin;
	}else{*/
		if (!(file = fopen(param->infile[1] , "r" ))){
			fprintf(stderr,"Cannot open sam file '%s'\n",param->infile[1]);
			exit(-1);
		}
	//}
	print_samheader(db,param);
	while ((numseq = read_sam_chunk(ri,db, param,file)) != 0){
		fprintf(stderr,"Reading\n");
		ri = realign_alignments(ri, db,param,numseq);
	}
	//if(param->infile[1]){
		fclose(file);
	//}
	
	for(i = 0; i < param->num_query;i++){
		free(ri[i]->strand); 
		free(ri[i]->hits);
		free(ri[i]->identity);
		free(ri[i]);
	}
	free(ri);
	return 1;
}

struct read_info** realign_alignments(struct read_info** ri, struct db* db,struct parameters* param, int n)
{
	int i,c;
	int hit = 0;
	int strand;
	unsigned char* seq = 0;
	//int num_scores = 0;

	char alphabet[] = "ACGTN";
	struct hirsch_mem* hm = 0;
	hm = hirsch_mem_alloc(hm,MAX_LINE);
	
	seq = malloc(sizeof(unsigned char) * 500);
	for(i = 0;i < n;i++){
		hit = 0;
		//num_scores = 0;
		//fprintf(stderr,"%d\n",i);
		while(ri[i]->identity[hit] != -2.0f){ // option != -2 
			if(ri[i]->strand[hit] != ri[i]->strand[0]){
				strand = 1;
			}else{
				strand = 0;
			}
			
			c = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
			if (!strand){
				align(
				      ri[i]->name,
				      ri[i]->seq, 
				      ri[i]->qual,
				      ri[i]->len,
				      get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[hit]-10), 
				      seq, 
				      ri[i]->len+20 ,
				      get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[hit]),
				      hm,
				      ri[i]->strand[hit]
				      );
			}else{
				
				
				
				//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
				align(
				      ri[i]->name,
				      reverse_complement2(ri[i]->seq,ri[i]->len), 
				      ri[i]->qual,
				      ri[i]->len,
				      get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[hit]-10), 
				      seq, 
				      ri[i]->len+20 ,
				      get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[hit]) ,
				      hm,
				      ri[i]->strand[hit]
				      );
				//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
				reverse_complement2(ri[i]->seq,ri[i]->len);				
			}
			fprintf(outfile ,"%s",hm->pline);
			hit++;
			if(hit == param->upper_limit){
				break;
			}
		}
		
		if(hit == 0){
			
			for(hit = 0; hit < ri[i]->len;hit++){
				ri[i]->seq[hit] = alphabet[ri[i]->seq[hit]];
			}
			
			if(ri[i]->qual){
				sprintf(hm->pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
						ri[i]->name,//1 QNAME Query NAME of the read or the read pair 
						4,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
						"*",//3 RNAME Reference sequence NAME 
						0,//4 POS 1-based leftmost POSition of clipped alignment 
						0,//5 MAPQ MAPping Quality (Phred-scaled) 
						"*",//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
						"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
						0,//8 MPOS 1-based leftmost Mate POSition 
						0,//9 ISIZE inferred Insert SIZE 
						ri[i]->seq,//10 SEQ query SEQuence on the same strand as the reference 
						ri[i]->qual//11 QUAL query QUALity (ASCII-33=Phred base quality) 
						);
			}else{
				sprintf(hm->pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
						ri[i]->name,//1 QNAME Query NAME of the read or the read pair 
						4,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
						"*",//3 RNAME Reference sequence NAME 
						0,//4 POS 1-based leftmost POSition of clipped alignment 
						0,//5 MAPQ MAPping Quality (Phred-scaled) 
						"*",//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
						"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
						0,//8 MPOS 1-based leftmost Mate POSition 
						0,//9 ISIZE inferred Insert SIZE 
						ri[i]->seq,//10 SEQ query SEQuence on the same strand as the reference 
						"*"//11 QUAL query QUALity (ASCII-33=Phred base quality) 
						);
			}
			hit = strlen(hm->pline);
			hm->pline[hit] = 0;
			fprintf(outfile, "%s\n",hm->pline);
		}
		
	}
	hirsch_mem_free(hm);
	free(seq);
	return ri;
}

struct read_info** get_posteriors(struct read_info** ri, struct hmm* hmm, struct db* db,struct parameters* param, int n, struct genome_probabilities* gp	)
{
	int i,j,c;
	int hit = 0;
	int strand;
	unsigned char* seq = 0;
	
	float* genome_posterior = 0;
	
	unsigned long int*  top_scores = 0;
	float* random_emission = 0;
	
	int a, b;
	
	int counter = 0;
		
	top_scores = malloc(sizeof(unsigned long int)* LIST_STORE_SIZE);
	random_emission = malloc(sizeof(float)*5);
	random_emission[0] = prob2scaledprob(0.2475f);
	random_emission[1] = prob2scaledprob(0.2475f);
	random_emission[2] = prob2scaledprob(0.2475f);
	random_emission[3] = prob2scaledprob(0.2475f);
	random_emission[4] = prob2scaledprob(0.01f);
	
	//fprintf(stderr,"%s\n",old_chr);
	seq = malloc(sizeof(unsigned char) * 500);
	genome_posterior = malloc(sizeof(float) * 500);
	for(i = 0;i < n;i++){
		hit = 0;
		counter++; 
		
		if(counter % 1000 == 0){
			fprintf(stderr,"%d\n",counter);
		}
		
		//fprintf(stderr,"%d\n",i);
		//while(ri[i]->identity[hit] != -2.0f){ // option != -2 
		if(ri[i]->identity[0] != -2.0f){
			gp->num_reads++;
			//current_chr = get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[0]-10);
			
			if(param->quality_cutoff <=  ri[i]->mapq){
			
				ri[i]->mapq  = prob2scaledprob(ri[i]->mapq);
				
				strand = get_sequence(db,seq,ri[i]->hits[0]-10, ri[i]->len+20,0);
				
				for(j = 0;j <= ri[i]->len+20;j++){
					genome_posterior[j] = -SCALEINFTY;
				}
				
				hmm = global_local_forward_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]);
				hmm = global_local_backward_log(hmm,random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]);
				
				/*if(!ri[i]->strand[hit]){
					j = 1;
				}else{
					j =  ri[i]->len;
				}*/
				
				for(j = 1; j <= ri[i]->len;j++){
					for (c = 1; c <= ri[i]->len+20;c++){
						genome_posterior[c] = logsum(genome_posterior[c],hmm->fM[j][c] + hmm->bM[j][c]);
						genome_posterior[c] = logsum(genome_posterior[c],hmm->fC[j][c] + hmm->bC[j][c]);
					}
				}
								
				for(j = 1; j <= ri[i]->len+20;j++){
					genome_posterior[j] = genome_posterior[j] + ri[i]->mapq - hmm->score;
				}
				
				a = (ri[i]->hits[0]-10) >> 10;
				b = (ri[i]->hits[0] + ri[i]->len+10) >> 10;
				
				if(!ri[i]->strand[hit]){
					
					if(a == b){
						if(!gp->plus[a]){
							gp->plus[a] = malloc(sizeof(float) * 1024);
							for(j = 0; j < 1024;j++){
								gp->plus[a][j] = -SCALEINFTY;
							}
						}
						c = a;
						//fprintf(stderr,"Region:%d\n",c);
						a = (ri[i]->hits[0]-10)  & 0x3FF;
						b = (ri[i]->hits[0] + ri[i]->len+10) & 0x3FF;
						//fprintf(stderr,"%d	%d\n",a,b);
						for(j = a;j <= b;j++){
							//	fprintf(stderr,"reding from:%d\n",j - a);
							gp->plus[c][j] = logsum(gp->plus[c][j],genome_posterior[j - a]);
						}
					}else{
						if(!gp->plus[a]){
							gp->plus[a] = malloc(sizeof(float) * 1024);
							for(j = 0; j < 1024;j++){
								gp->plus[a][j] = -SCALEINFTY;
							}
						}
						if(!gp->plus[b]){
							gp->plus[b] = malloc(sizeof(float) * 1024);
							for(j = 0; j < 1024;j++){
								gp->plus[b][j] = -SCALEINFTY;
							}
						}
						c = a;
						a = (ri[i]->hits[0]-10)  & 0x3FF;
						
						//fprintf(stderr,"%d	%d\n",a,b);
						for(j = a;j < 1024;j++){
							//fprintf(stderr,"reding from:%d	(%p)\n",j - a,gp->plus[c]);
							gp->plus[c][j] = logsum(gp->plus[c][j],genome_posterior[j - a]);
						}
						//fprintf(stderr,"break\n");
						c = b;
						b = (ri[i]->hits[0] + ri[i]->len+10) & 0x3FF;
						for(j = 0;j <= b;j++){
							//fprintf(stderr,"reding from:%d\n",j + 1024 - a);
							gp->plus[c][j] = logsum(gp->plus[c][j],genome_posterior[j + 1024 - a]);
						}
					}
				}else{
					if(a == b){
						if(!gp->minus[a]){
							gp->minus[a] = malloc(sizeof(float) * 1024);
							for(j = 0; j < 1024;j++){
								gp->minus[a][j] = -SCALEINFTY;
							}
						}
						c = a;
						a = (ri[i]->hits[0]-10)  & 0x3FF;
						b = (ri[i]->hits[0] + ri[i]->len+10) & 0x3FF;
						for(j = a;j <= b;j++){
							gp->minus[c][j] = logsum(gp->minus[c][j],genome_posterior[j - a]);
						}
					}else{
						if(!gp->minus[a]){
							gp->minus[a] = malloc(sizeof(float) * 1024);
							for(j = 0; j < 1024;j++){
								gp->minus[a][j] = -SCALEINFTY;
							}
						}
						if(!gp->minus[b]){
							gp->minus[b] = malloc(sizeof(float) * 1024);
							for(j = 0; j < 1024;j++){
								gp->minus[b][j] = -SCALEINFTY;
							}
						}
						c = a;
						a = (ri[i]->hits[0]-10)  & 0x3FF;
						
						
						for(j = a;j < 1024;j++){
							gp->minus[c][j] = logsum(gp->minus[c][j],genome_posterior[j - a]);
						}
						
						c = b;
						b = (ri[i]->hits[0] + ri[i]->len+10) & 0x3FF;
						for(j = 0;j <= b;j++){
							gp->minus[c][j] = logsum(gp->minus[c][j],genome_posterior[j + 1024 -a]);
						}
					}
				}
			}
		}
	}
	free(seq);
	free(genome_posterior);
	free(random_emission);
	free(top_scores);
	return ri;
}

void print_bedgraphs( struct db* db,struct parameters* param, struct genome_probabilities* gp,int size)
{
	int i,j;
	
	unsigned int c,old_c; 
	unsigned char plus[] = "plus.bed";
	unsigned char minus[] = "minus.bed";
	
	unsigned char plus_wig[] = "plus.wig";
	unsigned char minus_wig[] = "minus.wig";

	unsigned char* old_chr_name = 0;
	unsigned char* tmp_name = malloc(sizeof(unsigned char)* 500);
	
	float scaling = prob2scaledprob(1000000.0f);
	float divider = prob2scaledprob((float)gp->num_reads);
	
	
	double sum_prob = 0.0f;
	
	for(i = 0; i < size;i++){
		if(gp->plus[i]){
			for(j = 0; j < 1024;j++){
				//if(gp->plus[i][j]  > -15.7f){
					sum_prob += gp->plus[i][j];
				//}
			}
		}
		if(gp->minus[i]){
			for(j = 0; j < 1024;j++){
				//if(gp->minus[i][j]  > -15.7f){
					sum_prob += gp->minus[i][j];
				//}
			}
		}
	}
	
	fprintf(stderr,"sum_prob: %f\n", sum_prob);
	double tmp_prob = 0;
	
	old_chr_name = plus_wig;
	old_c = 0;
	j = 0;
	for(i = 0; i < (int)strlen(param->outfile);i++){
		tmp_name[j] = param->outfile[i];
		j++;
	}
	tmp_name[j] = '_';
	j++;
	if(!param->wiggle){
		for(i = 0; i < (int)strlen((char*)plus);i++){
			tmp_name[j] =plus[i];
			j++;
		}
	}else{
		for(i = 0; i < (int)strlen((char*)plus_wig);i++){
			tmp_name[j] =plus_wig[i];
			j++;
		}
	}
	tmp_name[j] = 0;
	if ((outfile = fopen((char*)tmp_name, "w")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	
	if(!param->wiggle){
		fprintf(outfile,"track type=bedGraph name=\"%s\" description=\"%s\" visibility=full color=200,100,0 altColor=0,100,200 priority=20\n",param->outfile,param->outfile);
		for(i = 0; i < size;i++){
			if(gp->plus[i]){
				for(j = 0; j < 1024;j++){
					if(gp->plus[i][j] != -SCALEINFTY){
						tmp_prob = scaledprob2prob(gp->plus[i][j] + scaling - divider);
						if(tmp_prob > 1e-6){
							c = ((i << 10) + j);
							fprintf(outfile,"%s\t%d\t%d\t%f\n",
									get_chr_name (db->seq_info,c),
									get_chr_pos (db->seq_info,c),
									get_chr_pos (db->seq_info,c)+1,
									tmp_prob);
						}
					}
				}
			}
		}
	}else{
		//fprintf(outfile,"track type=wiggle_0 name=\"%s\" description=\"%s\"\n",param->outfile,param->outfile);
		for(i = 0; i < size;i++){
			if(gp->plus[i]){
				for(j = 0; j < 1024;j++){
					//if(gp->plus[i][j]  > -15.7f){
						tmp_prob =  (double) (gp->plus[i][j]*1000000.0 / sum_prob)  ;
						
						//tmp_prob = floor(tmp_prob* 100000.0 + 0.5 ) / 100000.0;
						
						if(tmp_prob > 1e-6){
							c = ((i << 10) + j);
							if(strcmp((char*)old_chr_name,(char*) get_chr_name (db->seq_info,c)) != 0 || old_c+1 != c){
								old_chr_name = get_chr_name (db->seq_info,c);
								fprintf(outfile,"variableStep chrom=%s\n",old_chr_name);
							}//else{
								fprintf(outfile,"%d	%0.8f\n",get_chr_pos (db->seq_info,c)+1,tmp_prob);
							//}
							old_c = c;
						}
					//}
				}
			}
		}
	}
	
	
	
	fclose(outfile);
	old_c = 0;
	old_chr_name = plus_wig;
	j = 0;
	for(i = 0; i < (int)strlen(param->outfile);i++){
		tmp_name[j] = param->outfile[i];
		j++;
	}
	tmp_name[j] = '_';
	j++;
	if(!param->wiggle){
		for(i = 0; i < (int)strlen((char*)minus);i++){
			tmp_name[j] =minus[i];
			j++;
		}
	}else{
		for(i = 0; i < (int)strlen((char*)minus_wig);i++){
			tmp_name[j] = minus_wig[i];
			j++;
		}
	}
	tmp_name[j] = 0;
	if ((outfile = fopen((char*)tmp_name, "w")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	
	if(!param->wiggle){
		fprintf(outfile,"track type=bedGraph name=\"%s\" description=\"%s\" visibility=full color=200,100,0 altColor=0,100,200 priority=20\n",param->outfile,param->outfile);
		
		
		for(i = 0; i < size;i++){
			if(gp->minus[i]){
				for(j = 0; j < 1024;j++){
					if(gp->minus[i][j] != -SCALEINFTY){
						tmp_prob = scaledprob2prob(gp->minus[i][j] + scaling - divider);
						if(tmp_prob > 1e-6){
							c = ((i << 10) + j);
							fprintf(outfile,"%s\t%d\t%d\t%f\n",
									get_chr_name (db->seq_info,c),
									get_chr_pos (db->seq_info,c),
									get_chr_pos (db->seq_info,c)+1,
									tmp_prob);
						}
					}
				}
			}
		}
	}else{
		//fprintf(outfile,"track type=wiggle_0 name=\"%s\" description=\"%s\"\n",param->outfile,param->outfile);
		for(i = 0; i < size;i++){
			if(gp->minus[i]){
				for(j = 0; j < 1024;j++){
					//if(gp->minus[i][j]  > -15.7f){
						tmp_prob =  (double) (gp->minus[i][j]*1000000.0 / sum_prob)  ;
						//tmp_prob =  (double)  scaledprob2prob(gp->minus[i][j]+scaling  - sum_prob) ;
						//tmp_prob = floor(tmp_prob* 100000.0 + 0.5 ) / 100000.0;
						if(tmp_prob > 1e-6){
							c = ((i << 10) + j);
							if(strcmp((char*)old_chr_name,(char*) get_chr_name (db->seq_info,c)) != 0 || old_c+1 != c){
								old_chr_name = get_chr_name (db->seq_info,c);
								fprintf(outfile,"variableStep chrom=%s\n",old_chr_name);
							}//else{
								fprintf(outfile,"%d	%0.8f\n",get_chr_pos (db->seq_info, c),tmp_prob);
							//}
							old_c = c;
						}
					//}
				}
			}
		}
	}
	fclose(outfile);	
}


void* do_score_alignments_thread(void *threadarg)
{
	struct thread_data *data;
	data = (struct thread_data *) threadarg;
	
	struct read_info** ri = data->ri;
	struct hmm* hmm = data->hmm;
	struct db* db = data->db;
	const int start = data->start; 
	const int end = data->end;
	const struct parameters* param = data->param;
	int i,j,c;
	int hit = 0;
	int strand;
	unsigned char* seq = 0;
	unsigned char* aln = 0;
	int num_scores = 0;
	//int k = 0;
	unsigned long int*  top_scores = 0;
	unsigned long int top_cut = 0ul;
	float* random_emission = 0;
	unsigned char* read_posteriors_code;
	
	float* alignment_scores = 0;
	float* mapping_scores = 0;
	float** read_posterior = 0;
	char nuc[] = "ACGTN";
	//unsigned char noqual[] = "*";
	//unsigned char* qual = 0;
	float tmp;
	float sum,sum_alignment_scores;
	char pline[500];
	alignment_scores = malloc(sizeof(float)* LIST_STORE_SIZE);
	mapping_scores = malloc(sizeof(float)* LIST_STORE_SIZE);
	top_scores = malloc(sizeof(unsigned long int)* LIST_STORE_SIZE);
	random_emission = malloc(sizeof(float)*5);
	random_emission[0] = prob2scaledprob(0.2475f);
	random_emission[1] = prob2scaledprob(0.2475f);
	random_emission[2] = prob2scaledprob(0.2475f);
	random_emission[3] = prob2scaledprob(0.2475f);
	random_emission[4] = prob2scaledprob(0.01f);	
	
	read_posterior = malloc(sizeof(float*)*LIST_STORE_SIZE  );
	for(i = 0; i < LIST_STORE_SIZE;i++){
		read_posterior[i] = malloc(sizeof(float)* 500 );
	}
	seq = malloc(sizeof(unsigned char) * 500);
	
	read_posteriors_code = malloc(sizeof(unsigned char) * 500);
	for(i = start;i < end;i++){
		hit = 0;
		num_scores = 0;
		//fprintf(stderr,"%d\n",i);
		for(c = 0; c < LIST_STORE_SIZE; c++){
			for(j = 0;j <= ri[i]->len+2;j++){
				read_posterior[c][j] = -SCALEINFTY;
			}
		}
		
		if(ri[i]->identity[0] == -2.0f){
			
			for(hit = 0; hit < ri[i]->len;hit++){
				ri[i]->seq[hit] = nuc[(int)ri[i]->seq[hit]];
			}
			hit = 0;
			if(ri[i]->qual){
				sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
						ri[i]->name ,//1 QNAME Query NAME of the read or the read pair 
						4,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
						"*",//3 RNAME Reference sequence NAME 
						0,//4 POS 1-based leftmost POSition of clipped alignment 
						0,//5 MAPQ MAPping Quality (Phred-scaled) 
						"*",//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
						"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
						0,//8 MPOS 1-based leftmost Mate POSition 
						0,//9 ISIZE inferred Insert SIZE 
						ri[i]->seq,//10 SEQ query SEQuence on the same strand as the reference 
						ri[i]->qual//11 QUAL query QUALity (ASCII-33=Phred base quality) 
						);
			}else{
				sprintf(pline,"%s\t%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
						ri[i]->name ,//1 QNAME Query NAME of the read or the read pair 
						4,//2 FLAG bitwise FLAG (pairing, strand, mate strand, etc.) 
						"*",//3 RNAME Reference sequence NAME 
						0,//4 POS 1-based leftmost POSition of clipped alignment 
						0,//5 MAPQ MAPping Quality (Phred-scaled) 
						"*",//6 CIGAR extended CIGAR string (operations: MIDNSHP) 
						"*",//7 MRNM Mate Reference NaMe (‘=’ if same as RNAME) 
						0,//8 MPOS 1-based leftmost Mate POSition 
						0,//9 ISIZE inferred Insert SIZE 
						ri[i]->seq,//10 SEQ query SEQuence on the same strand as the reference 
						"*"//11 QUAL query QUALity (ASCII-33=Phred base quality) 
						);
			}
			c = strlen(pline);
			pline[c] = 0;
			fprintf(outfile, "%s\n",pline);
			
		}
		
		while(ri[i]->identity[hit] != -2.0f){ // option != -2 
			
			if(hit == 0){
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
			}else{
				if(ri[i]->strand[hit] != ri[i]->strand[0]){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
				}else{
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				}
			}
			hmm = global_local_forward_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]); 
			hmm = global_local_backward_log(hmm,random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]);
			
			for(j = 1; j <= ri[i]->len;j++){
				for (c = 1; c <= ri[i]->len+20;c++){
					read_posterior[hit][j] =  logsum(read_posterior[hit][j],hmm->fM[j][c] + hmm->bM[j][c]);
					read_posterior[hit][j] =  logsum(read_posterior[hit][j],hmm->fC[j][c] + hmm->bC[j][c]);
				}
			}
			
			for(j = 1; j <= ri[i]->len;j++){
				read_posterior[hit][j] =  read_posterior[hit][j] - hmm->score;
			}			
			
			mapping_scores[num_scores] = hmm->score + ri[i]->priors[hit];
			alignment_scores[num_scores] = hmm->score;
			
			if(param->viterbi){
				aln = global_local_viterbi_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]);
				mapping_scores[num_scores] = hmm->score ;
				free(aln);
			}
			
			
			num_scores++;
			hit++;
		}
		
		if(num_scores){
			sum = mapping_scores[0];
			sum_alignment_scores = alignment_scores[0];
			for(j = 1; j < num_scores;j++){
				sum = logsum(sum,mapping_scores[j]);
				sum_alignment_scores = logsum(sum_alignment_scores ,alignment_scores[j]);
			}
			
			/*if(num_scores > 1){
				for(j = 0; j < num_scores;j++){
					fprintf(stderr,"%f, ", scaledprob2prob(mapping_scores[j]-sum));
				}
				fprintf(stderr,"\n");
				fprintf(stderr,"%s\n",ri[i]->name);
				hit = 0;
				while(ri[i]->identity[hit] != -2.0f){ // option != -2 
					
					
					if(ri[i]->strand[hit] != 0){
						//reverse_complement2(ri[i]->seq,ri[i]->len);
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
						//hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ); 
						//reverse_complement2(ri[i]->seq,ri[i]->len);
						
						//reverse_complement2(ri[i]->seq,ri[i]->len);
						
					}else{
						strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
						//hmm = glocal_forward_2ndorder_log_Y(hmm,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20 ); 
					}
					
					for(j = 0; j <  ri[i]->len+20;j++){
						fprintf(stderr,"%d ",seq[j]);
					}
					fprintf(stderr,"\n");
					
					//mapping_scores[num_scores] = hmm->score + ri[i]->priors[hit];
					//num_scores++;
					hit++;
				}
				
				
			}	*/	
			//max = -1.0f;
			
			for(j = 0; j < LIST_STORE_SIZE;j++){
				top_scores[j] = 0ul;
			}
						
			
			for(j = 0; j < num_scores;j++){
				BinaryInsertionSortOne(top_scores,LIST_STORE_SIZE,(unsigned long int) (scaledprob2prob( mapping_scores[j] - sum) * 1000000.0f));
				//if(!strcmp(ri[i]->name, "B8A3AD_1")){
				//	fprintf(stderr,"PROBABILITY:%f\n",scaledprob2prob(mapping_scores[j]  -sum)* 1000000.0f);
				//}
			}
			
			j = 1;
			top_cut = top_scores[param->upper_limit-j];
			while(top_cut == 0.0f){
				j++;
				top_cut = top_scores[param->upper_limit-j];
			}
			
			
			//fprintf(stderr,"%s\n",ri[i]->name);
			// for(j = 0; j < num_scores;j++){
			// fprintf(stderr,"%f	%f	%f	%f	ident:%f\n",alignment_scores[j],sum, alignment_scores[j] - sum,scaledprob2prob( alignment_scores[j] - sum), ri[i]->identity[j]);
			// }
			 //fprintf(stderr,"\n");
			if(num_scores == 1){
				hit = 0;
				for(j = 0; j < num_scores;j++){
					if((unsigned long int) (scaledprob2prob( mapping_scores[j] - sum) * 1000000.0f) >= top_cut){
						if(ri[i]->strand[j] != ri[i]->strand[0]){
							strand = 1;
						}else{
							strand = 0;
						}
						
						
						for(c = 1; c <= ri[i]->len;c++){
							tmp = scaledprob2prob(read_posterior[j][c]);
							if(tmp >= 1.0 ){
								tmp = 99;
							}else{
								tmp = floor(-10.0 * log10 (1.0 - tmp));
							}
							//fprintf(stdout,"%d	%d	%f	%f	%f\n",c,num_scores,read_posterior[j][c],scaledprob2prob(read_posterior[j][c]), tmp);
							
							if(tmp <= 93 ){
								read_posteriors_code[c-1] = (char) (tmp + 33);
							}else{
								read_posteriors_code[c-1] = (char) (93 + 33);
							}
						}
						read_posteriors_code[ri[i]->len] = 0;
						
						
						c = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,0);
						if (!strand){
							aln = global_local_viterbi_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[j]);
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 ri[i]->mapq,
										 ri[i],
										 read_posteriors_code
										 );
							free(aln);
						}else{
							aln = global_local_viterbi_log(hmm, random_emission,  reverse_complement2(ri[i]->seq,ri[i]->len), seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[j]);
							//???
							//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
							
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int)ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 ri[i]->mapq,
										 ri[i],
										 reverse_without_complement(read_posteriors_code,ri[i]->len)
										 );
							//???
							//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
							reverse_complement2(ri[i]->seq,ri[i]->len);
							free(aln);
						}
						hit++;
						if(hit == param->upper_limit){
							break;
						}
					}
				}
			}else{
				hit = 0;
				for(j = 0; j < num_scores;j++){
					if((unsigned long int) (scaledprob2prob( mapping_scores[j] - sum) * 1000000.0f) >= top_cut){
						if(ri[i]->strand[j] != ri[i]->strand[0]){
							strand = 1;
						}else{
							strand = 0;
						}
						
						/*if(ri[i]->qual){
							if(strand){
								ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
							}
							qual = ri[i]->qual;
							
							
						}else{
							qual = noqual;
						}*/
						
						
						
						for(c = 1; c <= ri[i]->len;c++){
							tmp = scaledprob2prob(read_posterior[j][c]);
							if(tmp >= 1.0 ){
								tmp = 99;
							}else{
								tmp = floor(-10.0 * log10 (1.0 - tmp));
							}
							//fprintf(stdout,"%d	%d	%f	%f	%f\n",c,num_scores,read_posterior[j][c],scaledprob2prob(read_posterior[j][c]), tmp);
							
							if(tmp <= 93 ){
								read_posteriors_code[c-1] = (char) (tmp + 33);
							}else{
								read_posteriors_code[c-1] = (char) (93 + 33);
							}
						}
						read_posteriors_code[ri[i]->len] = 0;
						
						c = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,0);
						if (!strand){
							//if(!ri[i]->qual){
							//	ri[i]->qual = noqual;
							//}
							
							aln = global_local_viterbi_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[j]);
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 scaledprob2prob( mapping_scores[j] - sum),
										 ri[i],
										 read_posteriors_code
										 );
							free(aln);
						}else{
							aln = global_local_viterbi_log(hmm, random_emission,  reverse_complement2(ri[i]->seq,ri[i]->len), seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[j]);
							
							
							//???
							//if(ri[i]->qual){
								ri[i]->qual = reverse_without_complement(ri[i]->qual, strlen ((const char*) ri[i]->qual));
							//}else{
							//	ri[i]->qual = noqual;
							//}
							
							//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
							
							align_to_sam(
										 aln,
										 get_chr_name (db->seq_info,(unsigned int)ri[i]->hits[j]-10),
										 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
										 ri[i]->strand[j], 
										 scaledprob2prob( mapping_scores[j] - sum),
										 ri[i],
										 reverse_without_complement(read_posteriors_code,ri[i]->len)
										 );
							//???
							
							//if(ri[i]->qual){
							ri[i]->qual = reverse_without_complement(ri[i]->qual,strlen((const char*)  ri[i]->qual));
							//}
							
							//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
							reverse_complement2(ri[i]->seq,ri[i]->len);
							free(aln);
						}
						hit++;
						if(hit == param->upper_limit){
							break;
						}
					}
				}
			}
		}
	}
	
	free(mapping_scores);

	free(alignment_scores);
	free(random_emission);
	free(read_posteriors_code);
	for(i = 0; i < LIST_STORE_SIZE;i++){
		free(read_posterior[i]);
	}
	free(read_posterior);
	free(top_scores);	
	free(seq);
	pthread_exit((void *) 0);
}



struct read_info** score_alignments(struct read_info** ri, struct hmm* hmm, struct db* db,struct parameters* param, int n)
{
	int i,j,c;
	int hit = 0;
	int strand;
	unsigned char* seq = 0;
	int num_scores = 0;
	//int k = 0;
	unsigned long int*  top_scores = 0;
	unsigned long int top_cut = 0ul;
	float* random_emission = 0;
	unsigned char* read_posteriors_code;

	float* alignment_scores = 0;
	float** read_posterior = 0;
	
	float tmp;
	float sum;
	
	alignment_scores = malloc(sizeof(float)* LIST_STORE_SIZE);
	top_scores = malloc(sizeof(unsigned long int)* LIST_STORE_SIZE);
	random_emission = malloc(sizeof(float)*5);
	random_emission[0] = prob2scaledprob(0.2475f);
	random_emission[1] = prob2scaledprob(0.2475f);
	random_emission[2] = prob2scaledprob(0.2475f);
	random_emission[3] = prob2scaledprob(0.2475f);
	random_emission[4] = prob2scaledprob(0.01f);
	
	read_posterior = malloc(sizeof(float*)*LIST_STORE_SIZE);
	for(i = 0; i < LIST_STORE_SIZE;i++){
		read_posterior[i] = malloc(sizeof(float)* 500 );
	}
	seq = malloc(sizeof(unsigned char) * 500);
	
	read_posteriors_code = malloc(sizeof(unsigned char) * 500);
	
	for(i = 0;i < n;i++){
		hit = 0;
		num_scores = 0;
		//fprintf(stderr,"%d\n",i);
		for(c = 0; c < LIST_STORE_SIZE; c++){
			for(j = 0;j <= ri[i]->len+2;j++){
				read_posterior[c][j] = -SCALEINFTY;
			}
		}
		
		while(ri[i]->identity[hit] != -2.0f){ // option != -2 
			
			if(hit == 0){
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
			}else{
				if(ri[i]->strand[hit] != ri[i]->strand[0]){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
				}else{
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				}
			}
			hmm = global_local_forward_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]); 
			hmm = global_local_backward_log(hmm,random_emission,ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[hit]);
			
			
			for(j = 1; j <= ri[i]->len;j++){
				for (c = 1; c <= ri[i]->len+20;c++){
					//if (ri[i]->seq[j-1] == seq[c-1]){
					read_posterior[hit][j] =  logsum(read_posterior[hit][j],hmm->fM[j][c] + hmm->bM[j][c] - hmm->score);
					read_posterior[hit][j] =  logsum(read_posterior[hit][j],hmm->fC[j][c] + hmm->bC[j][c] - hmm->score);
					//}
				}
			}
			alignment_scores[num_scores] = hmm->score ;
			num_scores++;
			hit++;
		}
		
		if(num_scores){
			sum = -SCALEINFTY;
			for(j = 0; j < num_scores;j++){
			//	fprintf(stderr,"%f +	%f	= %f\n", sum ,alignment_scores[j],logsum(sum ,alignment_scores[j]) );
				sum = logsum(sum,alignment_scores[j]);
			}
			//max = -1.0f;
			/*sum = -SCALEINFTY;
			for(j = num_scores-1; j >= 0;j--){
				fprintf(stderr,"%f +	%f	= %f\n", sum ,alignment_scores[j],logsum(sum ,alignment_scores[j]) );
				sum = logsum(sum,alignment_scores[j]);
			}*/
			for(j = 0; j < LIST_STORE_SIZE;j++){
				top_scores[j] = 0ul;
			}
						
			
			for(j = 0; j < num_scores;j++){
				
				
				
				//alignment_scores[j] = alignment_scores[j] - sum;
				//if((alignment_scores[j] - sum) > max){
				//	max = alignment_scores[j] - sum;
				//}
				BinaryInsertionSortOne(top_scores,LIST_STORE_SIZE,(unsigned long int) (scaledprob2prob( alignment_scores[j] - sum) * 1000000.0f));
			}
			top_cut = top_scores[param->upper_limit-1];
			
			/*fprintf(stderr,"\n%s\n",ri[i]->name);
			for(j = 0; j < num_scores;j++){
				fprintf(stderr,"%f	%f	%f	%f	ident:%f		%lud	%lud\n",alignment_scores[j],sum, alignment_scores[j] - sum,scaledprob2prob( alignment_scores[j] - sum), ri[i]->identity[j],top_cut,top_scores[j]);
			}
			fprintf(stderr,"\n");*/
			
			
			
			hit = 0;
			for(j = 0; j < num_scores;j++){
				if((unsigned long int) (scaledprob2prob( alignment_scores[j] - sum) * 1000000.0f) >= top_cut){
					if(ri[i]->strand[j] != ri[i]->strand[0]){
						strand = 1;
					}else{
						strand = 0;
					}
					
					for(c = 1; c <= ri[i]->len;c++){
						tmp = scaledprob2prob(read_posterior[j][c]);
						if(tmp >= 1.0 ){
							tmp = 99;
						}else{
							tmp = floor(-10.0 * log10 (1.0 - tmp));
						}
						//fprintf(stdout,"%d	%d	%f	%f	%f\n",c,num_scores,read_posterior[j][c],scaledprob2prob(read_posterior[j][c]), tmp);
						
						if(tmp <= 93 ){
							read_posteriors_code[c-1] = (char) (tmp + 33);
						}else{
							read_posteriors_code[c-1] = (char) (93 + 33);
						}
					}
					read_posteriors_code[ri[i]->len] = 0;
					
					c = get_sequence(db,seq,ri[i]->hits[j]-10, ri[i]->len+20,0);
					if (!strand){
						
						
						
						
						align_to_sam(
						global_local_viterbi_log(hmm, random_emission, ri[i]->seq, seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[j]),
									 get_chr_name (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
									 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
									ri[i]->strand[j], 
									 scaledprob2prob( alignment_scores[j] - sum),
									 ri[i],
									read_posteriors_code
							     );
					}else{
						//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
						align_to_sam(
						global_local_viterbi_log(hmm, random_emission,  reverse_complement2(ri[i]->seq,ri[i]->len), seq, ri[i]->len, ri[i]->len+20,ri[i]->strand[j]),
									 get_chr_name (db->seq_info,(unsigned int)ri[i]->hits[j]-10),
									 get_chr_pos (db->seq_info,(unsigned int) ri[i]->hits[j]-10),
									ri[i]->strand[j], 
									 scaledprob2prob( alignment_scores[j] - sum),
									ri[i],
									reverse_without_complement(read_posteriors_code,ri[i]->len)
									 );
						reverse_complement2(ri[i]->seq,ri[i]->len);
						//ri[i]->qual = reverse_without_complement(ri[i]->qual, ri[i]->len);
					}
					hit++;
					if(hit == param->upper_limit){
						break;
					}
				}
			}
			/*if(!strcmp((char*)"chr1_4998659_4999201_1:0:0_2:0:0_129/1",(char*)ri[i]->name)){
				exit(0);
			}*/
		}
	}
	free(alignment_scores);
	free(random_emission);
	free(read_posteriors_code);
	free(read_posterior);
	free(top_scores);
	return ri;
}

struct read_info** run_score_alignments(struct hmm* hmm,struct read_info** ri,struct db* db,struct parameters* param,int size)
{
	struct thread_data* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	init_logsum();
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	//hmm = init_hmm(lena, lenb, 0.002f, 0.001f, 0.02f, 1);
	
	
	interval =  (int)((double)size /(double)param->num_threads);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].db = db;
		thread_data_array[t].hmm =  init_hmm(500, 500);
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].db = db;
	thread_data_array[param->num_threads - 1].hmm = init_hmm(500, 500);
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = size;
	thread_data_array[param->num_threads - 1].param = param;

	init_thread_hmms(thread_data_array,hmm,param->num_threads);
	
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, do_score_alignments_thread, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	
	for(t = 0;t < param->num_threads;t++) {
		free_hmm(thread_data_array[t].hmm);
	}
	free(thread_data_array);
	return ri;
}


struct hmm* run_pHMM(struct hmm* localhmm,struct read_info** ri,struct db* db,struct parameters* param,int size)
{
	struct thread_data* thread_data_array = 0; 
	thread_data_array = malloc(sizeof(struct thread_data)* param->num_threads);
	pthread_t threads[param->num_threads];
	pthread_attr_t attr;
	int t;
	int interval = 0;
	int rc;
	
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	//hmm = init_hmm(lena, lenb, 0.002f, 0.001f, 0.02f, 1);
	
	
	interval =  (int)((double)size /(double)param->num_threads);
	
	fprintf(stderr,"Starting with %d sequences\n",size);
	
	for(t = 0;t < param->num_threads - 1;t++) {
		thread_data_array[t].ri = ri;
		thread_data_array[t].db = db;
		thread_data_array[t].hmm = init_hmm(500, 500);
		thread_data_array[t].start = t*interval;
		thread_data_array[t].end = t*interval + interval;
		thread_data_array[t].param = param;
	}
	thread_data_array[param->num_threads - 1].ri = ri;
	thread_data_array[param->num_threads - 1].db = db;
	thread_data_array[param->num_threads - 1].hmm = init_hmm(500, 500);
	thread_data_array[param->num_threads - 1].start = t*interval;
	thread_data_array[param->num_threads - 1].end = size;
	thread_data_array[param->num_threads - 1].param = param;
	
	init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	
	rc = pthread_attr_init(&attr);
	if(rc){
		fprintf(stderr,"ERROR; return code from pthread_attr_init() is %d\n", rc);
		exit(-1);
	}
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
	
	for(t = 0;t < param->num_threads;t++) {
		rc = pthread_create(&threads[t], &attr, do_baum_welch_thread, (void *) &thread_data_array[t]);
		if (rc) {
			fprintf(stderr,"ERROR; return code from pthread_create() is %d\n", rc);
			exit(-1);
		}
	}
	
	pthread_attr_destroy(&attr);
	
	for (t = 0;t < param->num_threads;t++){
		rc = pthread_join(threads[t], NULL);
		if (rc){
			fprintf(stderr,"ERROR; return code from pthread_join()is %d\n", rc);
			exit(-1);
		}
	}
	
	entangle_hmms(thread_data_array,localhmm,param->num_threads);
	//init_thread_hmms(thread_data_array,localhmm,param->num_threads);
	for(t = 0;t < param->num_threads;t++) {
		free_hmm(thread_data_array[t].hmm);
		//print_hmm_param(thread_data_array[t].hmm);
	}
	
	free(thread_data_array);
	return localhmm;
}

int read_sam_chunk(struct read_info** ri,struct db* db,struct parameters* param,FILE* file)
{
	char tmp_text[MAX_LINE];
	char line[MAX_LINE];
	int column = 0; 
	int i,j,g,tmp;
	unsigned int pos;
	int read = 0;
	int c = 0;
	int hit = 0;
	
	for(i = 0; i < param->num_query;i++){
		if(ri[i]->seq){
			free(ri[i]->seq);
		}
		if(ri[i]->name){
			free(ri[i]->name);
		}
		if(ri[i]->qual){
			free(ri[i]->qual);
		}
		if(ri[i]->priors){
			free(ri[i]->priors);
		}
		//free(ri[i]->strand);
		ri[i]->priors = 0;
		ri[i]->seq = 0;
		ri[i]->name = 0;
		ri[i]->qual = 0;
		ri[i]->len = 0;
		//ri[i]->strand = 0;
	}
	
	while(fgets(line, MAX_LINE, file)){
		if(line[0] != '@'){
			column = 1; //<QNAME> 
			tmp = 0;
			hit = 0;
			pos = 0xFFFFFFFFu;
			for(j = 0;j < MAX_LINE;j++){
				tmp++;
				if(isspace((int)line[j])){
					break;
				}
			}
			
			ri[c]->name = malloc(sizeof(unsigned char)* tmp);
			for(j = 0;j < MAX_LINE;j++){
				
				if(isspace((int)line[j])){
					ri[c]->name[j] = 0;
					break;
				}
				ri[c]->name[j] = line[j];
			}
			
			for(i = 0; i < MAX_LINE;i++){
				if(line[i] == '\n'){
					break;
				}
				if(isspace((int)line[i])){
					column++;
					switch(column){
						case 2: // <FLAG>
							tmp = atoi(line+i+1);
							ri[c]->strand[hit] = (tmp & 0x10);
							break;
						case 3: // <RNAME> 
							if(line[i+1] != '*'){
								tmp = 0;
								for(j = i+1;j < MAX_LINE;j++){
									if(isspace((int)line[j])){
										tmp_text[tmp] = 0;
										break;
									}
									tmp_text[tmp] = line[j];
									tmp++;
								}
								pos = get_chr_offset(db->seq_info, tmp_text);
								if(pos == 0xFFFFFFFFu){
									fprintf(stderr,"ERROR: %s not found in indexed sequences\n",tmp_text);
									fprintf(stderr,"%s\n",line);
									exit(-1);
								}
							}
							break;
						case 4: // <POS>
							if(pos != 0xFFFFFFFFu){
								//fprintf(stderr,"Adding %u	%u	= %u\n",pos,(unsigned int) atoi(line + i + 1),(unsigned int) atoi(line + i + 1) + pos);
								pos += (unsigned int) atoi(line + i + 1);
								ri[c]->hits[hit] = pos-1;
								ri[c]->identity[hit] = -1.0f;
								hit++;
								
							}//else{ // Hack to count unmapped reads...  
							//	param->total_len++;
							//}
							break;
						case 5: //  <MAPQ>
							
							ri[c]->mapq =  1.0 - pow(10.0, -1.0 *   atof(line +i +1) / 10.0); 
							
							break;
						case 6: //  <CIGAR>
							break;
						case 7: //  <MRNM>
							break;
						case 8: //  <MPOS>
							break;
						case 9: //  <ISIZE>
							break;
						case 10: // <SEQ>
							
							tmp = 0;
							for(j = i+1;j < MAX_LINE;j++){
								tmp++;
								if(isspace((int)line[j])){
									break;
								}
							}
							
							ri[c]->seq = malloc(sizeof(unsigned char)* tmp);
							g = 0;
							for(j = i+1;j < MAX_LINE;j++){
								
								if(isspace((int)line[j])){
									ri[c]->seq[g] = 0;
									break;
								}
								ri[c]->seq[g] = nuc_code5[(int)line[j]];
								g++;
							}
							
							ri[c]->len = g;
							break;
						case 11: // <QUAL>
							tmp = 0;
							for(j = i+1;j < MAX_LINE;j++){
								tmp++;
								if(isspace((int)line[j])){
									break;
								}
							}
							//if(tmp == 2){
							//	ri[c]->qual = 0;
								//fprintf(stderr,"%d	%s",tmp,line +i+1);
							//}else{
								g= 0;
								ri[c]->qual = malloc(sizeof(unsigned char)* tmp);
								for(j = i+1;j < MAX_LINE;j++){
									
									if(isspace((int)line[j])){
										ri[c]->qual[g] = 0;
										break;
									}
									ri[c]->qual[g] = line[j];
									g++;
								}
							//}
							break;
						default: 
							
							
							tmp = byg_end("XA:Z:", line + i );
							if(tmp){
								
								//j = byg_count(";",line + tmp);
								g = 0;
								for(j = tmp + i;j < MAX_LINE;j++){
									if(isspace((int)line[j])){
										break; 
									}	
									
									if(line[j]  == ','){
										tmp_text[g] = 0;
										pos = get_chr_offset(db->seq_info, tmp_text);
										if(pos == 0xFFFFFFFFu){
											fprintf(stderr,"ERROR: %s not found in indexed sequences\n",tmp_text);
										}//else{
										//	fprintf(stderr,"Found: %s\n",tmp_text);
										//}
										tmp = atoi(line + j + 1);
										
										if(tmp > 0 ){
											ri[c]->strand[hit] = 0;
										}else{
											ri[c]->strand[hit] = 0x10;
											tmp = tmp *-1;
										}
										
										pos += tmp;
										ri[c]->hits[hit] = pos-1;
										ri[c]->identity[hit] = -1.0f;
										
										if(hit == LIST_STORE_SIZE){
											//i = MAX_LINE;
											break;
										}
										g = 0;
										
										j += byg_end(",", line + j +1 );
										
										j += byg_end(",", line + j +1 );
										ri[c]->identity[hit] = atoi(line + j + 1);
										hit++;
										
										j += byg_end(";", line + j +1 );
										
										j+= 1;
										if(isspace((int)line[j])){
											break; 
										}
										
									}
									tmp_text[g] = line[j];
									g++;
								}
								
							}
							tmp = byg_end("NM:i:", line + i );
							if(tmp){
								
								//fprintf(stderr,"%s errors:%d\n",line,atoi(line+tmp+i)); 
								//if(!param->seed_errors){
								if((double) atoi(line+tmp+i) / (double)   ri[c]->len > 0.1){
									//fprintf(stderr,"%f	%f	%f\n",(double) atoi(line+tmp+i) , (double)ri[c]->len , (double) atoi(line+tmp+i) / (double) ri[c]->len );
									param->total_len++;
								}
								ri[c]->identity[0] =  atoi(line+tmp+i);// -1.0f;
								//}
								
								
							}
							tmp = byg_end("XP:Z:", line + i );
							if(tmp){
								
								/*$q = chr(($Q<=93? $Q : 93) + 33);
								where chr() is the Perl function to convert an integer to a character based on the ASCII table.
								Conversely, given a character $q, the corresponding Phred quality can be calculated with:
								$Q = ord($q) - 33;
								*/
								ri[c]->priors = malloc(sizeof(float) * (ri[c]->len+1));
								//ri[c]->priors[0] = 0;
								g= 0;
								//fprintf(stderr,"%s\n",line);
								for(j = tmp+i;j < MAX_LINE;j++){
									if(isspace((int)line[j])){
										break;
									}
								//	fprintf(stderr,"Reading:%d	%c\n",g, line[j]);
									ri[c]->priors[g] = (int)(line[j]) - 33; 
									g++;
								}
							}else{
								ri[c]->priors = malloc(sizeof(unsigned int)* (LIST_STORE_SIZE+1)); // dirty fix for resuing prior array to store alignment quality during training....
								
								for(j = 0; j < LIST_STORE_SIZE+1;j++){
									ri[c]->priors[j] = 0;
								}
								
							}
							
							
							i = MAX_LINE;
							break;
					}
				}
			}
			ri[c]->hits[hit] = 0xFFFFFFFFu;
			ri[c]->identity[hit] = -2.0f;
			/*fprintf(stderr,"%s\n",line);
			for(i = 0; i < hit;i++){
				fprintf(stderr,"%d ",ri[c]->strand[i] );
			}
			fprintf(stderr,"\n");
			for(i = 0; i < hit;i++){
				fprintf(stderr,"%f ",ri[c]->identity[i] );
			}
			fprintf(stderr,"\n\n");
			*/
			
			//fprintf(stderr,"%f\n",ri[c]->mapq);
			c++;
			read++;
			if(c == param->num_query){
				return c;
			}
		}
	}
	return c;
}

struct read_info** add_k(struct read_info** ri, struct db* db,int n)
{
	int i;
	int hit = 0;
	int strand;
	unsigned char* seq = 0;
	
	int k = 0;
	
	seq = malloc(sizeof(unsigned char) * 500);
	for(i = 0;i < n;i++){
		hit = 0;
		
		
		//fprintf(stderr,"%s\n",ri[i]->name);
		while(ri[i]->identity[hit] == -1.0f){
			if(ri[i]->hits[hit] < 10){
				ri[i]->hits[hit] = 10;
			}
			
			if(hit == 0){
				strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
			}else{
				if(ri[i]->strand[hit] != ri[i]->strand[0]){
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,1);
				}else{
					strand = get_sequence(db,seq,ri[i]->hits[hit]-10, ri[i]->len+20,0);
				}
			}
			k = (int) bpm((unsigned char*)seq,(unsigned char*)ri[i]->seq,ri[i]->len+20,ri[i]->len);
			ri[i]->identity[hit] = (float)(ri[i]->len - k) / (float)ri[i]->len;
		//	fprintf(stderr,"%f\n",ri[i]->identity[hit]);
			hit++;
		}
	}
	free(seq);
	return ri;
}

struct read_info** add_uniform_prior(struct read_info** ri, int n)
{
	int i,j;
	
	float base_line = 0.0f;
	
	for(i = 0;i < n;i++){
		base_line = prob2scaledprob(1.0f / (float) ri[i]->len   );
		ri[i]->priors = malloc(sizeof(float) * ri[i]->len);
		
		for(j = 0;j < ri[i]->len;j++){
			ri[i]->priors[j] = base_line;
		}
		
	}
	return ri;
}



int byg_end(const char* pattern,const char*text)
{
	const char* tmp = 0;
	int Tc;
	int i  = 0;
	int s = 0;
	int T[256];
	for (i = 0;i < 256;i++){ 
		T[i] = 0; 
	}
	
	int m = (int)strlen(pattern);
	int n = (int)strlen(text);
	if (m > n){
		i = m;
		m = n;
		n = i;
		tmp = text;
		text = pattern;
		pattern = tmp;
	}
	
	int mb = (1 << (m-1));
	
	for (i= 0;i < m;i++){
		T[(int)pattern[i]] |= (1 << i);
	}
	
	for (i = 0;i < n;i++){
		s <<= 1;
		s |= 1;
		if(!text[i]){
			return -1;
		}
		Tc = T[(int)text[i]];
		s &= Tc;
		if(s & mb){
			return i+1;
		}
	}
	return 0;
}


